﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;

namespace AirFusion.Dandelion
{
	/// <summary>
	/// Provides the web API of Dandelion.
	/// </summary>
	class DandelionAPI
	{
		private string DAPI_TEMPLATE_DEVICEREGISTER = "/pair/v1/devices/";
		private string DAPI_TEMPLATE_DEVICEPAIRING = "/pair/v1/clash/";
		private string DAPI_TEMPLATE_DEVICEPROFILE = "/af/v1/profile/";

		private Logger _logger;

		public string _dandelionServerAddress;

		/// <summary>
		/// Initializes a new instance of <see cref="DandelionAPI" /> object.
		/// </summary>
		/// <param name="dandelionHost"> The host name of the web server. </param>
		/// <param name="dandelionPort"> Default is null. The port of the web server. </param>
		/// <param name="isSecure"> Default is true. The connection is secure or not. </param>
		/// <exception cref="ArgumentNullException"></exception>
		public DandelionAPI(string dandelionHost, ushort? dandelionPort = null, bool isSecure = true)
		{
			if (string.IsNullOrEmpty(dandelionHost))
			{
				throw new ArgumentNullException("dandelionServerHost");
			}

			this._dandelionServerAddress = (isSecure ? "https://" : "http://") + dandelionHost.Trim('/') + (dandelionPort.HasValue ? ":" + dandelionPort : "");

			string tmpTime = DateTime.Now.ToString("MMdd HH-mm-ss.fff");
			string logFileName = tmpTime + "_DandelionAPI.txt";
			this._logger = new Logger(Common.LogFolderPath + logFileName);
			_logger.WriteNewLine("=====");
		}


		#region Dandelion API for update/get profile

		/// <summary>
		/// Updates the profile asynchronously.
		/// </summary>
		/// <param name="deviceId"> The device ID. </param>
		/// <param name="instanceId"> The instance ID. </param>
		/// <param name="profile"> The profile. </param>
		public async Task<Dictionary<int, string>> UpdateProfileAsync(string deviceId, string instanceId, string profile)
		{
			_logger.WriteNewLine("[UpdateProfileAsync] profile = " + profile);

			JObject data = new JObject(
				new JProperty("device_id", deviceId),
				new JProperty("instance_id", instanceId),
				new JProperty("profile", profile)
				);

			string url = this._dandelionServerAddress + DAPI_TEMPLATE_DEVICEPROFILE;
			if (Common.EnableAPIKey)
			{
				url += "?" + Common.DandelionAirFusionAppKey;
			}

			url += "&instance_id=" + instanceId;
			
			//Posting the Data to Dandelion
			string response = await HttpUtils.SendHttpRequestAsync(Common.HttpMethodPost, url,
															Common.HttpContentTypeJson, Common.SocketTimeout, null,
															data.ToString()).ConfigureAwait(false);

			_logger.WriteNewLine("[UpdateProfileAsync] Response = " + response);
			Dictionary<int, string> tmpDC = new Dictionary<int, string>();
			tmpDC.Add(HttpUtils.LastResponseCode, HttpUtils.LastResponseMessage);
			return tmpDC;
		}


		/// <summary>
		/// Gets the profile asynchronously.
		/// </summary>
		/// <param name="deviceId"> The device ID. </param>
		/// <param name="instanceId"> The instance ID. </param>
		public async Task<Dictionary<Dictionary<int, string>, JObject>> GetProfileAsync(string deviceId,
																									string instanceId)
		{
			_logger.WriteNewLine("[GetProfileAsync] device_id = " + deviceId + ", instance_id = " + instanceId);
			Dictionary<Dictionary<int, string>, JObject> tmpDC = new Dictionary<Dictionary<int, string>, JObject>();

			string url = this._dandelionServerAddress + DAPI_TEMPLATE_DEVICEPROFILE + "?";

			if (!string.IsNullOrEmpty(deviceId))
			{
				url += "device_id=" + deviceId;
			}
			else if (!string.IsNullOrEmpty(instanceId))
			{
				url += "instance_id=" + instanceId;
			}

			if (Common.EnableAPIKey)
			{
				url += "&" + Common.DandelionAirFusionAppKey;
			}

			//Posting the Data to Dandelion
			string response = await HttpUtils.SendHttpRequestAsync(Common.HttpMethodGet, url, "", Common.SocketTimeout, null, null).ConfigureAwait(false);

			Dictionary<int, string> tmpStatusDC = new Dictionary<int, string>();
			if (!string.IsNullOrEmpty(response))
			{
				_logger.WriteNewLine("[GetProfileAsync] Response = " + response);
				JObject jObject = JsonConvert.DeserializeObject<JObject>(response);
				tmpStatusDC.Add(HttpUtils.LastResponseCode, HttpUtils.LastResponseMessage);
				tmpDC.Add(tmpStatusDC, jObject);
				return tmpDC;
			}
			tmpStatusDC.Add(HttpUtils.LastResponseCode, HttpUtils.LastResponseMessage);
			tmpDC.Add(tmpStatusDC, null);
			return tmpDC;
		}

		#endregion


		#region Dandelion API for register/unregister Device

		/// <summary>
		/// Registers the device to Dandelion with login.
		/// </summary>
		public string RegisterDeviceByLogin(string instanceId, string accessToken)
		{
			_logger.WriteNewLine("RegisterDeviceByLogin ..... ");

			return null;
		}

		/// <summary>
		/// Registered the device to Dandelion asynchronously.
		/// </summary>
		public async Task<string> RegisterDeviceAsync()
		{
			_logger.WriteNewLine("RegisterDeviceAsync ..... ");

			JObject data = new JObject();
			DeviceInfo u = new DeviceInfo();
			string softID = u._securityIdentifier;
			try
			{
				softID = u.GetSecurityIdentifier();
			}
			catch (Exception e)
			{
				_logger.WriteNewLine("[DandelionAPI][GetSecurityIdentifier] Exception: " + e.ToString());
			}
			data.Add("soft_id", softID);

			string _serial = u._deviceUuid;
			try
			{
				_serial = u.GetDeviceUUID();
			}
			catch (Exception e)
			{
				_logger.WriteNewLine("[DandelionAPI][GetDeviceUUID] Exception: " + e.ToString());
			}
			data.Add("serial", _serial);

			string _brand = u._deviceBrand;
			try
			{
				_brand = u.GetDeviceBrand();
			}
			catch (Exception e)
			{
				_logger.WriteNewLine("[DandelionAPI][GetDeviceBrand] Exception: " + e.ToString());
			}
			data.Add("brand", _brand);

			string model = u._deviceModelName;
			try
			{
				model = u.GetDeviceModel();
			}
			catch (Exception e)
			{
				_logger.WriteNewLine("[DandelionAPI][GetDeviceModel] Exception: " + e.ToString());
			}
			data.Add("model", model);

			string type = u._deviceChassisTypes;
			try
			{
				type = u.GetDeviceChassisTypes();
			}
			catch (Exception e)
			{
				_logger.WriteNewLine("[DandelionAPI][GetSecurityIdentifier] Exception: " + e.ToString());
			}
			data.Add("type", type);

			string url = this._dandelionServerAddress + DAPI_TEMPLATE_DEVICEREGISTER;
			if (Common.EnableAPIKey)
			{
				url += "?" + Common.DandelionPairAppKey;
			}

			string response = await HttpUtils.SendHttpRequestAsync(Common.HttpMethodPost, url,
															Common.HttpContentTypeJson, Common.SocketTimeout, null, data.ToString()).ConfigureAwait(false);

			_logger.WriteNewLine("[RegisterDeviceAsync] LastResponseCode = " + HttpUtils.LastResponseCode);

			if (!string.IsNullOrEmpty(response))
			{
				string device_id = string.Empty;
				JObject jObject = JObject.Parse(response);

				var msgProperty = jObject.Property("device_id");
				if (msgProperty != null)
				{
					device_id = msgProperty.Value.ToString();
				}
				_logger.WriteNewLine("[RegisterDeviceAsync] device_id = " + device_id);

				return device_id;
			}
			return null;
		}

		/// <summary>
		/// De-registers device from Dandelion asynchronous.
		/// </summary>
		/// <param name="deviceId">The device identifier.</param>
		/// <param name="accessToken">The access token.</param>
		/// <returns></returns>
		public async Task<bool> UnRegisterDeviceAsync(string deviceId, string accessToken, string refreshToken)
		{
			_logger.WriteNewLine("UnRegisterDeviceAsync ..... ");

			return false;
		}

		#endregion


		#region Dandelion API for start pairing/pairing request

		/// <summary>
		/// Starts pairing asynchronously.
		/// </summary>
		/// <param name="deviceId"> The device ID. </param>
		/// <param name="pairTimeout"> The time span that is hold on the server, in seconds. </param>
		/// <param name="bssidList"> The list of nearby devices. </param>
		public async Task<Dictionary<Dictionary<int, string>, List<string>>> pairingStartAsync(string deviceId,
																				long pairTimeout, List<string> bssidList)
		{
			_logger.WriteNewLine("pairingStartAsync ..... ");

			string tmpClashTime = DateTime.Now.ToString("yyyy-MM-dd'T'HH:mm:ss.fff'Z'", DateTimeFormatInfo.InvariantInfo);
			int tmpPairMethod = Common.PairMethodBSSID;

			JObject data = new JObject(
				new JProperty("device_id", deviceId),
				new JProperty("clash_time", tmpClashTime),
				new JProperty("pair_time", pairTimeout),
				new JProperty("bssid_info_list", bssidList),
				new JProperty("pair_method", tmpPairMethod)
				);

			string url = this._dandelionServerAddress + DAPI_TEMPLATE_DEVICEPAIRING;
			if (Common.EnableAPIKey)
			{
				url += "?" + Common.DandelionPairAppKey;
			}

			int timeout = (int)pairTimeout * 1000 + Common.SocketTimeout;

			//Posting the Data to Dandelion
			string response = await HttpUtils.SendHttpRequestAsync(Common.HttpMethodPost, url,
															Common.HttpContentTypeJson, timeout, null,
															data.ToString()).ConfigureAwait(false);
			Dictionary<Dictionary<int, string>, List<string>> tmpDC = new Dictionary<Dictionary<int, string>,
																										List<string>>();
			Dictionary<int, string> tmpStatusDC = new Dictionary<int, string>();
			if (!string.IsNullOrEmpty(response))
			{
				_logger.WriteNewLine("[pairingStartAsync] Response = " + response);

				string tmpDeviceId = string.Empty;
				string tmpType = string.Empty;
				string tmpTransactionId = string.Empty;
				List<string> nearbyDeviceList = new List<string>();

				JObject jObject = JObject.Parse(response);

				var deviceIDProperty = jObject.Property("device_id");
				if (deviceIDProperty != null)
				{
					tmpDeviceId = deviceIDProperty.Value.ToString();
				}

				var typeProperty = jObject.Property("type");
				if (typeProperty != null)
				{
					tmpType = typeProperty.Value.ToString();
				}

				var transactionIDProperty = jObject.Property("transaction_id");
				if (transactionIDProperty != null)
				{
					tmpTransactionId = transactionIDProperty.Value.ToString();
				}

				var deviceListProperty = jObject.Property("candidate");
				if (deviceListProperty != null)
				{
					nearbyDeviceList = JsonConvert.DeserializeObject<List<string>>(jObject.SelectToken("candidate").ToString());
				}

				tmpStatusDC.Add(HttpUtils.LastResponseCode, HttpUtils.LastResponseMessage);
				tmpDC.Add(tmpStatusDC, nearbyDeviceList);
				return tmpDC;
			}
			tmpStatusDC.Add(HttpUtils.LastResponseCode, HttpUtils.LastResponseMessage);
			tmpDC.Add(tmpStatusDC, null);
			return tmpDC;
		}

		/// <summary>
		/// Requests pairing asynchronously.
		/// </summary>
		/// <param name="deviceId"> The device ID. </param>
		/// <param name="pairTimeout"> The time span that is hold on the server, in seconds. </param>
		/// <param name="instanceId"> The instance ID of target device. </param>
		public async Task<Dictionary<Dictionary<int, string>, string>> pairingRequestAsync(string deviceId,
																				long pairTimeout, string instanceId)
		{
			string tmpClashTime = DateTime.Now.ToString("yyyy-MM-dd'T'HH:mm:ss.fff'Z'", DateTimeFormatInfo.InvariantInfo);
			int tmpPairMethod = Common.PairMethodSonudTrigger;

			JObject data = new JObject(
				new JProperty("device_id", deviceId),
				new JProperty("clash_time", tmpClashTime),
				new JProperty("pair_time", pairTimeout),
				new JProperty("soundtrigger_key", instanceId),
				new JProperty("pair_method", tmpPairMethod)
				);

			Dictionary<Dictionary<int, string>, string> tmpDC = new Dictionary<Dictionary<int, string>, string>();

			string url = this._dandelionServerAddress + DAPI_TEMPLATE_DEVICEPAIRING;
			if (Common.EnableAPIKey)
			{
				url += "?" + Common.DandelionPairAppKey;
			}

			int timeout = (int)pairTimeout * 1000 + Common.SocketTimeout;

			//Posting the Data to Dandelion
			string response = await HttpUtils.SendHttpRequestAsync(Common.HttpMethodPost, url,
															Common.HttpContentTypeJson, timeout, null, data.ToString()).ConfigureAwait(false);
			Dictionary<int, string> tmpStatusDC = new Dictionary<int, string>();

			if (!string.IsNullOrEmpty(response))
			{
				_logger.WriteNewLine("[pairingRequestAsync] Response = " + response);

				string tmpTargetDeviceId = string.Empty;

				JObject jObject = JObject.Parse(response);

				var deviceIDProperty = jObject.Property("device_id");
				if (deviceIDProperty != null)
				{
					tmpTargetDeviceId = deviceIDProperty.Value.ToString();
				}

				tmpStatusDC.Add(HttpUtils.LastResponseCode, HttpUtils.LastResponseMessage);
				tmpDC.Add(tmpStatusDC, tmpTargetDeviceId);
				return tmpDC;
			}

			tmpStatusDC.Add(HttpUtils.LastResponseCode, HttpUtils.LastResponseMessage);
			tmpDC.Add(tmpStatusDC, null);
			return tmpDC;
		}
		#endregion


		#region Dandelion API for notification/notification icon/ notification action

		/// <summary>
		/// Gets the notification.
		/// </summary>
		/// <param name="url">The URL.</param>
		/// <param name="accessToken">The access token.</param>
		/// <param name="refreshToken">The refresh token.</param>
		/// <returns></returns>
		public string GetNotification(string url, string accessToken, string refreshToken)
		{
			_logger.WriteNewLine("GetNotification ..... ");

			return null;
		}

		/// <summary>
		/// Gets the notification icon.
		/// </summary>
		/// <param name="key">The key.</param>
		/// <param name="accessToken">The access token.</param>
		/// <param name="refreshToken">The refresh token.</param>
		/// <returns></returns>
		public async Task<string> GetNotificationIcon(string key, string accessToken, string refreshToken)
		{
			_logger.WriteNewLine("GetNotificationIcon ..... ");

			return null;
		}

		/// <summary>
		/// Posts the notification action.
		/// </summary>
		/// <param name="targetDeviceID">The target device identifier.</param>
		/// <param name="localDeviceID">The local device identifier.</param>
		/// <param name="actionBody">The action body.</param>
		/// <param name="accessToken">The access token.</param>
		/// <param name="refreshToken">The refresh token.</param>
		/// <returns></returns>
		public async Task<Dictionary<int, string>> PostNotificationAction(string targetDeviceID, string localDeviceID, string actionBody, string accessToken, string refreshToken)
		{
			_logger.WriteNewLine("PostNotificationAction ..... ");

			return null;
		}

		#endregion


		#region Dandelion API for refresh token

		/// <summary>
		/// Refresh the token.
		/// </summary>
		public string RefreshAccessToken(string refreshToken)
		{
			_logger.WriteNewLine("RefreshAccessToken ..... ");

			return null;
		}

		#endregion


		#region Dandelion API for PnServices

		/// <summary>
		/// Registers the device to PnServices.
		/// </summary>
		public Dictionary<int, string> RegisterDeviceForPnServices(string deviceId, string userId)
		{
			_logger.WriteNewLine("RegisterDeviceForPnServices ..... ");

			return null;
		}

		#endregion
	}
}
