﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Microsoft.Win32;
using Newtonsoft.Json.Linq;
using System.Threading;
//using System.Threading.Tasks;

namespace AirFusion.Dandelion
{
	#region Thread Process for update/get profile

	/// <summary>
	/// The process of update device profile.
	/// </summary>
	class ProcessWithUpdateDeviceProfile
	{
		private string _deviceId;
		private string _instanceId;
		private string _profile;

		private string _dandelionHost;
		private ushort? _dandelionPort;
		private bool _isSecure;

		/// <summary>
		/// Initializes a new instance of <see cref="ProcessWithUpdateDeviceProfile" /> object.
		/// </summary>
		/// <param name="cloudHost"> The host name of the web server. </param>
		/// <param name="cloudPort"> Default is null. The port of the web server. </param>
		/// <param name="isSecure"> Default is true. The connection is secure or not. </param>
		public ProcessWithUpdateDeviceProfile(string cloudHost, ushort? cloudPort, bool isSecure)
		{
			this._dandelionHost = cloudHost;
			this._dandelionPort = cloudPort;
			this._isSecure = isSecure;
		}

		/// <summary>
		/// Sets the property of <see cref="ProcessWithUpdateDeviceProfile" /> object.
		/// </summary>
		/// <param name="deviceId"> The device ID of local device. </param>
		/// <param name="instanceId"> The instance ID of specific device. </param>
		/// <param name="profile"> The profile of specific device. </param>
		public void SetUpdateDeviceProfileFields(string deviceId, string instanceId, string profile)
		{
			this._deviceId = deviceId;
			this._instanceId = instanceId;
			this._profile = profile;
		}

		/// <summary>
		/// Updates device profile asynchronously.
		/// </summary>
		public async Task<AODCommon.UpdateDeviceProfileResult> UpdateDeviceProfileProcAsync()
		{
			AODCommon.UpdateDeviceProfileResult taskResult = new AODCommon.UpdateDeviceProfileResult();

			if (string.IsNullOrEmpty(this._deviceId))
			{
				taskResult.Profile = null;
				taskResult.IsSuccess = false;
				taskResult.StatusCode = AODCommon.DandelionStatusCode.ControllerStatusCodeInternetFailure;
				taskResult.Exception = new Exception("The operation is failed, might due to the initialization hasn't finished.");
				return taskResult;
			}

			DandelionAPI tmpAPI = new DandelionAPI(this._dandelionHost, this._dandelionPort, this._isSecure);
			Dictionary<int, string> tmpDC = await tmpAPI.UpdateProfileAsync(this._deviceId, this._instanceId,
																								this._profile).ConfigureAwait(false);
			if (tmpDC != null)
			{
				KeyValuePair<int, string> tmpKV = tmpDC.ElementAt(0);

				taskResult.Profile = null;
				if (tmpKV.Key == (int)Common.HttpStatusCode.HttpStatusCodeOK)
				{
					taskResult.IsSuccess = true;
				}
				else
				{
					taskResult.IsSuccess = false;
					taskResult.StatusCode = (AODCommon.DandelionStatusCode)tmpKV.Key;
					taskResult.Exception = new Exception("DandelionServer fail return:" + tmpKV.Value + "(" + tmpKV.Key + ")");
				}
			}
			else
			{
				taskResult.Profile = null;
				taskResult.IsSuccess = false;
				taskResult.StatusCode = AODCommon.DandelionStatusCode.ControllerStatusCodeServerFailure;
				taskResult.Exception = new Exception("DandelionServer fail: Unexpected error.");
			}

			return taskResult;
		}
	}

	/// <summary>
	/// Provides the process of profile retrieval.
	/// </summary>
	class ProcessWithGetDeviceProfile
	{
		private string _instanceId;

		private string _dandelionHost;
		private ushort? _dandelionPort;
		private bool _isSecure;

		/// <summary>
		/// Initializes a new instance of <see cref="ProcessWithGetDeviceProfile" /> object.
		/// </summary>
		/// <param name="cloudHost"> The host name of the web server. </param>
		/// <param name="cloudPort"> Default is null. The port of the web server. </param>
		/// <param name="isSecure"> Default is true. The connection is secure or not. </param>
		public ProcessWithGetDeviceProfile(string cloudHost, ushort? cloudPort, bool isSecure)
		{
			this._dandelionHost = cloudHost;
			this._dandelionPort = cloudPort;
			this._isSecure = isSecure;
		}

		/// <summary>
		/// Sets the property of <see cref="ProcessWithGetDeviceProfile" /> object.
		/// </summary>
		/// <param name="instanceId"> The instance ID of specific device. </param>
		public void SetUpdateDeviceProfileFields(string instanceId)
		{
			this._instanceId = instanceId;
		}

		/// <summary>
		/// Gets device profile asynchronously.
		/// </summary>
		public async Task<AODCommon.GetDeviceProfileResult> GetDeviceProfileProcAsync()
		{
			DandelionAPI tmpAPI = new DandelionAPI(this._dandelionHost, this._dandelionPort, this._isSecure);
			Dictionary<Dictionary<int, string>, JObject> tmpDC = await tmpAPI.GetProfileAsync(null, _instanceId).ConfigureAwait(false);

			AODCommon.GetDeviceProfileResult task = new AODCommon.GetDeviceProfileResult();
			if (tmpDC != null)
			{
				KeyValuePair<Dictionary<int, string>, JObject> tmpKV = tmpDC.ElementAt(0);
				KeyValuePair<int, string> tmpStatusKV = tmpKV.Key.ElementAt(0);
				if (tmpKV.Value != null)
				{
					var msgProperty = tmpKV.Value.Property("profile");
					if (msgProperty != null)
					{
						task.Profile = msgProperty.Value.ToString();
					}
					else
					{
						task.Profile = null;
					}

					if (tmpStatusKV.Key == (int)Common.HttpStatusCode.HttpStatusCodeOK)
					{
						task.IsSuccess = true;
					}
					else
					{
						task.IsSuccess = false;
						task.StatusCode = (AODCommon.DandelionStatusCode)tmpStatusKV.Key;
						task.Exception = new Exception("DandelionServer fail return:" + tmpStatusKV.Value + "(" + tmpStatusKV.Key + ")");
					}
				}
				else
				{
					task.Profile = null;
					task.IsSuccess = false;
					task.StatusCode = (AODCommon.DandelionStatusCode)tmpStatusKV.Key;
					task.Exception = new Exception("DandelionServer fail return:" + tmpStatusKV.Value + "(" + tmpStatusKV.Key + ")");
				}
			}
			else
			{
				task.Profile = null;
				task.IsSuccess = false;
				task.StatusCode = AODCommon.DandelionStatusCode.ControllerStatusCodeServerFailure;
				task.Exception = new InvalidOperationException("DandelionServer fail: Unexpected error.");
			}

			return task;
		}
	}

	#endregion


	#region Thread Process for pairing

	/// <summary>
	/// Provide the process of pairing start.
	/// </summary>
	class ProcessWithPairingStart
	{
		private string _deviceId;
		private string _instanceId;
		private long _pairingTimes;
		private AODCommon.PairingEventHandler _startPairingEventHandler;

		private string _dandelionHost;
		private ushort? _dandelionPort;
		private bool _isSecure;

		/// <summary>
		/// Initializes a new instance of <see cref="ProcessWithPairingStart" /> object.
		/// </summary>
		/// <param name="cloudHost"> The host name of the web server. </param>
		/// <param name="cloudPort"> Default is null. The port of the web server. </param>
		/// <param name="isSecure"> Default is true. The connection is secure or not. </param>
		public ProcessWithPairingStart(string cloudHost, ushort? cloudPort, bool isSecure)
		{
			this._dandelionHost = cloudHost;
			this._dandelionPort = cloudPort;
			this._isSecure = isSecure;
		}

		/// <summary>
		/// Sets the property of <see cref="ProcessWithPairingStart" /> object.
		/// </summary>
		/// <param name="deviceId"> The device ID of local device. </param>
		/// <param name="instanceId"> The instance ID of specific device. </param>
		/// <param name="pairingTimes"> The time span that is hold on the server, in seconds. </param>
		/// <param name="startPairingEventHandler"> The near devices' profile callback. </param>
		public void SetPairingStartFields(string deviceId, string instanceId, long pairingTimes,
															AODCommon.PairingEventHandler startPairingEventHandler)
		{
			this._deviceId = deviceId;
			this._instanceId = instanceId;
			this._pairingTimes = pairingTimes;
			this._startPairingEventHandler = startPairingEventHandler;
		}

		/// <summary>
		/// Starts pairing process asynchronously.
		/// </summary>
		public async Task PairingStartProcAsync(bool isAdvertiser)
		{
			if (string.IsNullOrEmpty(this._deviceId) || this._deviceId.Equals("null"))
			{
				if (this._startPairingEventHandler != null)
				{
					AODCommon.PairingResultEventArgs argsError = new AODCommon.PairingResultEventArgs();
					argsError.IsSuccess = false;
					argsError.InstanceID = string.Empty;
					argsError.Profile = string.Empty;
					argsError.IsCompleted = true;
					argsError.StatusCode = AODCommon.DandelionStatusCode.ControllerStatusCodeInternetFailure;
					argsError.Exception = new InvalidOperationException("The pairing start fail, because the internet error.");
					this._startPairingEventHandler(this, argsError);
				}
				return;
			}

			List<string> bssidList = new List<string>();
			try
			{
				WirelessUtils w = new WirelessUtils();
				List<byte[]> tmpList;
				w.GetNearbyWirelessAccessPoints(out tmpList);
				if (tmpList != null && tmpList.Count > 0)
				{
					foreach (byte[] item in tmpList)
					{
						//byte convert to string
						var macAddrLen = (uint)item.Length;
						var str = new string[(int)macAddrLen];
						for (int i = 0; i < macAddrLen; i++)
						{
							str[i] = item[i].ToString("x2");
						}
						string mac = string.Join("", str);
						string macNew = mac;
						int insertedCount = 0;
						for (int i = 2; i < mac.Length; i = i + 2)
						{
							macNew = macNew.Insert(i + insertedCount++, ":");
						}
						bssidList.Add(macNew);
					}
				}
			}
			catch (Exception e)
			{
				if (this._startPairingEventHandler != null)
				{
					AODCommon.PairingResultEventArgs argsError = new AODCommon.PairingResultEventArgs();
					argsError.IsSuccess = false;
					argsError.IsCompleted = true;
					argsError.InstanceID = string.Empty;
					argsError.Profile = string.Empty;
					argsError.StatusCode = AODCommon.DandelionStatusCode.ControllerStatusCodeWifiFailure;
					argsError.Exception = new Exception("Failed getting WiFi information.", e);
					this._startPairingEventHandler(this, argsError);
				}
				return;
			}

			DandelionAPI tmpAPI = new DandelionAPI(this._dandelionHost, this._dandelionPort, this._isSecure);

			Dictionary<Dictionary<int, string>, List<string>> tmpDC;
			tmpDC = await tmpAPI.pairingStartAsync(this._deviceId, this._pairingTimes, bssidList).ConfigureAwait(false);

			AODCommon.PairingResultEventArgs args = new AODCommon.PairingResultEventArgs();
			if (tmpDC != null)
			{
				KeyValuePair<Dictionary<int, string>, List<string>> tmpKV = tmpDC.ElementAt(0);
				KeyValuePair<int, string> tmpStatusKV = tmpKV.Key.ElementAt(0);

				if (!isAdvertiser && tmpKV.Value != null)
				{
					ThreadWithGetNearDeviceProfile tmpGetNearDeviceProfile = new ThreadWithGetNearDeviceProfile
																									(this._dandelionHost,
																									this._dandelionPort,
																									this._isSecure,
																									tmpKV.Value,
																						this._startPairingEventHandler);

					tmpGetNearDeviceProfile.GetNearDeviceProfileProcAsync().ConfigureAwait(false);
					args.IsSuccess = true;
					args.IsCompleted = false;
				}
				else
				{
					if (tmpStatusKV.Key == (int)Common.HttpStatusCode.HttpStatusCodeOK ||
						tmpStatusKV.Key == (int)Common.HttpStatusCode.HttpStatusCodeNoContent)
					{
						args.IsSuccess = true;
					}
					else
					{
						args.IsSuccess = false;
						args.StatusCode = (AODCommon.DandelionStatusCode)tmpStatusKV.Key;
						args.Exception = new Exception("DandelionServer fail return: " + tmpStatusKV.Value + "(" + tmpStatusKV.Key + ")");
					}
					args.IsCompleted = true;
				}
			}
			else
			{
				args.IsSuccess = false;
				args.StatusCode = AODCommon.DandelionStatusCode.ControllerStatusCodeServerFailure;
				args.Exception = new Exception("DandelionServer fail: Unexpected error.");
				args.IsCompleted = true;
			}

			if (this._startPairingEventHandler != null)
			{
				args.InstanceID = string.Empty;
				args.Profile = string.Empty;
				this._startPairingEventHandler(this, args);
			}
		}
	}

	/// <summary>
	/// Provides the process of pairing request.
	/// </summary>
	class ProcessWithPairingRequest
	{
		private string _deviceId;
		private string _instanceId;

		private string _dandelionHost;
		private ushort? _dandelionPort;
		private bool _isSecure;

		/// <summary>
		/// Initializes a new instance of <see cref="ProcessWithPairingRequest" /> object.
		/// </summary>
		/// <param name="cloudHost"> The host name of the web server. </param>
		/// <param name="cloudPort"> Default is null. The port of the web server. </param>
		/// <param name="isSecure"> Default is true. The connection is secure or not. </param>
		public ProcessWithPairingRequest(string cloudHost, ushort? cloudPort, bool isSecure)
		{
			this._dandelionHost = cloudHost;
			this._dandelionPort = cloudPort;
			this._isSecure = isSecure;
		}

		/// <summary>
		/// Sets the property of <see cref="ProcessWithPairingRequest" /> object.
		/// </summary>
		/// <param name="deviceId"> The device ID of local device. </param>
		/// <param name="instanceId"> The instance ID of specific device. </param>
		public void SetPairingRequestFields(string deviceId, string instanceId)
		{
			this._deviceId = deviceId;
			this._instanceId = instanceId;
		}

		/// <summary>
		/// Requests pairing asynchronously.
		/// </summary>
		public async Task<AODCommon.PairingInvitationTaskResult> PairingRequestProcAsync()
		{
			if (string.IsNullOrEmpty(this._deviceId))
			{
				throw (new AODException("The pairing request fail, because the internet error.",
																					AODExceptionStatus.ERR_Internet));
			}

			DandelionAPI tmpAPI = new DandelionAPI(this._dandelionHost, this._dandelionPort, this._isSecure);
			Dictionary<Dictionary<int, string>, string> tmpDC = await tmpAPI.pairingRequestAsync(this._deviceId, Common.DefaultHoldTime,
																									this._instanceId).ConfigureAwait(false);

			AODCommon.PairingInvitationTaskResult task = new AODCommon.PairingInvitationTaskResult();
			if (tmpDC != null)
			{
				KeyValuePair<Dictionary<int, string>, string> tmpKV = tmpDC.ElementAt(0);
				KeyValuePair<int, string> tmpStatusKV = tmpKV.Key.ElementAt(0);

				if (tmpStatusKV.Key == (int)Common.HttpStatusCode.HttpStatusCodeOK)
				{
					task.IsSuccess = true;
				}
				else
				{
					task.IsSuccess = false;
					task.StatusCode = (AODCommon.DandelionStatusCode)tmpStatusKV.Key;
					task.Exception = new Exception("DandelionServer fail return:" + tmpStatusKV.Value + "(" + tmpStatusKV.Key + ")");
				}
			}
			else
			{
				task.IsSuccess = false;
				task.StatusCode = AODCommon.DandelionStatusCode.ControllerStatusCodeServerFailure;
				task.Exception = new Exception("DandelionServer fail: Unexpected error.");
			}

			return task;
		}
	}

	#endregion


	#region Thread Process for send message/get notification icon

	/// <summary>
	/// Provides the process of notification.
	/// </summary>
	class ProcessWithGetNotificationIcon
	{
		private string _dandelionHost;
		private ushort? _dandelionPort;
		private bool _isSecure;

		/// <summary>
		/// Initializes a new instance of <see cref="ProcessWithGetNotificationIcon" /> object.
		/// </summary>
		/// <param name="cloudHost"> The host name of the web server. </param>
		/// <param name="cloudPort"> Default is null. The port of the web server. </param>
		/// <param name="isSecure"> Default is true. The connection is secure or not. </param>
		public ProcessWithGetNotificationIcon(string cloudHost, ushort? cloudPort = null, bool isSecure = true)
		{
			this._dandelionHost = cloudHost;
			this._dandelionPort = cloudPort;
			this._isSecure = isSecure;
		}

		/// <summary>
		/// Gets notification icon asynchronously.
		/// </summary>
		/// <param name="accessToken">The access token.</param>
		/// <param name="refreshToken">The refresh token.</param>
		/// <param name="key">The key.</param>
		/// <returns></returns>
		public async Task<string> GetNotificationIconProcAsync(string accessToken, string refreshToken, string key)
		{
			DandelionAPI tmpAPI = new DandelionAPI(this._dandelionHost, this._dandelionPort, this._isSecure);
			string response = await tmpAPI.GetNotificationIcon(key, accessToken, refreshToken).ConfigureAwait(false);
			return response;
		}
	}

	/// <summary>
	/// Provides the process of send message.
	/// </summary>
	class ProcessWithSendMessage
	{
		private string _dandelionHost;
		private ushort? _dandelionPort;
		private bool _isSecure;

		/// <summary>
		/// Initializes a new instance of <see cref="ProcessWithSendMessage" /> object.
		/// </summary>
		/// <param name="cloudHost"> The host name of the web server. </param>
		/// <param name="cloudPort"> Default is null. The port of the web server. </param>
		/// <param name="isSecure"> Default is true. The connection is secure or not. </param>
		public ProcessWithSendMessage(string cloudHost, ushort? cloudPort = null, bool isSecure = true)
		{
			this._dandelionHost = cloudHost;
			this._dandelionPort = cloudPort;
			this._isSecure = isSecure;
		}

		/// <summary>
		/// Sends message asynchronously.
		/// </summary>
		/// <param name="accessToken">The access token.</param>
		/// <param name="refreshToken">The refresh token.</param>
		/// <param name="instanceID">The instance identifier.</param>
		/// <param name="message">The message.</param>
		/// <returns></returns>
		public async Task<AODCommon.SendMessageTaskResult> SendMessageProcAsync(string accessToken, string refreshToken, string localDeviceID, string instanceID, string message)
		{
			DandelionAPI tmpAPI = new DandelionAPI(this._dandelionHost, this._dandelionPort, this._isSecure);
			Dictionary<Dictionary<int, string>, JObject> tmpDC = await tmpAPI.GetProfileAsync(null, instanceID).ConfigureAwait(false);

			AODCommon.SendMessageTaskResult task = new AODCommon.SendMessageTaskResult();
			if (tmpDC != null)
			{
				KeyValuePair<Dictionary<int, string>, JObject> tmpKV = tmpDC.ElementAt(0);
				KeyValuePair<int, string> tmpStatusKV = tmpKV.Key.ElementAt(0);
				if (tmpKV.Value != null)
				{
					string targetDeviceID = string.Empty;

					var idProperty = tmpKV.Value.Property("device_id");
					if (idProperty != null)
					{
						targetDeviceID = idProperty.Value.ToString();
					}

					if (!string.IsNullOrEmpty(targetDeviceID))
					{
						Dictionary<int, string> sendMessageDC = await tmpAPI.PostNotificationAction(targetDeviceID, localDeviceID, message, accessToken, refreshToken).ConfigureAwait(false);

						if (sendMessageDC != null)
						{
							KeyValuePair<int, string> sendMessageKV = sendMessageDC.ElementAt(0);
							if (sendMessageKV.Key == (int)Common.HttpStatusCode.HttpStatusCodeOK)
							{
								task.IsSuccess = true;
							}
							else
							{
								task.IsSuccess = false;
								task.Exception = new Exception("DandelionServer fail return:" + sendMessageKV.Value + "(" + sendMessageKV.Key + ")");
							}
						}
					}
					else
					{
						task.IsSuccess = false;
						task.Exception = new Exception("DandelionServer fail return:" + tmpStatusKV.Value + "(" + tmpStatusKV.Key + ")");
					}
				}
				else
				{
					task.IsSuccess = false;
					task.Exception = new Exception("DandelionServer fail return:" + tmpStatusKV.Value + "(" + tmpStatusKV.Key + ")");
				}
			}
			else
			{
				task.IsSuccess = false;
				task.Exception = new InvalidOperationException("DandelionServer fail: Unexpected error.");
			}

			return task;
		}
	}

	#endregion


	/// <summary>
	/// Provides the process of device ID retrieval.
	/// </summary>
	class ProcessWithGetDeviceID
	{
		/// <summary>
		/// Gets the device ID asynchronously.
		/// </summary>
		/// <param name="cloudHost"> The host name of the web server. </param>
		/// <param name="cloudPort"> Default is null. The port of the web server. </param>
		/// <param name="isSecure"> Default is true. The connection is secure or not. </param>
		public async Task<string> GetDeviceIDProcAsync(string cloudHost, ushort? cloudPort, bool isSecure)
		{
			DandelionAPI tmpAPI = new DandelionAPI(cloudHost, cloudPort, isSecure);
			return await tmpAPI.RegisterDeviceAsync().ConfigureAwait(false);
		}
	}


	/// <summary>
	/// The thread of get device's profile.
	/// </summary>
	class ThreadWithGetNearDeviceProfile
	{
		private string _dandelionHost;
		private ushort? _dandelionPort;
		private bool _isSecure;

		private List<string> _deviceList;
		private AODCommon.PairingEventHandler _startPairingEventHandler;

		/// <summary>
		/// Initializes a new instance of <see cref="ThreadWithGetNearDeviceProfile" /> object.
		/// </summary>
		/// <param name="cloudHost"> The host name of the web server. </param>
		/// <param name="cloudPort"> Default is null. The port of the web server. </param>
		/// <param name="isSecure"> Default is true. The connection is secure or not. </param>
		/// <param name="deviceList"> The list of nearby device </param>
		/// <param name="startPairingEventHandler"> The near devices' profile callback. </param>
		public ThreadWithGetNearDeviceProfile(string cloudHost, ushort? cloudPort, bool isSecure, List<string> deviceList,
															AODCommon.PairingEventHandler startPairingEventHandler)
		{
			this._dandelionHost = cloudHost;
			this._dandelionPort = cloudPort;
			this._isSecure = isSecure;

			this._deviceList = deviceList;
			this._startPairingEventHandler = startPairingEventHandler;
		}

		/// <summary>
		///  Get nearby device profile asynchronously.
		/// </summary>
		public async Task GetNearDeviceProfileProcAsync()
		{
			if (this._deviceList != null && this._deviceList.Count > 0)
			{
				int items = this._deviceList.Count;

				DandelionAPI tmpAPI = new DandelionAPI(this._dandelionHost, this._dandelionPort, this._isSecure);
				foreach (string tmpDeviceItem in this._deviceList)
				{
					AODCommon.PairingResultEventArgs args = new AODCommon.PairingResultEventArgs();
					KeyValuePair<Dictionary<int, string>, JObject> tmpDeviceKV = new
																		KeyValuePair<Dictionary<int, string>, JObject>();
					Dictionary<Dictionary<int, string>, JObject> tmpDeviceDC = await tmpAPI.GetProfileAsync(tmpDeviceItem, "").ConfigureAwait(false);

					if (tmpDeviceDC != null)
					{
						tmpDeviceKV = tmpDeviceDC.ElementAt(0);
						KeyValuePair<int, string> tmpStatusKV = tmpDeviceKV.Key.ElementAt(0);
						if (tmpDeviceKV.Value != null)
						{
							string tmpInstanceId = string.Empty;
							string tmpProfile = string.Empty;

							var instanceIDProperty = tmpDeviceKV.Value.Property("instance_id");
							if (instanceIDProperty != null)
							{
								tmpInstanceId = instanceIDProperty.Value.ToString();
							}

							var profileProperty = tmpDeviceKV.Value.Property("profile");
							if (profileProperty != null)
							{
								tmpProfile = profileProperty.Value.ToString();
							}

							args.IsSuccess = true;
							args.InstanceID = tmpInstanceId;
							args.Profile = tmpProfile;
						}
						else
						{
							args.IsSuccess = false;
							args.InstanceID = string.Empty;
							args.Profile = string.Empty;
							args.StatusCode = (AODCommon.DandelionStatusCode)tmpStatusKV.Key;
							args.Exception = new Exception("DandelionServer fail return: " + tmpStatusKV.Value + "(" + tmpStatusKV.Key + ")");
						}
					}

					args.IsCompleted = true;
					items--;
					if (items > 0)
					{
						args.IsCompleted = false;
					}

					if (this._startPairingEventHandler != null)
					{
						this._startPairingEventHandler(this, args);
					}
				}
			}
		}
	}

}
