﻿using AirFusion.Common;
using AirFusion.Manager.DeviceManager;
using AirFusion.Manager.FeatureHelper;
using AirFusion.WebService.Adapters;
using System;
using System.Diagnostics;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.ConnectionManager
{
	internal class CheckRemoteControlStatusManager : ConnectionInterface
	{
		#region define some const value

		public RemoteDevicePermission remotedevicepermission = RemoteDevicePermission.HavePermission;

		private static CheckRemoteControlStatusManager _pCheckRemoteControlStatusManager = null;
		private static object _pInstanceLock = new object();
		private ConnectControlStatusType connStatusType;
		private bool IsDisconnect = false;

		#endregion define some const value

		private Process _ADBdaemonprocess = null;

		public Process ADBdaemonprocess
		{
			get
			{
				return _ADBdaemonprocess;
			}

			set
			{
				_ADBdaemonprocess = value;
			}
		}

		#region public method

		public event EventHandler<ConnectEventArgs> ConnectionFlowEvent;

		public event EventHandler<AppEventArgs> ConnectStepEvent;

		/// <summary>
		/// For avoid CheckRemoteDeviceStatusManager double create.
		/// </summary>
		public static CheckRemoteControlStatusManager Instance
		{
			get
			{
				if (_pCheckRemoteControlStatusManager == null)
				{
					lock (_pInstanceLock)
					{
						if (_pCheckRemoteControlStatusManager == null) _pCheckRemoteControlStatusManager = new CheckRemoteControlStatusManager();
					}
				}
				return _pCheckRemoteControlStatusManager;
			}
		}

		private CheckRemoteControlStatusManager()
		{
			this.connStatusType = ConnectControlStatusType.None;
		}

		public void Connect(AppDeviceInfo deviceInfo, FlowControlManager controlmanager)
		{
			LogtoFile.TrackEnter();

			this.IsDisconnect = false;

			//First need to check remote device permission
			this.remotedevicepermission = this.CheckRemoteDeviceSystemPermission(deviceInfo.Device);

			//Set Screencast Window Info
			controlmanager._currentScreenWindow.SetInitialScreencastWindowInfo(deviceInfo, remotedevicepermission);

			//Reset connStatusType to default value
			this.connStatusType = ConnectControlStatusType.None;

			#region Try to Connect to Remote Device Input Control Service

			if (this.IsDisconnect)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Before ConnecRemoteInputControl, because the user has been disconnect screencast, we should stop to connect quickly.", "");
				return;
			}

			///Check AcerEXTEND can connecton to remote device input control service,
			///if Yes, It will  return ConnectionControlStatusType.InputControlConnectSuccess
			///   No,  It will return Error code=>ConnectionControlStatusType.InputControlConnectFail or ConnectionControlStatusType.InitialInputControlFail
			ConnectControlStatusType connectinputControlStatus = this.ConnecRemoteInputControl(deviceInfo, controlmanager);

			this.connStatusType = connectinputControlStatus;

			//check Success or not
			if (connectinputControlStatus == ConnectControlStatusType.InputControlConnectSuccess)
			{
				///Check remote device input daemon control status.
				///In normal case, it will return ConnectionControlStatusType.InputDaemonRunning
				///If fail, it will return ConnectControlStatusType.InputDaemonNotRunning or
				///ConnectControlStatusType.CheckInputDaemonTimeout
				///If return ConnectControlStatusType.CheckInputDaemonWebException, need to double check webservice fucntion.
				///If return ConnectControlStatusType.CheckInputDaemonException, represent the function return unexpect exception, need to double check

				#region Try to check remote device input daemon status

				this.CheckRemoteInputDaemonStatus(deviceInfo).ContinueWith(delegate(Task<ConnectControlStatusType> CheckInputDaemontask)
				{
					ConnectControlStatusType inputdaemonStatus = CheckInputDaemontask.Result;

					this.connStatusType = inputdaemonStatus;

					if (inputdaemonStatus == ConnectControlStatusType.InputDaemonRunning)
					{
						#region Try to check remote device Screencast daemon status

						///Check remote device screencast daemon control status.
						///In normal case, it will return ConnectionControlStatusType.ScreencastDaemonRunning
						///If fail, it will return ConnectControlStatusType.ScreencastDaemonNotRunning or
						///ConnectControlStatusType.CheckScreencastDaemonTimeout
						///If return ConnectControlStatusType.CheckScreencastDaemonWebException, need to double check webservice fucntion.
						///If return ConnectControlStatusType.CheckScreencastDaemonException, represent the function return unexpect exception, need to double check

						this.CheckRemoteScreencastDaemonStatus(deviceInfo).ContinueWith(delegate(Task<ConnectControlStatusType> CheckScreencastDaemontask)
						{
							ConnectControlStatusType screencastStatus = CheckScreencastDaemontask.Result;

							this.connStatusType = screencastStatus;

							if (screencastStatus == ConnectControlStatusType.ScreencastDaemonRunning)
							{
								this.Connected();
							}
							else
							{
								this.Disconnect(deviceInfo);
							}
						}).ConfigureAwait(continueOnCapturedContext: false);

						#endregion Try to check remote device Screencast daemon status
					}
					else
					{
						this.Disconnect(deviceInfo);
					}
				}).ConfigureAwait(continueOnCapturedContext: false);

				#endregion Try to check remote device input daemon status
			}
			else
			{
				this.Disconnect(deviceInfo);
			}

			#endregion Try to Connect to Remote Device Input Control Service
		}

		/// <summary>
		/// When Check Remote control finish, need to create a call back funciton to info the result
		/// </summary>
		public void Connected()
		{
			this.SendConnectionEvent(new ConnectEventArgs(ConnectionEventType.CheckRemoteControlStatusSuccess));
		}

		/// <summary>
		/// The Function will try to disconnecting about connction work. Because in the Manager is handle Check
		/// remote device control manager.
		/// </summary>
		public void Disconnect(AppDeviceInfo deviceInfo, bool IsAuto = true)
		{
			this.IsDisconnect = true;

			//For Non-Acer Device Use
			if (this.remotedevicepermission == RemoteDevicePermission.NoPermission)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Try to Kill ADB daemon process", "");
				CleanProcess(this.ADBdaemonprocess);
			}

			LogtoFile.TrackEnter();
			DisconnectionRemoteInputControl();
			this.Disconnected(IsAuto);
			LogtoFile.TrackLeave();
		}

		/// <summary>
		/// When Disconnected finished, need to crate a call back function to info the result.
		/// </summary>
		/// <param name="deviceInfoStatus"></param>
		public void Disconnected(bool IsAuto)
		{
			LogtoFile.TrackEnter();
			//Need to send event to info UI about disconnected is success.
			if (IsAuto)
				this.SendConnectionEvent(new ConnectEventArgs(ConnectionEventType.CheckRemoteControlStatusFail, this.connStatusType));
			LogtoFile.TrackLeave();
		}

		#endregion public method

		#region private method

		/// <summary>
		/// Checks the remote device system permission.
		/// First priority : read registry value to check it. (only for debug use)
		/// Second priority : By remote device info to check system permission
		/// </summary>
		/// <param name="device">The device.</param>
		/// <returns> RemoteDevicePermission value
		///
		/// RemoteDevicePermission.HavePermission
		/// RemoteDevicePermission.NoPermission
		///
		/// </returns>
		private RemoteDevicePermission CheckRemoteDeviceSystemPermission(AFUserDevice device)
		{
			LogtoFile.TrackEnter();

			try
			{
				//For easy change setting to try has/no system permission linking mehtod
				RemoteDevicePermission enPermissionCheck = CustomRegSetting.GetRemoteDevicePermission();

				if (enPermissionCheck != RemoteDevicePermission.AutoDetect)
				{
					return enPermissionCheck;
				}
				else
				{
					if (null != device)
					{
						if (device.HasSystemPermission.HasValue)
						{
							if (device.HasSystemPermission.Value)
							{
								return RemoteDevicePermission.HavePermission;
							}
							else
							{
								return RemoteDevicePermission.NoPermission;
							}
						}

						LogtoFile.WriteLog("", Configs.DebugLevelValue, "CheckRemoteDeviceSystemPermission device.HasSystemPermission has not value", "");
					}
					else
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "CheckRemoteDeviceSystemPermission device is null", "");
					}
				}

				LogtoFile.WriteLog("", Configs.DebugLevelValue, "CheckRemoteDeviceSystemPermission return default value NoPermission", "");
				return RemoteDevicePermission.NoPermission;
			}
			finally
			{
				LogtoFile.TrackLeave();
			}
		}

		/// <summary>
		/// Connections the remote input control.
		/// </summary>
		/// <param name="deviceInfo">The current connect device information. Because we need to Know
		/// Server IP, ServerVNCPort, VNCPassword, about these information is store in AppDeviceInfo, so
		/// we need to input the struct for reference value</param>
		/// <param name="controlmanager">The flow controlmanager. Becasue we need to check InputControl handler
		/// has been create, but the control is create by flow control, so we need to input the structure for
		/// reference value</param>
		/// <returns>
		/// Success, ConnectControlStatusType.InputControlConnectSuccess.
		/// Fail, ConnectControlStatusType.RunInputControlFailed and ConnectControlStatusType.InitialInputControlFail
		/// </returns>
		private ConnectControlStatusType ConnecRemoteInputControl(AppDeviceInfo deviceInfo, FlowControlManager controlmanager)
		{
			#region Connect to Remote device VNC

			LogtoFile.TrackEnter();

			try
			{
				//For Temp use, it shoudl be modify in after scheduler
				InitialStatus initializeResult = InitialStatus.InitializeInputControl;

				// try to connect input control
				for (int i = 0; i < Configs.RetryTimesForInputControl; i++)
				{
					FlowControlManager.Instance.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializeConnectTOVNC, i));

					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Try to connect to server for input control, time: " + i, "");

					if (null == controlmanager.InputController)
					{
						controlmanager.InputController = new VNCController(deviceInfo.ServerIP, deviceInfo.ServerVNCPort, deviceInfo.ServerVNCPassword, Configs.DownloadFileFolderPath);
					}

					initializeResult = controlmanager.InputController.InitializeVNC();

					//represent Initial Input control success.

					if (controlmanager.InputController == null)
						break;

					if (controlmanager.InputController.IsInitializationSuccess &&
						controlmanager.InputController.IsRunSuccess &&
						initializeResult == InitialStatus.ConnectToInputControlCompleted)
					{
						break;
					}

					Thread.Sleep(Configs.RetryTimePeriodForInputControl);
				}

				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Try to ConnectionRemoteInputControl result : " + initializeResult, "");

				if (initializeResult == InitialStatus.ConnectToInputControlCompleted)
				{
					return ConnectControlStatusType.InputControlConnectSuccess;
				}
				else if (initializeResult == InitialStatus.RunInputControlFailed)
				{
					return ConnectControlStatusType.InputControlConnectFail;
				}
				else
				{
					return ConnectControlStatusType.InitialInputControlFail;
				}
			}
			finally
			{
				LogtoFile.TrackLeave();
			}

			#endregion Connect to Remote device VNC
		}

		/// <summary>
		/// Disconnect Remote Input Control
		/// </summary>
		private void DisconnectionRemoteInputControl()
		{
			try
			{
				LogtoFile.TrackEnter();

				// Close VNC
				if (null != FlowControlManager.Instance.InputController)
				{
					if (null != FlowControlManager.Instance.InputController.VNCClient)
					{
						FlowControlManager.Instance.InputController.VNCClient.CloseVNC();
					}

					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Clear VNCController", "");
				}
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
			}
			finally
			{
				FlowControlManager.Instance.InputController = null;
				LogtoFile.TrackLeave();
			}
		}

		/// <summary>
		/// Checks the remote input daemon status by WebService Query result
		/// </summary>
		/// <param name="deviceInfo">The device information.</param>
		/// <returns>
		/// If return "ConnectControlStatusType.InputDaemonRunning" represent remote device input daemon is running
		/// If return "ConnectControlStatusType.InputDaemonNotRunning" represent remote device input daemon is not running
		/// If return "ConnectControlStatusType.CheckInputDaemonTimeout" represent timeout by webservice(maybe network has some issue)
		/// If return "ConnectControlStatusType.CheckInputDaemonWebException" represent WebService has some issue, need to double check.
		/// If return "ConnectControlStatusType.CheckInputDaemonException" represent the function has unexpect exception issue, need to double check.
		/// </returns>
		private async Task<ConnectControlStatusType> CheckRemoteInputDaemonStatus(AppDeviceInfo deviceInfo)
		{
			#region Check Input Daemon Status

			ConnectControlStatusType inputDaemonStatusCheck = ConnectControlStatusType.None;
			bool isInputDaemonAlreadyRunning = false;

			try
			{
				LogtoFile.TrackEnter();

				// Try to check input daemon status with the specific times
				for (int i = 0; i < Configs.CheckRemoteInputDaemonStatusRetryTimes; i++)
				{
					FlowControlManager.Instance.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializeCheckVNCDaemonStatus, i));

					try
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> Call GetInputDaemonStatusAsync()", "");
						isInputDaemonAlreadyRunning = await WebServiceGetInputDaemonStatusAsync(deviceInfo.Device).ConfigureAwait(false);

						inputDaemonStatusCheck = isInputDaemonAlreadyRunning ? ConnectControlStatusType.InputDaemonRunning : ConnectControlStatusType.InputDaemonNotRunning;
					}
					catch (TimeoutException)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> Call GetInputDaemonStatusAsync() Timeout", "");
						inputDaemonStatusCheck = ConnectControlStatusType.CheckInputDaemonTimeout;
					}
					catch (WebServiceException Wex)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> Call GetInputDaemonStatusAsync() WebServiceException" + Wex.ToString(), "");
						inputDaemonStatusCheck = ConnectControlStatusType.CheckInputDaemonWebException;
						break;
					}
					catch (Exception ex)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
						inputDaemonStatusCheck = ConnectControlStatusType.CheckInputDaemonException;
					}

					if (isInputDaemonAlreadyRunning)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "input daemon is already running", "");
						break;
					}
					else
					{
						if (this.remotedevicepermission == RemoteDevicePermission.NoPermission)
						{
							ConnectControlStatusType launchInputDaemonByAdbStatus = LaunchInputDaemonByADB(deviceInfo);

							if (launchInputDaemonByAdbStatus != ConnectControlStatusType.LaunchInputDaemonByADBSuccess)
							{
								inputDaemonStatusCheck = launchInputDaemonByAdbStatus;

								//LogtoFile.WriteLog("", Configs.DebugLevelValue, "input daemon is not running, wait 1s", "");
							}
						}

						LogtoFile.WriteLog("", Configs.DebugLevelValue, "input daemon is not running, wait 1s", "");
						//Thread.Sleep(Configs.CheckRemoteInputDaemonStatusRetryTimePeriod);
						await Task.Delay(Configs.CheckRemoteInputDaemonStatusRetryTimePeriod);
					}
				}

				if (isInputDaemonAlreadyRunning)
				{
					return ConnectControlStatusType.InputDaemonRunning;
				}
				else
				{
					if (this.remotedevicepermission == RemoteDevicePermission.NoPermission)
					{
						try
						{
							CleanProcess(this.ADBdaemonprocess);
						}
						catch (Exception ex)
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "_Adbinputdaemonprocess fail ex: " + ex.ToString(), "");
						}
					}

					return inputDaemonStatusCheck;
				}
			}
			finally
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> Call GetInputDaemonStatusAsync() result" + inputDaemonStatusCheck.GetStringValue(), "");

				LogtoFile.TrackLeave();
			}

			#endregion Check Input Daemon Status
		}

		/// <summary>
		/// Checks the remote screencast daemon status by WebService Query result
		/// </summary>
		/// <param name="deviceInfo">The device information.</param>
		/// <returns>
		/// If return "ConnectControlStatusType.ScreencastDaemonRunning" represent remote device screencast daemon is running
		/// If return "ConnectControlStatusType.CheckScreencastDaemonTimeout" represent timeout by webservice(maybe network has some issue)
		/// If return "ConnectControlStatusType.CheckScreencastDaemonWebException" represent WebService has some issue, need to double check.
		/// If return "ConnectControlStatusType.CheckScreencastDaemonException" represent the function has unexpect exception issue, need to double check.
		/// </returns>
		private async Task<ConnectControlStatusType> CheckRemoteScreencastDaemonStatus(AppDeviceInfo deviceInfo)
		{
			#region Check Input Daemon Status

			ConnectControlStatusType screencastDaemonResult = ConnectControlStatusType.None;
			DaemonStatus screencastDaemonStatus = DaemonStatus.NotRunning;

			try
			{
				LogtoFile.TrackEnter();

				// Try to check input daemon status with the specific times
				for (int i = 0; i < Configs.CheckRemoteScreencasetDaemonStatusRetryTimes; i++)
				{
					FlowControlManager.Instance.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializeCheckVNCDaemonStatus, i));

					try
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> Call GetInputDaemonStatusAsync()", "");
						screencastDaemonStatus = await this.WebServiceGetScreencastDaemonStatusAsync(deviceInfo.Device).ConfigureAwait(false);

						screencastDaemonResult = screencastDaemonStatus != DaemonStatus.NotRunning ? ConnectControlStatusType.ScreencastDaemonRunning : ConnectControlStatusType.ScreencastDaemonNotRunning;
					}
					catch (TimeoutException)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> Call GetInputDaemonStatusAsync() Timeout", "");
						screencastDaemonResult = ConnectControlStatusType.CheckScreencastDaemonTimeout;
					}
					catch (WebServiceException Wex)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> Call GetInputDaemonStatusAsync() WebServiceException" + Wex.ToString(), "");
						screencastDaemonResult = ConnectControlStatusType.CheckScreencastDaemonWebException;
						break;
					}
					catch (Exception ex)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
						screencastDaemonResult = ConnectControlStatusType.CheckScreencastDaemonException;
					}

					if (screencastDaemonStatus != DaemonStatus.NotRunning)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "screencast daemon is already running", "");
						break;
					}
					else
					{
						if (this.remotedevicepermission == RemoteDevicePermission.NoPermission)
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "The device No Permission, need to launch ADB to do it", "");

							ConnectControlStatusType launchScreencastDaemonByAdbStatus = LaunchScreencastDaemonByADB();

							if (launchScreencastDaemonByAdbStatus != ConnectControlStatusType.LaunchScreencastDaemonByADBSuccess)
							{
								screencastDaemonResult = launchScreencastDaemonByAdbStatus;

								//LogtoFile.WriteLog("", Configs.DebugLevelValue, "input daemon is not running, wait 1s", "");
							}
						}

						LogtoFile.WriteLog("", Configs.DebugLevelValue, "screencast daemon is not running, wait 1s", "");
						await Task.Delay(Configs.CheckRemoteScreencasetDaemonStatusRetryTimePeriod);
					}
				}

				if (screencastDaemonStatus != DaemonStatus.NotRunning)
				{
					return ConnectControlStatusType.ScreencastDaemonRunning;
				}
				else
				{
					if (this.remotedevicepermission == RemoteDevicePermission.NoPermission)
					{
						try
						{
							CleanProcess(this.ADBdaemonprocess);
						}
						catch (Exception ex)
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "CleanProcess fail ex: " + ex.ToString(), "");
						}
					}

					return screencastDaemonResult;
				}
			}
			finally
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> Call CheckRemoteScreencastDaemonStatus result" + screencastDaemonResult.GetStringValue(), "");

				LogtoFile.TrackLeave();
			}

			#endregion Check Input Daemon Status
		}

		private void InitialADBdaemon(ref Process process)
		{
			process = new Process();
			// The configurations of ADB command process
			process.StartInfo.CreateNoWindow = true;
			process.StartInfo.UseShellExecute = false;
			process.StartInfo.WindowStyle = ProcessWindowStyle.Hidden;
			process.StartInfo.RedirectStandardOutput = true;
			process.StartInfo.RedirectStandardError = true;
			process.StartInfo.FileName = Configs.ADBFilePath;
		}

		private void CleanProcess(Process _inputprocess)
		{
			try
			{
				LogtoFile.TrackEnter();

				if (_inputprocess != null)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "CleanProcess _inputprocess != null", "");

					try
					{
						_inputprocess.Kill();
					}
					catch (Exception ex)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "Kill _inputprocess fail ex: " + ex.ToString(), "");
					}

					_inputprocess = null;
				}
			}
			finally
			{
				LogtoFile.TrackLeave();
			}
		}

		private bool RunADBEvent(Process process, string argumnets)
		{
			try
			{
				LogtoFile.TrackEnter();

				process.StartInfo.Arguments = argumnets;
				process.Start();

				#region for Debug use

				string[] outputMessage;
				string[] errorMessage;
				outputMessage = process.StandardOutput.ReadToEnd().Split(new string[] { "\r\n" }, StringSplitOptions.RemoveEmptyEntries);
				errorMessage = process.StandardError.ReadToEnd().Split(new string[] { "\r\n" }, StringSplitOptions.RemoveEmptyEntries);

				LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent arguments: " + argumnets, "");

				if (outputMessage.Length != 0)
				{
					foreach (var output in outputMessage)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent outputMessage: " + output, "");
					}
				}

				if (errorMessage.Length != 0)
				{
					foreach (var error in errorMessage)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent errorMessage: " + error, "");
					}
				}

				#endregion for Debug use

				process.WaitForExit();

				if (process.ExitCode == 0)
				{
					return true;
				}
				else
				{
					return false;
				}
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent exception: " + ex.ToString(), "");

				return false;
			}
			finally
			{
				LogtoFile.TrackLeave();
			}
		}

		#region Input Daemon By ADB Command

		private ConnectControlStatusType LaunchInputDaemonByADB(AppDeviceInfo deviceInfo)
		{
			try
			{
				#region Initial Run ADB Input Daemon process parameter

				this.CleanProcess(this.ADBdaemonprocess);

				this.InitialADBdaemon(ref this._ADBdaemonprocess);

				#endregion Initial Run ADB Input Daemon process parameter

				#region ADB Kill Service

				if (this.ADBKillService(deviceInfo))
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent ADBKillService Success", "");
				}
				else
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent ADBKillService Fail", "");
				}

				#endregion ADB Kill Service

				#region ADB List Exist Device

				if (this.ADBListDevice())
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent ADBListDevice Success", "");
				}
				else
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent ADBListDevice Fail", "");
				}

				#endregion ADB List Exist Device

				#region Check Remote Device is Exist

				if (this.ADBCheckRemoteDeviceExist(deviceInfo))
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent Device Found Success", "");

					#region Pull Daemon from Remote Device

					if (this.ADBPullDaemonToRemoteDevice(deviceInfo))
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent Pull Input Daemon Success", "");

						#region Push Daemon to Remote Device

						if (this.ADBPushDaemonToRemoteDevice(deviceInfo))
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent Push Input Daemon Success", "");

							#region Shell Set Privilege daemon To Remote Device

							if (this.ADBShellSetPrivilegeDaemonToRemoteDevice(deviceInfo))
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent Shell Set Privilege Input Daemon Success", "");

								string argumnets = Configs.ShellInputDaemonCommand ?? String.Empty;
								this.ADBdaemonprocess.StartInfo.Arguments = argumnets;
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "StartInfo.Arguments: " + argumnets, "");
								this.ADBdaemonprocess.Start();

								return ConnectControlStatusType.LaunchInputDaemonByADBSuccess;
							}
							else
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent Shell Set Privilege Input Daemon Fail", "");
								return ConnectControlStatusType.ADBShellSetPrivilegeInputDaemonFail;
							}

							#endregion Shell Set Privilege daemon To Remote Device
						}
						else
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent Push Input Daemon Fail", "");
							return ConnectControlStatusType.ADBPushInputDaemonFail;
						}

						#endregion Push Daemon to Remote Device
					}
					else
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent Pull Input Daemon Fail", "");
						return ConnectControlStatusType.ADBPullInputDaemonFail;
					}

					#endregion Pull Daemon from Remote Device
				}
				else
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent Device Found Fail", "");
					return ConnectControlStatusType.ADBRemoteDeviceNotFound;
				}

				#endregion Check Remote Device is Exist
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "LaunchInputDaemonByADB Exception : " + ex.ToString(), "");

				return ConnectControlStatusType.LaunchInputDaemonByADBException;
			}
			finally
			{
				LogtoFile.TrackLeave();
			}
		}

		private bool ADBKillService(AppDeviceInfo deviceInfo)
		{
			string arguments = string.Format(Configs.KillServerCommand, deviceInfo.Device.SerialNumber);

			return this.RunADBEvent(this.ADBdaemonprocess, arguments);
		}

		private bool ADBListDevice()
		{
			string arguments = string.Format(Configs.DeviceCommand);

			return this.RunADBEvent(this.ADBdaemonprocess, arguments);
		}

		private bool ADBCheckRemoteDeviceExist(AppDeviceInfo deviceInfo)
		{
			string arguments = string.Format(Configs.DeviceExistCommand, deviceInfo.Device.SerialNumber);

			return this.RunADBEvent(ADBdaemonprocess, arguments);
		}

		private bool ADBPullDaemonToRemoteDevice(AppDeviceInfo deviceInfo)
		{
			string arguments = string.Format(Configs.PullInputDaemonCommand, deviceInfo.Device.SerialNumber);

			return this.RunADBEvent(ADBdaemonprocess, arguments);
		}

		private bool ADBPushDaemonToRemoteDevice(AppDeviceInfo deviceInfo)
		{
			string arguments = string.Format(Configs.PushInputDaemonCommand, deviceInfo.Device.SerialNumber);

			return this.RunADBEvent(this.ADBdaemonprocess, arguments);
		}

		private bool ADBShellSetPrivilegeDaemonToRemoteDevice(AppDeviceInfo deviceInfo)
		{
			string arguments = string.Format(Configs.ShellSetPrivilegeInputDaemonCommand, deviceInfo.Device.SerialNumber);

			return this.RunADBEvent(this.ADBdaemonprocess, arguments);
		}

		#endregion Input Daemon By ADB Command

		#region Screencast Daemon By ADB Command

		private ConnectControlStatusType LaunchScreencastDaemonByADB()
		{
			try
			{
				LogtoFile.TrackEnter();

				#region Initial Run ADB Input Daemon process parameter

				this.CleanProcess(this.ADBdaemonprocess);

				this.InitialADBdaemon(ref this._ADBdaemonprocess);

				#endregion Initial Run ADB Input Daemon process parameter

				#region Pull Screencast Daemon from Remote Device

				if (this.ADBPullScreencastDaemonToRemoteDevice())
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent Pull Screencast Daemon to Remote Device Success", "");

					#region Push Screencast Daemon to Remote Device

					if (this.ADBPushScreencastDaemonToRemoteDevice())
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent Push Screencast Daemon to Remote Device Success", "");

						#region Set Privilege Screencast Daemon To Remote Device

						if (this.ADBShellSetPrivilegeScreencastDaemonToRemoteDevice())
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent Shell Set Privilege Screencast Daemon Success", "");

							string arguments = Configs.ShellscreenCastingDaemonCommand ?? String.Empty;
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "StartInfo.Arguments: " + arguments, "");
							this.ADBdaemonprocess.StartInfo.Arguments = arguments;
							this.ADBdaemonprocess.Start();

							return ConnectControlStatusType.LaunchScreencastDaemonByADBSuccess;
						}
						else
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent Shell Set Privilege Screencast Daemon Fail", "");
							return ConnectControlStatusType.ADBShellSetPrivilegeScreencastDaemonFail;
						}

						#endregion Set Privilege Screencast Daemon To Remote Device
					}
					else
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent Push Screencast Daemon to Remote Device Fail", "");

						return ConnectControlStatusType.ADBPushScreencastDaemonFail;
					}

					#endregion Push Screencast Daemon to Remote Device
				}
				else
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "RunADBEvent Pull Screencast Daemon to Remote Device Fail", "");
					return ConnectControlStatusType.ADBPullScreencastDaemonFail;
				}

				#endregion Pull Screencast Daemon from Remote Device
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "LaunchScreencastDaemonByADB Exception : " + ex.ToString(), "");

				return ConnectControlStatusType.LaunchScreencastDaemonByADBException;
			}
			finally
			{
				LogtoFile.TrackLeave();
			}
		}

		private bool ADBPullScreencastDaemonToRemoteDevice()
		{
			string arguments = Configs.PullScreenCastingCommand ?? String.Empty;

			return this.RunADBEvent(this.ADBdaemonprocess, arguments);
		}

		private bool ADBPushScreencastDaemonToRemoteDevice()
		{
			string arguments = Configs.PushScreenCastingCommand ?? String.Empty;

			return this.RunADBEvent(this.ADBdaemonprocess, arguments);
		}

		private bool ADBShellSetPrivilegeScreencastDaemonToRemoteDevice()
		{
			string arguments = Configs.ShellSetPrivilegeScreenCastingDaemonCommand ?? String.Empty;

			return this.RunADBEvent(this.ADBdaemonprocess, arguments);
		}

		#endregion Screencast Daemon By ADB Command

		#region Event Handle

		private void SendConnectionEvent(ConnectEventArgs args)
		{
			try
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Ready to send ConnectionEvent: " + args.EventType, "");

				if (null != this.ConnectionFlowEvent)
				{
					var receivers = this.ConnectionFlowEvent.GetInvocationList();
					foreach (EventHandler<ConnectEventArgs> receiver in receivers)
					{
						receiver.BeginInvoke(this, args, null, null);
						//receiver.Invoke(this, args);
					}
				}
				else
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "this.ConnectionFlowEvent is null", "");
				}
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
			}
		}

		#endregion Event Handle

		#region Call WebService to check information

		/// <summary>
		/// Call WebService function to get remote Input Daemon Status
		/// </summary>
		/// <param name="targetDevice"></param>
		/// <returns>
		/// AFUserDevice.GetInputDaemonStatusAsync
		/// </returns>
		/// <exception cref="System.TimeoutException">TimeoutException</exception>
		/// <exception cref="System.WebServiceException">WebServiceException</exception>
		private async Task<bool> WebServiceGetInputDaemonStatusAsync(AFUserDevice targetDevice)
		{
			return await targetDevice.GetInputDaemonStatusAsync().ConfigureAwait(false);
		}

		/// <summary>
		/// Call WebService function to get remote Screencast Daemon Status
		/// </summary>
		/// <param name="targetDevice"></param>
		/// <returns>
		/// AFUserDevice.GetScreencastDaemonStatusAsync
		/// </returns>
		/// <exception cref="System.TimeoutException">TimeoutException</exception>
		/// <exception cref="System.WebServiceException">WebServiceException</exception>
		private async Task<DaemonStatus> WebServiceGetScreencastDaemonStatusAsync(AFUserDevice targetDevice)
		{
			return await targetDevice.GetScreencastDaemonStatusAsync().ConfigureAwait(false);
		}

		#endregion Call WebService to check information

		#endregion private method
	}
}