﻿using AirFusion.Common;
using AirFusion.Manager.DeviceManager;
using AirFusion.Manager.FeatureHelper;
using AirFusion.WebService.Adapters;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Threading;

namespace AirFusion.ConnectionManager
{
	internal class ScreencastManager : ConnectionInterface
	{
		#region define some const value

		private static ScreencastManager _pScreencastManager = null;
		private static object _pInstanceLock = new object();
		private ScreencastStatusType connStatusType;
		private bool IsDisconnect = false;

		#endregion define some const value

		#region public method

		public event EventHandler<ConnectEventArgs> ConnectionFlowEvent;

		public event EventHandler<AppEventArgs> ConnectStepEvent;

		/// <summary>
		/// For avoid ScreencastManager double create.
		/// </summary>
		public static ScreencastManager Instance
		{
			get
			{
				if (_pScreencastManager == null)
				{
					lock (_pInstanceLock)
					{
						if (_pScreencastManager == null) _pScreencastManager = new ScreencastManager();
					}
				}
				return _pScreencastManager;
			}
		}

		/// <summary>
		/// Connects the specified device information.
		/// </summary>
		/// <param name="deviceInfo">The device information.</param>
		/// <param name="controlmanager">The flow controlmanager.</param>
		public void Connect(AppDeviceInfo deviceInfo, FlowControlManager controlmanager)
		{
			this.IsDisconnect = false;

			//Need to check some parameter, default Protocol Type is TCP
			VideoProtocol protocol = VideoProtocol.TCP;

			//The fucntion will try to get protocol/Width/Height/BiteRate/KeyFrameInterval value from registry
			//If registry not assign any value, will based on device info to set default value
			//Protocol default is TCP
			//Witdh and Height default is device Support resoultion big value
			//BiteRate default is 0.5M
			//KeyFrameInterval is 15
			this.InitialVideoParameter(deviceInfo, ref protocol,
									   ref controlmanager.m_iVideoWidth, ref controlmanager.m_iVideoHeight,
									   ref controlmanager.m_BiteRate, ref controlmanager.m_iKeyFrameInterval);

			//Because DShow Filter need to know our Video Width and Height, so need to assign value.
			this.SettingDSVideoSize(controlmanager);

			if (this.IsDisconnect)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Before StartReceiving, because the user has been disconnect screencast, we should stop to connect quickly.", "");
				return;
			}

			//Add Direct Show Source filter
			if (controlmanager._currentScreenWindow.dsElement != null)
			{
				controlmanager._currentScreenWindow.dsElement.StartReceiving((ConnectionType)protocol);
			}

			//Initialize Screencast control info, it will insert previous check intial parameter, so need to check
			//double check previous initial parameter is right
			ScreencastControlInfo control = new ScreencastControlInfo()
			{
				DestinationPort = (ushort)deviceInfo.ServerRTPPort,
				Protocol = (ConnectionType)protocol == ConnectionType.UDP ? ScreencastControlInfo.ConnectionProtocol.Udp : ScreencastControlInfo.ConnectionProtocol.Tcp,
				Resolution = new ScreencastControlInfo.ResolutionInfo
				{
					Width = controlmanager.m_iVideoWidth,
					Height = controlmanager.m_iVideoHeight,
				},
				KeyFrameInterval = controlmanager.m_iKeyFrameInterval,
				Bitrate = controlmanager.m_BiteRate,
			};

			if (this.IsDisconnect)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Before DoPlayScreencast, because the user has been disconnect screencast, we should stop to connect quickly.", "");
				return;
			}

			//Control(play) remote device screencast by webservice
			DoPlayScreencast(deviceInfo, control).ContinueWith(delegate(Task<ScreencastStatusType> playScreencastTask)
			{
				ScreencastStatusType palyScreencast = playScreencastTask.Result;

				this.connStatusType = palyScreencast;

				//Check play remote device screencast result
				if (palyScreencast == ScreencastStatusType.ScreencastControlSuccess)
				{
					this.Connected();
				}
				else
				{
					this.Disconnect(deviceInfo);
				}
			}).ConfigureAwait(continueOnCapturedContext: false);
		}

		/// <summary>
		/// When Remote screencast control finish, need to create a call back funciton to info the result
		/// </summary>
		public void Connected()
		{
			this.SendConnectionEvent(new ConnectEventArgs(ConnectionEventType.ScreencastControlSuccess));
		}

		/// <summary>
		/// The Function will try to disconnecting about connction work.
		/// Need to Stop Screencast when disconnect
		/// </summary>
		public void Disconnect(AppDeviceInfo deviceInfo, bool IsAuto = true)
		{
			LogtoFile.TrackEnter();

			this.IsDisconnect = true;

			DoStopScreencast(deviceInfo, new ScreencastControlInfo()).ContinueWith(delegate(Task<ScreencastStatusType> stopScreencastTask)
			{
				LogtoFile.TrackLeave();
				this.Disconnected(IsAuto);
			}).ConfigureAwait(continueOnCapturedContext: false);
		}

		/// <summary>
		/// When Disconnected finished, If it is auto disconnect need to
		/// crate a call back function to info the result.
		/// </summary>
		/// <param name="deviceInfoStatus"></param>
		public void Disconnected(bool IsAuto)
		{
			LogtoFile.TrackEnter();

			if (IsAuto)
			{
				this.SendConnectionEvent(new ConnectEventArgs(ConnectionEventType.ScreencastControlFail, this.connStatusType));
			}

			LogtoFile.TrackLeave();
		}

		#endregion public method

		#region private method

		/// <summary>
		/// Initializes a new instance of the <see cref="ScreencastManager"/> class.
		/// </summary>
		private ScreencastManager()
		{
			this.connStatusType = ScreencastStatusType.None;
		}

		/// <summary>
		/// Initials the video parameter.
		/// </summary>
		/// <param name="deviceInfo">The device information.</param>
		/// <param name="protocol">The protocol.</param>
		/// <param name="videoWidth">Width of the video.</param>
		/// <param name="videoHeight">Height of the video.</param>
		/// <param name="videoBiteRate">The video bite rate.</param>
		/// <param name="videoKeyFrameInterval">The video key frame interval.</param>
		/// The funciton will try to get value from registry, If not found any value in registry key
		/// Will based on AppDeviceInfo to get/set default value
		/// Protocol default is TCP
		/// Witdh and Height default is device Support resoultion big value
		/// BiteRate default is 0.5M
		/// KeyFrameInterval is 15
		private void InitialVideoParameter(AppDeviceInfo deviceInfo, ref VideoProtocol protocol,
										   ref int videoWidth, ref int videoHeight,
										   ref uint videoBiteRate, ref uint videoKeyFrameInterval)
		{
			LogtoFile.TrackEnter();

			#region Get Video Protocol

			protocol = CheckVideoProtocol();

			#endregion Get Video Protocol

			#region Get Video Width/Height

			bool bVideoSize = GetVideoSize(deviceInfo, ref videoWidth, ref videoHeight);

			#endregion Get Video Width/Height

			#region Get Video BiteRate

			this.GetVideoBiteRate(ref videoBiteRate);

			#endregion Get Video BiteRate

			#region Get Video KeyFrame Interval

			this.GetVideoKeyFrameInterval(ref videoKeyFrameInterval);

			#endregion Get Video KeyFrame Interval

			#region for debug use, log some information

			LogtoFile.WriteLog("", Configs.DebugLevelValue, "Current Connection Protocol is : " + protocol.GetStringValue(), "");

			if (!bVideoSize)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "GetVideoSize fail reset to deafult 720p Width : " + videoWidth + " Height : " + videoHeight, "");
			}
			else
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "GetVideoSize Width : " + videoWidth + " Height : " + videoHeight, "");
			}

			LogtoFile.WriteLog("", Configs.DebugLevelValue, "GetVideoBiteRate value : " + videoBiteRate, "");

			LogtoFile.WriteLog("", Configs.DebugLevelValue, "GetVideoKeyFrameInterval value : " + videoKeyFrameInterval, "");

			#endregion for debug use, log some information

			LogtoFile.TrackLeave();
		}

		/// <summary>
		/// Checks the video protocol.
		/// </summary>
		/// <returns>VideoProtocol.TCP or VideoProtocol.UDP</returns>
		/// default vlaue is TCP, if registry has some special setting, will change to UDP
		private VideoProtocol CheckVideoProtocol()
		{
			LogtoFile.TrackEnter();
			VideoProtocol playVideoProtocol = VideoProtocol.TCP;
			try
			{
				playVideoProtocol = CustomRegSetting.GetVideoProtocol();
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "CheckVideoProtocol has exception : " + ex.ToString(), "");
			}
			finally
			{
				LogtoFile.TrackLeave();
			}
			return playVideoProtocol;
		}

		/// <summary>
		/// Gets the size of the video.
		/// </summary>
		/// <param name="deviceInfo">The device information.</param>
		/// <param name="videoWidth">Width of the video.</param>
		/// <param name="videoHeight">Height of the video.</param>
		/// <returns>Video Width/Height</returns>
		/// The funciton will try to get speical video width/height from registry
		/// If registry not set any value, it will based on AppDeviceInfo.Device.Services.Screencast.SupportedResolutions
		/// to set big resoultion
		private bool GetVideoSize(AppDeviceInfo deviceInfo, ref int videoWidth, ref int videoHeight)
		{
			LogtoFile.TrackEnter();
			try
			{
				CustomRegSetting.GetVideoSize(ref videoWidth, ref videoHeight);

				if (videoWidth == 0 || videoHeight == 0)
				{
					if (deviceInfo != null)
					{
						if (deviceInfo.Device.Services.Screencast.SupportedResolutions != null)
						{
							int _iDeviceSupportResolution = deviceInfo.Device.Services.Screencast.SupportedResolutions.Count;
							if (_iDeviceSupportResolution > 0)
							{
								videoWidth = deviceInfo.Device.Services.Screencast.SupportedResolutions[_iDeviceSupportResolution - 1].Width;
								videoHeight = deviceInfo.Device.Services.Screencast.SupportedResolutions[_iDeviceSupportResolution - 1].Height;
								return true;
							}
							else
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "GetVideoSize: deviceInfo.Device.Services.Screencast.SupportedResolutions is null", "");
							}
						}
						else
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "GetVideoSize: prepare set default width and height value", "");
						}
					}
				}
				else
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "GetVideoSize: video Width and Height has been assign value, don't need to change.", "");
					return true;
				}

				int _iHeight = deviceInfo.Device.Display.Height > deviceInfo.Device.Display.Width ? deviceInfo.Device.Display.Height.Value : deviceInfo.Device.Display.Width.Value;
				int _iWidth = deviceInfo.Device.Display.Width < deviceInfo.Device.Display.Height ? deviceInfo.Device.Display.Width.Value : deviceInfo.Device.Display.Height.Value;

				//Set default to 720p
				videoWidth = 720;
				videoHeight = (720 * _iHeight) / _iWidth;
				return false;
			}
			finally
			{
				LogtoFile.TrackLeave();
			}
		}

		/// <summary>
		/// Gets the video bite rate.
		/// </summary>
		/// <param name="videoBiteRate">The video bite rate.</param>
		/// It will get speical video biterate from registry. Default is 0.5M
		private void GetVideoBiteRate(ref uint videoBiteRate)
		{
			LogtoFile.TrackEnter();

			CustomRegSetting.GetVideoBiteRate(ref videoBiteRate);

			LogtoFile.TrackLeave();
		}

		/// <summary>
		/// Gets the video key frame interval.
		/// </summary>
		/// <param name="videoKeyFrameInterval">The video key frame interval.</param>
		/// It will get speical video key frame interval from registry. Default is 15
		private void GetVideoKeyFrameInterval(ref uint videoKeyFrameInterval)
		{
			LogtoFile.TrackEnter();
			CustomRegSetting.GetVideoKeyFrameInterval(ref videoKeyFrameInterval);
			LogtoFile.TrackLeave();
		}

		/// <summary>
		/// Settings the size of the ds video.
		/// </summary>
		/// <param name="controlmanager">The controlmanager.</param>
		/// Because DShow Filter need to know our Video Width and Height, so need to assign value.
		private void SettingDSVideoSize(FlowControlManager controlmanager)
		{
			controlmanager._currentScreenWindow.dsElement.VideoWidth = controlmanager.m_iVideoWidth;
			controlmanager._currentScreenWindow.dsElement.VideoHeight = controlmanager.m_iVideoHeight;
		}

		/// <summary>
		/// Play the remote screencast daemon status by WebService Query result
		/// </summary>
		/// <param name="deviceInfo">The device information.</param>
		/// <param name="control">The control is play screencast structure</param>
		/// <returns>
		/// If return "ScreencastStatusType.ScreencastControlSuccess" represent remote device screencast control is success
		/// If return "ScreencastStatusType.ScreencastControlTimeout" represent timeout by webservice(maybe network has some issue)
		/// If return "ScreencastStatusType.ScreencastControlWebServiceException" represent WebService has some issue, need to double check.
		/// If return "ScreencastStatusType.ScreencastControlException" represent the function has unexpect exception issue, need to double check.
		/// </returns>
		private async Task<ScreencastStatusType> DoPlayScreencast(AppDeviceInfo deviceInfo, ScreencastControlInfo control)
		{
			#region Play Screencast

			ScreencastStatusType playScreencastResult = ScreencastStatusType.None;

			try
			{
				LogtoFile.TrackEnter();

				// Try to check input daemon status with the specific times
				for (int i = 0; i < Configs.PlayRemoteSceencastRetryTimes; i++)
				{
					FlowControlManager.Instance.SendAppEventAsync(new AppEventArgs(AppEventType.InitialStatus, InitialStatus.InitializePlayScreencast, i));

					try
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> Call WebServiceControlScreenCastingDaemonAsync()", "");
						playScreencastResult = await this.WebServiceScreencastingControlAsync(deviceInfo.Device, ScreencastControlInfo.ScreencastOperation.Play, control).ConfigureAwait(false);
					}
					catch (TimeoutException)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> Call WebServiceControlScreenCastingDaemonAsync() Timeout", "");
						playScreencastResult = ScreencastStatusType.ScreencastControlTimeout;
					}
					catch (WebServiceException Wex)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> Call WebServiceControlScreenCastingDaemonAsync() WebServiceException" + Wex.ToString(), "");
						playScreencastResult = ScreencastStatusType.ScreencastControlWebServiceException;
						break;
					}
					catch (Exception ex)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
						playScreencastResult = ScreencastStatusType.ScreencastControlException;
					}

					if (playScreencastResult == ScreencastStatusType.ScreencastControlSuccess)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "screencast control is success", "");
						break;
					}
					else
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "screencast control is fail, will try again", "");

						await Task.Delay(Configs.PlayRemoteScreencastRetryTimePeriod);
					}
				}

				if (playScreencastResult != ScreencastStatusType.ScreencastControlSuccess)
				{
					return playScreencastResult;
				}
				else
				{
					return playScreencastResult;
				}
			}
			finally
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "DoPlayScreencast result" + playScreencastResult.GetStringValue(), "");

				LogtoFile.TrackLeave();
			}

			#endregion Play Screencast
		}

		/// <summary>
		/// Stop the remote screencast daemon status by WebService Query result
		/// </summary>
		/// <param name="deviceInfo">The device information.</param>
		/// <param name="control">The control is stop screencast structure</param>
		/// <returns>
		/// If return "ScreencastStatusType.ScreencastControlSuccess" represent remote device screencast control is success
		/// If return "ScreencastStatusType.ScreencastControlTimeout" represent timeout by webservice(maybe network has some issue)
		/// If return "ScreencastStatusType.ScreencastControlWebServiceException" represent WebService has some issue, need to double check.
		/// If return "ScreencastStatusType.ScreencastControlException" represent the function has unexpect exception issue, need to double check.
		/// </returns>
		private async Task<ScreencastStatusType> DoStopScreencast(AppDeviceInfo deviceInfo, ScreencastControlInfo control)
		{
			#region Stop Screencast

			ScreencastStatusType stopScreencastResult = ScreencastStatusType.None;

			try
			{
				LogtoFile.TrackEnter();

				try
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> Call WebServiceControlScreenCastingDaemonAsync()", "");
					stopScreencastResult = await this.WebServiceScreencastingControlAsync(deviceInfo.Device, ScreencastControlInfo.ScreencastOperation.Stop, control).ConfigureAwait(false);
				}
				catch (TimeoutException)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> Call WebServiceControlScreenCastingDaemonAsync() Timeout", "");
					stopScreencastResult = ScreencastStatusType.ScreencastControlTimeout;
				}
				catch (WebServiceException Wex)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> Call WebServiceControlScreenCastingDaemonAsync() WebServiceException" + Wex.ToString(), "");
					stopScreencastResult = ScreencastStatusType.ScreencastControlWebServiceException;
				}
				catch (Exception ex)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
					stopScreencastResult = ScreencastStatusType.ScreencastControlException;
				}

				if (stopScreencastResult != ScreencastStatusType.ScreencastControlSuccess)
				{
					return stopScreencastResult;
				}
				else
				{
					return stopScreencastResult;
				}
			}
			finally
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "DoStopScreencast result" + stopScreencastResult.GetStringValue(), "");

				LogtoFile.TrackLeave();
			}

			#endregion Stop Screencast
		}

		#region Event Handle

		private void SendConnectionEvent(ConnectEventArgs args)
		{
			try
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Ready to send ConnectionEvent: " + args.EventType, "");

				if (null != this.ConnectionFlowEvent)
				{
					var receivers = this.ConnectionFlowEvent.GetInvocationList();
					foreach (EventHandler<ConnectEventArgs> receiver in receivers)
					{
						receiver.BeginInvoke(this, args, null, null);
						//receiver.Invoke(this, args);
					}
				}
				else
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "this.ConnectionFlowEvent is null", "");
				}
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
			}
		}

		#endregion Event Handle

		#region Call WebService to do something

		/// <summary>
		/// Call WebService to control remote screen casting daemon.
		/// </summary>
		/// <param name="targetDevice">The target device.</param>
		/// <param name="operation">The operation.</param>
		/// <param name="control">The control.</param>
		/// <returns>The control remote device screencast result.
		/// If return "ScreencastStatusType.ScreencastControlSuccess" represent we can control remote device screencast.
		/// If return "ScreencastStatusType.ScreencastControlTimeout" represent timeout by webservice(maybe network has some issue)
		/// If return "ScreencastStatusType.ScreencastControlWebServiceException" represent WebService has some issue, need to double check.
		/// If return "ScreencastStatusType.ScreencastControlException" represent the function has some issue, need to double check.
		/// </returns>
		private async Task<ScreencastStatusType> WebServiceScreencastingControlAsync(AFUserDevice targetDevice, ScreencastControlInfo.ScreencastOperation operation, ScreencastControlInfo control)
		{
			LogtoFile.TrackEnter();

			LogtoFile.WriteLog("", Configs.DebugLevelValue, "WebServiceControlScreenCastingDaemonAsync operation : " + operation.GetStringValue(), "");

			try
			{
				await targetDevice.ControlScreencastAsync(operation, control).ConfigureAwait(false);
				return ScreencastStatusType.ScreencastControlSuccess;
			}
			catch (TimeoutException ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "WebServiceControlScreenCastingDaemonAsync Timout exception : " + ex.ToString(), "");
				return ScreencastStatusType.ScreencastControlTimeout;
			}
			catch (WebServiceException ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "WebServiceControlScreenCastingDaemonAsync WebServiceException : " + ex.ToString(), "");
				return ScreencastStatusType.ScreencastControlWebServiceException;
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
				return ScreencastStatusType.ScreencastControlException;
			}
			finally
			{
				LogtoFile.TrackLeave();
			}
		}

		#endregion Call WebService to do something

		#endregion private method
	}
}