﻿using AirFusion.Manager.DeviceManager;
using DirectShowLib;
using System;
using System.Drawing;
using System.Runtime.InteropServices;
using WPFMediaKit.DirectShow.Controls;
using WPFMediaKit.DirectX;

namespace AirFusion
{
	/// <summary>
	/// The DSElement class provides the WPF UI element for rendering video
	/// </summary>
	public class DSElement : D3DRenderer
	{
		#region Definitions

		private const int MaximumScreenCastFrameInterval = 1000;

		#endregion Definitions

		#region Members

		// This user id is dedicated to assign value to allocator of render
		private static IntPtr _userID = new IntPtr(unchecked((int)0xDEADBEEF));

		// This guid is for pull source filter, it must be the same with the definition in filter
		private static Guid _pullSourceFilterCLSID = new Guid("{B313B152-A57B-4051-8418-F7AB3250AB98}");

		// This guid is for LAV splitter, it must be the same with the definition in filter
		private static Guid _lavSplitterCLSID = new Guid("{11A35D23-7338-4B82-BA5B-BF8358461905}");

		// This guid is for LAV video, it must be the same with the definition in filter
		private static Guid _lavVideoDecoderCLSID = new Guid("{E01757C6-116F-4BDE-ADDA-0A2C23B01E5C}");

		// This guid is for LAV audio, it must be the same with the definition in filter
		private static Guid _lavAudioDecoderCLSID = new Guid("{ED2D8F44-BACA-4C21-873A-D3EC66AB337F}");

		// This guid is for video render, it must be the same with the definition in filter
		private static Guid _videoRendererCLSID = new Guid("{B87BEB7B-8D29-423F-AE4D-6582C10175AC}");

		// This IBaseFilter is for source filter
		private IBaseFilter _pullSourceFilter = null;

		private IBaseFilter _pLAVSplitter = null;
		private IBaseFilter _pLAVVideoDecoder = null;
		private IBaseFilter _pLAVAudioDecoder = null;
		private IBaseFilter _pvideoRenderer = null;
		private IBaseFilter _pdefaultDirectSoundDevice = null;
		private IVMRSurfaceAllocatorNotify9 _pIVMRSurfaceAllocatornotify = null;
		private Vmr9Allocator _pVmr9allocator = null;

		private bool _isClosed = true;
		private object _closeLockObject = new object();

		private DateTime? startTime = null, endTime = null;

		#endregion Members

		#region Events

		/// <summary>
		/// Fires when network is unstable.
		/// </summary>
		public event Action NetworkUnstable;

		#endregion Events

		#region Properties

		// This boolean value is for determining if the graph is running
		private bool _isGraphRunning = false;

		public bool IsGraphRunning
		{
			get
			{
				return _isGraphRunning;
			}
		}

		// This object is for DirectShow graph
		private object _graphLockObj = new object();

		private IGraphBuilder _graph = null;

		private IGraphBuilder Graph
		{
			get
			{
				lock (_graphLockObj)
				{
					return _graph;
				}
			}

			set
			{
				lock (_graphLockObj)
				{
					_graph = value;
				}
			}
		}

		public int VideoWidth;
		public int VideoHeight;

		// The timer to detect video frame interval.
		private System.Timers.Timer _screenCastFrameIntervalTimer;

		#endregion Properties

		#region Constructors

		/// <summary>
		/// Constructor of the <see cref="DSElement"/> class.
		/// </summary>
		public DSElement()
			: base()
		{
			this.InitializeVideoFrameIntervalTimer();
		}

		#endregion Constructors

		#region Public Methods

		/// <summary>
		/// Initializes required filters before running DirectShow graph
		/// </summary>
		public void Initialize()
		{
			LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start", "");

			try
			{
				// Initialize global filters
				_pullSourceFilter = (IBaseFilter)Activator.CreateInstance(Type.GetTypeFromCLSID(_pullSourceFilterCLSID));
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Initialize global filters -- Done!", "");

				int hr = 0;

				// graph builder
				ICaptureGraphBuilder2 pBuilder = (ICaptureGraphBuilder2)new CaptureGraphBuilder2();
				Graph = new FilterGraph() as IGraphBuilder;
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start adding graph builder", "");
				hr = pBuilder.SetFiltergraph(Graph);
				this.CheckHRValue(hr);
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Add graph builder -- Done!", "");

				// add Pull Source Filter
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start adding source filter", "");
				hr = Graph.AddFilter(_pullSourceFilter, "Pull Source Filter");
				this.CheckHRValue(hr);
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Add source filter -- Done!", "");

				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Add local filters -- Done!", "");
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
			}

			LogtoFile.WriteLog("", Configs.DebugLevelValue, "End", "");
		}

		/// <summary>
		/// Starts receiving network package based on the specific protocol
		/// </summary>
		/// <param name="connectionType"></param>
		public void StartReceiving(ConnectionType connectionType)
		{
			LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start", "");

			this.Dispatcher.Invoke(delegate()
			{
				try
				{
					this._isClosed = false;

					// Initialize global filters
					if (_pullSourceFilter == null)
						_pullSourceFilter = (IBaseFilter)Activator.CreateInstance(Type.GetTypeFromCLSID(_pullSourceFilterCLSID));
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Initialize global filters -- Done!", "");

					int hr = 0;

					// graph builder
					if (Graph == null)
					{
						ICaptureGraphBuilder2 pBuilder = (ICaptureGraphBuilder2)new CaptureGraphBuilder2();
						Graph = new FilterGraph() as IGraphBuilder;
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start adding graph builder", "");
						hr = pBuilder.SetFiltergraph(Graph);
						this.CheckHRValue(hr);
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "Add graph builder -- Done!", "");
					}
					// add Pull Source Filter
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start adding source filter", "");

					hr = Graph.AddFilter(_pullSourceFilter, "Pull Source Filter");
					this.CheckHRValue(hr);

					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Add source filter -- Done!", "");

					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Add local filters -- Done!", "");

					//LogtoFile.WriteLog("", Configs.DebugLevelValue, "Get interface", "");
					ISourceFilter iSourceFilter = (ISourceFilter)_pullSourceFilter;

					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Call ReConnecting", "");
					iSourceFilter.ReConnecting((int)connectionType);

					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Call StartReceiving", "");
					iSourceFilter.StartReceiving();

					LogtoFile.WriteLog("", Configs.DebugLevelValue, "End interface", "");
				}
				catch (Exception ex)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
				}
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "End", "");
			});
		}

		/// <summary>
		/// Runs DirectShow graph
		/// </summary>
		/// <param name="isAudioEnabled">
		/// The boolean value is to do if enabling audio
		/// </param>
		public void Run(bool isAudioEnabled)
		{
			LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start", "");

			this.Dispatcher.Invoke(delegate()
			{
				try
				{
					Guid CLSID_VideoMixingRenderer9 = new Guid("{51B4ABF3-748F-4E3B-A276-C828330E926A}"); //quartz.dll

					_pLAVSplitter = (IBaseFilter)Activator.CreateInstance(Type.GetTypeFromCLSID(_lavSplitterCLSID));
					_pLAVVideoDecoder = (IBaseFilter)Activator.CreateInstance(Type.GetTypeFromCLSID(_lavVideoDecoderCLSID));
					_pLAVAudioDecoder = (IBaseFilter)Activator.CreateInstance(Type.GetTypeFromCLSID(_lavAudioDecoderCLSID));
					_pdefaultDirectSoundDevice = (IBaseFilter)new DSoundRender();
					_pvideoRenderer = CreateRenderer() as IBaseFilter;

					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Initialize local filters -- Done!", "");

					int hr = 0;

					// add LAV Splitter
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start adding splitter filter", "");
					hr = Graph.AddFilter(_pLAVSplitter, "LAV Splitter");
					this.CheckHRValue(hr);
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Add splitter filter -- Done!", "");

					// setting LAV Splitter Dimension
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start set splitter filter dimension", "");
					ILAVSplitterSettingFilter _iLAVSplitterSetting = _pLAVSplitter as ILAVSplitterSettingFilter;
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "this.VideoWidth: " + this.VideoWidth, "");
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "this.VideoHeight: " + this.VideoHeight, "");

					_iLAVSplitterSetting.InitialdimensionSetting(this.VideoWidth, this.VideoHeight);
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Set splitter dimension -- Done!", "");

					// add LAV Video Decoder
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start adding video codec filter", "");
					hr = Graph.AddFilter(_pLAVVideoDecoder, "LAV Video Decoder");
					this.CheckHRValue(hr);
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Add video codec filter -- Done!", "");

					if (isAudioEnabled)
					{
						try
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "Audio is enabled", "");
							// add LAV Audio Decoder
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start adding audio filter", "");
							hr = Graph.AddFilter(_pLAVAudioDecoder, "LAV Audio Decoder");
							this.CheckHRValue(hr);
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "Add audio filter", "");

							// add Default DirectSound Device
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start adding audio device filter", "");
							hr = Graph.AddFilter(_pdefaultDirectSoundDevice, "Default DirectSound Device");
							this.CheckHRValue(hr);
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "Add audio device filter", "");
						}
						catch (Exception ex)
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
						}
					}
					else
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "Audio is disable", "");
					}

					// add Video Renderer
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start adding video render filter", "");
					hr = Graph.AddFilter(_pvideoRenderer, "Video Renderer");
					this.CheckHRValue(hr);
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Add video render filter", "");

					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Add local filters -- Done!", "");

					// connect UDP Pull Source Filter and LAV Splitter
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start connect souce filter to splitter filter", "");
					hr = Graph.ConnectDirect(GetPin(_pullSourceFilter, "Output"), GetPin(_pLAVSplitter, "Input"), null);
					this.CheckHRValue(hr);
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Connect souce filter to splitter filter -- Done!", "");

					// connect LAV Splitter and LAV Video Decoder
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start connect splitter filter to video codec filter", "");
					hr = Graph.ConnectDirect(GetPin(_pLAVSplitter, "Video"), GetPin(_pLAVVideoDecoder, "Input"), null);
					this.CheckHRValue(hr);
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Connect splitter filter to video codec filter -- Done!", "");

					if (isAudioEnabled)
					{
						try
						{
							// connect LAV Splitter and LAV Audio Decoder
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start connect splitter filter to audio codec filter", "");
							hr = Graph.ConnectDirect(GetPin(_pLAVSplitter, "Audio"), GetPin(_pLAVAudioDecoder, "Input"), null);
							this.CheckHRValue(hr);
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "Connect splitter filter to audio codec filter -- Done!", "");

							// connect LAV Audio Decoder and Default DirectSound Device
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start connect audio codec filter to audio device filter", "");
							hr = Graph.ConnectDirect(GetPin(_pLAVAudioDecoder, "Output"), GetPin(_pdefaultDirectSoundDevice, "Audio Input pin (rendered)"), null);
							this.CheckHRValue(hr);
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "Connect audio codec filter to audio device -- Done!", "");
						}
						catch (Exception ex)
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
						}
					}

					// connect LAV Video Decoder and Video Renderer
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start connect video codec filter to video render filter", "");
					hr = Graph.ConnectDirect(GetPin(_pLAVVideoDecoder, "Output"), GetPin(_pvideoRenderer, "VMR Input0"), null);
					this.CheckHRValue(hr);
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Connect video codec filter to video render filter -- Done!", "");

					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start _mediaControl.Run()", "");

					// Set sync source to null for lag issue on full screen mode.
					IMediaFilter mediaFilter = Graph as IMediaFilter;
					if (null != mediaFilter)
					{
						mediaFilter.SetSyncSource(null);
					}

					IMediaControl mediaControl = Graph as IMediaControl;
					if (null != mediaControl)
					{
						hr = mediaControl.Run();
						_isGraphRunning = true;
						this.CheckHRValue(hr);
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "mediaControl.Run() -- Done!", "");

						startTime = DateTime.Now;
					}
					else
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "mediaControl is null", "");
					}
				}
				catch (Exception ex)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
				}
			});

			LogtoFile.WriteLog("", Configs.DebugLevelValue, "End", "");
		}

		/// <summary>
		/// Closes DirectShow graph
		/// </summary>
		public void Close()
		{
			LogtoFile.WriteLog("", Configs.DebugLevelValue, "DSElement Close Start", "");

			lock (_closeLockObject)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "this._isClosed: " + this._isClosed, "");

				if (!this._isClosed)
				{
					try
					{
						if (startTime != null)
						{
							endTime = DateTime.Now;
                            double diffTime = ((TimeSpan)(endTime - startTime)).TotalMinutes;
                            string usedTime = Math.Floor(diffTime).ToString();
                            LogtoFile.WriteLog("", Configs.DebugLevelValue, "Screencast used Time = " + usedTime + " mins", "");
                            GoogleAnalytics.GoogleAnalyticSend("Screencast Time", usedTime);
						}
						else
						{
							LogtoFile.WriteLog("", Configs.DebugLevelValue, "Screencast Start Time is null", "");
						}
						_isGraphRunning = false;

						if (this._screenCastFrameIntervalTimer.Enabled)
						{
							this._screenCastFrameIntervalTimer.Stop();
						}

						//0522 Johnon add
						Configs.ApplicationReference._isShowTCPDisconnection = false;

						this.Dispatcher.Invoke(delegate()
						{
							if (_pullSourceFilter != null)
							{
								ISourceFilter iSourceFilter = (ISourceFilter)_pullSourceFilter;
								iSourceFilter.SendDummyData();
								_pullSourceFilter.Stop();
							}
						});

						if (this._graph != null)
						{
							IMediaControl mediaControl = this._graph as IMediaControl;
							mediaControl.Stop();
						}

						this.Dispatcher.Invoke(delegate()
						{
							if (_pullSourceFilter != null)
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release PullSourceFilter Start", "");
								int hr = Graph.RemoveFilter(_pullSourceFilter);
								CheckHRValue(hr);
								Marshal.ReleaseComObject(_pullSourceFilter);
								_pullSourceFilter = null;
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release PullSourceFilter End", "");
							}

							if (_pLAVSplitter != null)
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release LAVSplitter Start", "");
								int hr = Graph.RemoveFilter(_pLAVSplitter);
								CheckHRValue(hr);
								//Marshal.ReleaseComObject(_pLAVSplitter);
								_pLAVSplitter = null;
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release LAVSplitter End", "");
							}

							if (_pLAVVideoDecoder != null)
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release LAVVideoDecoder Start", "");
								int hr = Graph.RemoveFilter(_pLAVVideoDecoder);
								CheckHRValue(hr);
								//Marshal.ReleaseComObject(_pLAVVideoDecoder);
								_pLAVVideoDecoder = null;
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release LAVVideoDecoder End", "");
							}

							if (_pvideoRenderer != null)
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release videoRenderer Start", "");
								int hr = Graph.RemoveFilter(_pvideoRenderer);
								CheckHRValue(hr);
								//Marshal.ReleaseComObject(_pvideoRenderer);
								_pvideoRenderer = null;
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release videoRenderer End", "");
							}

							if (_pLAVAudioDecoder != null)
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release LAVAudioDecoder Start", "");
								int hr = Graph.RemoveFilter(_pLAVAudioDecoder);
								CheckHRValue(hr);
								//Marshal.ReleaseComObject(_pLAVAudioDecoder);
								_pLAVAudioDecoder = null;
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release LAVAudioDecoder End", "");
							}

							if (_pdefaultDirectSoundDevice != null)
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release DefaultDirectSoundDevice Start", "");
								int hr = Graph.RemoveFilter(_pdefaultDirectSoundDevice);
								CheckHRValue(hr);
								//Marshal.ReleaseComObject(_pdefaultDirectSoundDevice);
								_pdefaultDirectSoundDevice = null;
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release DefaultDirectSoundDevice End", "");
							}

							if (_pIVMRSurfaceAllocatornotify != null)
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release IVMRSurfaceAllocatornotify Start", "");
								Marshal.ReleaseComObject(_pIVMRSurfaceAllocatornotify);
								_pIVMRSurfaceAllocatornotify = null;
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release IVMRSurfaceAllocatornotify End", "");
							}

							if (_pVmr9allocator != null)
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release Vmr9allocator Start", "");
								base.SetIsFrontBufferAvailableChangedEventStatus(false);
								_pVmr9allocator.NewAllocatorFrame -= new Action(this.NewAllocatorFrameCallback);
								_pVmr9allocator.NewAllocatorSurface -= new NewAllocatorSurfaceDelegate(this.NewAllocatorSurfaceCallback);
								_pVmr9allocator.Dispose();
								_pVmr9allocator = null;
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release Vmr9allocator End", "");
							}

							if (null != this._graph)
							{
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release this._graph Start", "");
								Marshal.ReleaseComObject(this._graph);
								this._graph = null;
								LogtoFile.WriteLog("", Configs.DebugLevelValue, "Release this._graph End", "");
							}
						});
					}
					catch (Exception ex)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
					}

					// To reset images
					this.Dispatcher.Invoke(delegate()
					{
						this.CleanImage();
					});

					this._isClosed = true;
				}
			}

			LogtoFile.WriteLog("", Configs.DebugLevelValue, "DSElement Close End", "");
		}

		/// <summary>
		/// Sets the image can be referred
		/// </summary>
		public void SetCanRenderred()
		{
			try
			{
				if (!this.IsCanRenderred)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "IsCanRenderred is false, set to true", "");
					this.CountOfFrame = 0;      // To reset CountOfFrame to be zero
					this.IsCanRenderred = true; // To set the image can be renderred
				}
				else
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "IsCanRenderred is true", "");
				}
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
			}
		}

		/// <summary>
		/// Cleans the image in render
		/// </summary>
		public void CleanImage()
		{
			LogtoFile.WriteLog("", Configs.DebugLevelValue, ">>>>>> CleanImage()", "");

			try
			{
				this.IsCanRenderred = false;
				this.CountOfFrame = 0;      // To reset CountOfFrame to be zero
				this.VideoImage.Source = null;
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
			}
		}

		/// <summary>
		/// Save the screen snapshot to local as PNG.
		/// </summary>
		/// <param name="outputFolder">
		/// The ouput folder of the snapshot.
		/// </param>
		/// <param name="displayRotation">
		/// The current display rotation of Air Fusion.
		/// </param>
		/// <returns>
		/// The absolute path of the created snapshot.
		/// </returns>
		public string SnapShot(string ouputFolder, AFUserDevice.DisplayInfo.DisplayRotation displayRotation)
		{
			switch (displayRotation)
			{
				case AFUserDevice.DisplayInfo.DisplayRotation.Rotation0:
					return base.SnapCurrentD3DImage(ouputFolder, RotateFlipType.RotateNoneFlipNone);

				case AFUserDevice.DisplayInfo.DisplayRotation.Rotation90:
					return base.SnapCurrentD3DImage(ouputFolder, RotateFlipType.Rotate90FlipNone);

				case AFUserDevice.DisplayInfo.DisplayRotation.Rotation180:
					return base.SnapCurrentD3DImage(ouputFolder, RotateFlipType.Rotate180FlipNone);

				case AFUserDevice.DisplayInfo.DisplayRotation.Rotation270:
					return base.SnapCurrentD3DImage(ouputFolder, RotateFlipType.Rotate270FlipNone);

				default:
					return base.SnapCurrentD3DImage(ouputFolder, RotateFlipType.RotateNoneFlipNone);
			}
		}

		/// <summary>
		/// Checks is the primary display adapter lost.
		/// </summary>
		/// <returns>True: Adapter lost; False: Adapter still alive</returns>
		public bool CheckIsPrimaryDisplayAdapterLost()
		{
			if (this._pVmr9allocator != null)
			{
				return this._pVmr9allocator.CheckIsPrimaryDisplayAdapterLost();
			}
			else
			{
				return false;
			}
		}

		/// <summary>
		/// Pauses directshow.
		/// </summary>
		public void Pause()
		{
			lock (_closeLockObject)
			{
				if (!this._isClosed)
				{
					_isGraphRunning = false;

					if (this._graph != null)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "DSElement paused!!!", "");

						IMediaControl mediaControl = this._graph as IMediaControl;
						mediaControl.Pause();
					}
				}
			}
		}

		#endregion Public Methods

		#region Private Methods

		/// <summary>
		/// Creates a render filter
		/// </summary>
		/// <returns>
		/// VideoMixingRenderer9 filter
		/// </returns>
		private VideoMixingRenderer9 CreateRenderer()
		{
			var result = new VideoMixingRenderer9();
			var cfg = result as IVMRFilterConfig9;
			cfg.SetNumberOfStreams(1);
			cfg.SetRenderingMode(VMR9Mode.Renderless);

			_pIVMRSurfaceAllocatornotify = result as IVMRSurfaceAllocatorNotify9;
			_pVmr9allocator = new Vmr9Allocator();
			_pIVMRSurfaceAllocatornotify.AdviseSurfaceAllocator(_userID, _pVmr9allocator);
			_pVmr9allocator.AdviseNotify(_pIVMRSurfaceAllocatornotify);
			_pVmr9allocator.NewAllocatorFrame += new Action(this.NewAllocatorFrameCallback);
			_pVmr9allocator.NewAllocatorSurface += new NewAllocatorSurfaceDelegate(this.NewAllocatorSurfaceCallback);
			base.SetIsFrontBufferAvailableChangedEventStatus(true);
			return result;
		}

		/// <summary>
		/// Gets the pin of filter
		/// </summary>
		/// <param name="filter">
		/// Target filter
		/// </param>
		/// <param name="pinname">
		/// Pin name
		/// </param>
		/// <returns>
		/// Target pin
		/// </returns>
		private IPin GetPin(IBaseFilter filter, string pinname)
		{
			LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start, Pin Name: " + pinname, "");

			try
			{
				IEnumPins epins;
				int hr = filter.EnumPins(out epins);

				if (hr < 0)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Can't enumerate pins, Pin Name: " + pinname, "");
					return null;
				}

				IntPtr fetched = Marshal.AllocCoTaskMem(4);
				IPin[] pins = new IPin[1];
				while (epins.Next(1, pins, fetched) == 0)
				{
					PinInfo pinfo;
					pins[0].QueryPinInfo(out pinfo);
					bool found = (pinfo.name == pinname);
					DsUtils.FreePinInfo(pinfo);
					if (found)
					{
						return pins[0];
					}
				}
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
			}
			finally
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "End, Pin Name: " + pinname, "");
			}

			LogtoFile.WriteLog("", Configs.DebugLevelValue, "Pin not found, Pin Name: " + pinname, "");
			return null;
		}

		/// <summary>
		/// Checks the DirectShow operation result
		/// </summary>
		/// <param name="hr">
		/// return integer
		/// </param>
		private void CheckHRValue(int hr)
		{
			LogtoFile.WriteLog("", Configs.DebugLevelValue, "hr: " + hr, "");

			if (hr < 0)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, DsError.GetErrorText(hr), "");
				DsError.ThrowExceptionForHR(hr);
			}
		}

		/// <summary>
		/// Intializes the video frame interval timer.
		/// </summary>
		private void InitializeVideoFrameIntervalTimer()
		{
			this._screenCastFrameIntervalTimer = new System.Timers.Timer();
			this._screenCastFrameIntervalTimer.Interval = MaximumScreenCastFrameInterval;
			this._screenCastFrameIntervalTimer.Elapsed += delegate(object sender, System.Timers.ElapsedEventArgs e)
			{
				Action NetworkUnstabledHandler = this.NetworkUnstable;
				if (NetworkUnstabledHandler != null)
				{
					NetworkUnstabledHandler();
				}
			};
		}

		#endregion Private Methods

		#region Callbacks

		private int _debugCheckCount = 0;

		private void NewAllocatorFrameCallback()
		{
			if (_debugCheckCount > 300)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Renderring frame", "");
				_debugCheckCount = 0;
			}

			++_debugCheckCount;

			if (this._isGraphRunning)
			{
				if (this._screenCastFrameIntervalTimer.Enabled)
				{
					this._screenCastFrameIntervalTimer.Stop();
				}
				this._screenCastFrameIntervalTimer.Start();
			}

			base.InvalidateVideoImage();
		}

		private void NewAllocatorSurfaceCallback(object sender, IntPtr pSurface)
		{
			LogtoFile.WriteLog("", Configs.DebugLevelValue, "Start", "");

			try
			{
				base.SetBackBuffer(pSurface);
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
			}
			finally
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "End", "");
			}
		}

		#endregion Callbacks
	}
}