﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Diagnostics;
using System.Windows;
using AirFusion.Manager;
using AirFusion.Manager.DeviceManager;

namespace AirFusion
{
	/// <summary>
	/// The class provides the control logics of RTSP connection
	/// </summary>
	public class RTSPController
	{
		#region Members

		private string _queryM1StringFormat = "RTSP/1.0 200 OK\r\ncseq: {0}\r\nDate: Wed, 30 Jul 2014 05:42:09 +0000 \r\nUser-Agent: stagefright/1.2 (Linux;Android 4.4.2)\r\nPublic: org.wfa.wfd1.0, GET_PARAMETER, SET_PARAMETER\r\n\r\n";
		private string _queryM2StringFormat = "OPTIONS * RTSP/1.0\r\ncseq: {0}\r\nRequire: org.wfa.wfd1.0\r\n\r\n";
		private string _queryM3StringFormat = "RTSP/1.0 200 OK\r\nCSeq: {0}\r\nContent-Type: text/parameters\r\nContent-Length: {1}\r\n\r\n";
		private string _queryM4StringFormat = "RTSP/1.0 200 OK\r\nCSeq: {0}\r\n\r\n";
		private string _queryM5StringFormat = "RTSP/1.0 200 OK\r\nCSeq: {0}\r\n\r\n";
		private string _queryM6StringFormat = "SETUP rtsp://{0}/wfd1.0/streamid=0 RTSP/1.0\r\nCSeq: {1}\r\nTransport: RTP/AVP/UDP;unicast;client_port={2}\r\n\r\n";
		private string _queryM7StringFormat = "PLAY rtsp://{0}/wfd1.0/streamid=0 RTSP/1.0\r\nCSeq: {1}\r\nSession: {2}\r\n\r\n";

		private string _executeTearDownFormat = "TEARDOWN rtsp://{0}/wfd1.0/streamid=0 RTSP/1.0\r\nCSeq: {1}\r\nSession: {2}\r\n\r\n";
		private string _executePauseFormat = "PAUSE rtsp://{0}/wfd1.0/streamid=0 RTSP/1.0\r\nCSeq: {1}\r\nSession: {2}\r\n\r\n";
		private string _executePlayFormat = "PLAY rtsp://{0}/wfd1.0/streamid=0 RTSP/1.0\r\nCSeq: {1}\r\nSession: {2}\r\n\r\n";

		// video resolution is 1280 x 720 with 30 FPS
		private string _videoFormatInfo_1280x720 = "wfd_video_formats: 00 00 01 01 00000020 00000000 00000000 00 0000 0000 00 none none\r\n" +
										  "wfd_audio_codecs: LPCM 00000003 02, AAC 00000001 02, AC3 00000000 02\r\n" +
										  "wfd_client_rtp_ports: RTP/AVP/UDP;unicast 24030 0 mode=play\r\n" +
										  "wfd_uibc_capability: none\r\n";


		// video resolution is 720 x 480 with 60 FPS
		private string _videoFormatInfo_720x480 = "wfd_video_formats: 00 00 01 01 00000002 00000000 00000000 00 0000 0000 00 none none\r\n" +
										  "wfd_audio_codecs: LPCM 00000003 02, AAC 00000001 02, AC3 00000000 02\r\n" +
										  "wfd_client_rtp_ports: RTP/AVP/UDP;unicast 24030 0 mode=play\r\n" +
										  "wfd_uibc_capability: none\r\n";



		// video resolution is 640 x 480 with 60 FPS
		private string _videoFormatInfo_640x480 = "wfd_video_formats: 00 00 01 01 00000001 00000000 00000000 00 0000 0000 00 none none\r\n" +
										  "wfd_audio_codecs: LPCM 00000003 02, AAC 00000001 02, AC3 00000000 02\r\n" +
										  "wfd_client_rtp_ports: RTP/AVP/UDP;unicast 24030 0 mode=play\r\n" +
										  "wfd_uibc_capability: none\r\n";

		private Socket _socket = null;
		private string _session = string.Empty;

		private int _nCseqi = 0;
		private int _nCseqj = 0;

		private Byte[] _bytesReceived = new Byte[4096];

		#endregion

		#region Events

		public event EventHandler<RTSPControllerEventArgs> RTSPEvent;

		#endregion

		#region Properties

		private string _serverIP = string.Empty;
		public string ServerIP
		{
			get
			{
				return _serverIP;
			}
		}

		private int _rtspPort = 0;
		public int RTSPPort
		{
			get
			{
				return _rtspPort;
			}
		}

		private int _rtpPort = 0;
		public int RTPPort
		{
			get
			{
				return _rtpPort;
			}
		}

		private bool _isRTSPSocketConnected = false;
		public bool IsConnected
		{
			get
			{
				bool isConntected = false;

				if (null != this._socket)
				{
					isConntected = this._socket.Connected;
				}

				return isConntected && _isRTSPSocketConnected;
			}
		}

		#endregion

		#region Private Methods

		private string FindSession(string recvstring)
		{
			string result = string.Empty;

			int indexStart = recvstring.IndexOf("Session");
			if (indexStart == -1)
			{
				return result;
			}

			int indexEnd = recvstring.IndexOf(";", indexStart);
			if (indexEnd == -1)
			{
				return result;
			}

			result = recvstring.Substring(indexStart + 9, (indexEnd - (indexStart + 9)));

			return result;
		}

		#endregion

		#region Public Methods

		public RTSPController(string serverIP, int rtspport, int rtpport)
		{
			this._serverIP = serverIP;
			this._rtspPort = rtspport;
			this._rtpPort = rtpport;
		}

		public void InitializeSocket()
		{
			if (this._serverIP == string.Empty || this._rtspPort == 0)
			{
				return;
			}

			if (null == this._socket)
			{
				this._socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
				this._socket.ReceiveTimeout = Configs.ReceiveTimeoutForRTSPSocket;
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Create new socket...", "");
			}
			else 
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Use existed socket...", "");
			}

			if (!this._socket.Connected)
			{
				try
				{

					// Get server network IP
					IPAddress serverIP = null;
					if (IPAddress.TryParse(this._serverIP, out serverIP))
					{
						this._socket.Connect(serverIP, this._rtspPort);
						_isRTSPSocketConnected = true;
					}
				}
				catch(Exception ex)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, ex.ToString(), "");
				}
			}
			else
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "this._socket.Connected is true", "");
				_isRTSPSocketConnected = true;
			}

			return;
		}

		public string Send(VideoResolutionOptions options)
		{
			LogtoFile.WriteLog("", Configs.DebugLevelValue, "Request Video Resolution: " + options.ToString(), "");

			// Initialize video resolution
			string videoFormatInfo = string.Empty;
			switch (options)
			{
				case VideoResolutionOptions.LandscapeResolution1280x720:
					{ 
						videoFormatInfo = _videoFormatInfo_1280x720; 
					}
					break;

				case VideoResolutionOptions.LandscapeResolution640x480:
					{
						videoFormatInfo = _videoFormatInfo_640x480;
					}
					break;

				case VideoResolutionOptions.LandscapeResolution720x480:
					{
						videoFormatInfo = _videoFormatInfo_720x480;
					}
					break;

				default:
					{
						videoFormatInfo = _videoFormatInfo_1280x720;
					}
					break;
			}


			StringWriter sw = new StringWriter();

			_nCseqi = 0;
			_nCseqj = 0;

			// Got M1 request
			int bytes = 0;
			string receivedData = string.Empty;
			string request = string.Empty;
			Byte[] bytesSent;



			try
			{
				bytes = this._socket.Receive(_bytesReceived, _bytesReceived.Length, 0);
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Got M1 request failure, Error Message: " + ex.ToString(), "");
				return string.Empty;
			}
			receivedData = receivedData + Encoding.ASCII.GetString(_bytesReceived, 0, bytes);

			LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "\nbytes: " + bytes + "\n", "");
			LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "M1 request: " + receivedData, "");

			// Send M1 response
			this._nCseqi = 0;
			this._nCseqj = 0;

			this._nCseqi += 1;
			request = string.Format(this._queryM1StringFormat, this._nCseqi);
			bytesSent = Encoding.ASCII.GetBytes(request);

			LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "*M1 response: " + request, "");

			if (_socket.Send(bytesSent, bytesSent.Length, 0) == bytesSent.Length)
			{

			}
			else
			{

				return string.Empty;
			}

			this._session = FindSession(receivedData);
			sw.WriteLine(receivedData);





			// Send M2 request
			this._nCseqj = 2;
			request = string.Format(this._queryM2StringFormat, this._nCseqj);
			bytesSent = Encoding.ASCII.GetBytes(request);

			LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "*M2 request: " + request, "");

			if (this._socket.Send(bytesSent, bytesSent.Length, 0) == bytesSent.Length)
			{
				bytes = 0;
				receivedData = string.Empty;

				// Got M2 response
				try
				{
					bytes = this._socket.Receive(_bytesReceived, _bytesReceived.Length, 0);
				}
				catch (Exception ex)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Got M2 response failure, Error Message: " + ex.ToString(), "");
					return string.Empty;
				}
				receivedData = receivedData + Encoding.ASCII.GetString(_bytesReceived, 0, bytes);

				LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "\nbytes: " + bytes + "\n", "");
				LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "M2 response: " + receivedData, "");

				this._session = FindSession(receivedData);
				sw.WriteLine(receivedData);
			}
			else
			{
				sw.WriteLine("Send M2 Request Fail");


				return string.Empty;
			}





			// Got M3 request
			bytes = 0;
			receivedData = string.Empty;

			try
			{
				bytes = this._socket.Receive(_bytesReceived, _bytesReceived.Length, 0);
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Got M3 request failure, Error Message: " + ex.ToString(), "");
				return string.Empty;
			}
			receivedData = receivedData + Encoding.ASCII.GetString(_bytesReceived, 0, bytes);

			LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "\nbytes: " + bytes + "\n", "");
			LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "M3 request: " + receivedData, "");

			// Send M3 response
			this._nCseqi += 1;

			request = string.Format(this._queryM3StringFormat, this._nCseqi, videoFormatInfo.Length);
			request += videoFormatInfo;
			bytesSent = Encoding.ASCII.GetBytes(request);

			LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "*M3 response: " + request, "");

			if (this._socket.Send(bytesSent, bytesSent.Length, 0) == bytesSent.Length)
			{

			}
			else
			{

				return string.Empty;
			}

			this._session = FindSession(receivedData);
			sw.WriteLine(receivedData);



			// Got M4 request
			bytes = 0;
			receivedData = string.Empty;

			try
			{
				bytes = this._socket.Receive(_bytesReceived, _bytesReceived.Length, 0);
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Got M4 request failure, Error Message: " + ex.ToString(), "");
				return string.Empty;
			}
			receivedData = receivedData + Encoding.ASCII.GetString(_bytesReceived, 0, bytes);

			LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "\nbytes: " + bytes + "\n", "");
			LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "M4 request: " + receivedData, "");

			// Send M4 response
			this._nCseqi += 1;
			request = string.Format(this._queryM4StringFormat, this._nCseqi);
			bytesSent = Encoding.ASCII.GetBytes(request);

			LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "*M4 response: " + request, "");

			if (this._socket.Send(bytesSent, bytesSent.Length, 0) == bytesSent.Length)
			{

			}
			else
			{
				return string.Empty;
			}

			this._session = FindSession(receivedData);
			sw.WriteLine(receivedData);





			// Got M5 request
			bytes = 0;
			receivedData = string.Empty;

			try
			{
				bytes = this._socket.Receive(_bytesReceived, _bytesReceived.Length, 0);
			}
			catch (Exception ex)
			{
				LogtoFile.WriteLog("", Configs.DebugLevelValue, "Got M5 request failure, Error Message: " + ex.ToString(), "");
				return string.Empty;
			}

			receivedData = receivedData + Encoding.ASCII.GetString(_bytesReceived, 0, bytes);

			LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "\nbytes: " + bytes + "\n", "");
			LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "M5 request: " + receivedData, "");

			// Send M5 response
			this._nCseqi += 1;
			request = string.Format(this._queryM5StringFormat, this._nCseqi);
			bytesSent = Encoding.ASCII.GetBytes(request);

			LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "*M5 response: " + request, "");

			if (this._socket.Send(bytesSent, bytesSent.Length, 0) == bytesSent.Length)
			{

			}
			else
			{
				return string.Empty;
			}

			this._session = FindSession(receivedData);
			sw.WriteLine(receivedData);




			// Send M6 request
			this._nCseqj += 1;
			request = string.Format(this._queryM6StringFormat, this._serverIP, this._nCseqj, this._rtpPort);
			bytesSent = Encoding.ASCII.GetBytes(request);

			LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "*M6 request: " + request, "");

			if (this._socket.Send(bytesSent, bytesSent.Length, 0) == bytesSent.Length)
			{
				bytes = 0;
				receivedData = string.Empty;

				// Got M6 response
				try
				{
					bytes = this._socket.Receive(_bytesReceived, _bytesReceived.Length, 0);
				}
				catch (Exception ex)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Got M6 response failure, Error Message: " + ex.ToString(), "");
					return string.Empty;
				}
				receivedData = receivedData + Encoding.ASCII.GetString(_bytesReceived, 0, bytes);

				LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "\nbytes: " + bytes + "\n", "");
				LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "M6 response: " + receivedData, "");

				this._session = FindSession(receivedData);
				sw.WriteLine(receivedData);
			}
			else
			{

				sw.WriteLine("Send M6 Request Fail");

				return string.Empty;
			}




			// Send M7 request
			this._nCseqj += 1;
			request = string.Format(_queryM7StringFormat, this._serverIP, this._nCseqj, this._session);
			bytesSent = Encoding.ASCII.GetBytes(request);

			LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "*M7 request: " + request, "");

			if (this._socket.Send(bytesSent, bytesSent.Length, 0) == bytesSent.Length)
			{
				bytes = 0;
				receivedData = string.Empty;

				// Got M7 response
				try
				{
					bytes = this._socket.Receive(_bytesReceived, _bytesReceived.Length, 0);
				}
				catch (Exception ex)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Got M7 response failure, Error Message: " + ex.ToString(), "");
					return string.Empty;
				}
				receivedData = receivedData + Encoding.ASCII.GetString(_bytesReceived, 0, bytes);

				LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "\nbytes: " + bytes + "\n", "");
				LogtoFile.WriteLog("", DebugLevel.OnlyDebugWrite, "M7 response: " + request, "");


				this._session = FindSession(receivedData);
				sw.WriteLine(receivedData);
			}
			else
			{
				sw.WriteLine("Send M7 Request Fail");

				return string.Empty;
			}

			return sw.ToString();
		}

		public string Pause()
		{
			// Send pause request
			this._nCseqj += 1;
			string request = string.Format(this._executePauseFormat, this._serverIP, this._nCseqj, this._session);
			Byte[] bytesSent = Encoding.ASCII.GetBytes(request);

			if (this._socket.Send(bytesSent, bytesSent.Length, 0) == bytesSent.Length)
			{
				int bytes = 0;
				string receivedData = string.Empty;

				// Got pause response
				try
				{
					bytes = this._socket.Receive(_bytesReceived, _bytesReceived.Length, 0);
				}
				catch (Exception ex)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Got pause response failure, Error Message: " + ex.ToString(), "");
					return string.Empty;
				}

				receivedData = receivedData + Encoding.ASCII.GetString(_bytesReceived, 0, bytes);

				this._session = FindSession(receivedData);
				return receivedData;
			}
			else
			{
				return "Send Pause Request Fail";
			}
		}

		public string Play()
		{
			// Send play request
			this._nCseqj += 1;
			string request = string.Format(this._executePlayFormat, this._serverIP, this._nCseqj, this._session);
			Byte[] bytesSent = Encoding.ASCII.GetBytes(request);

			if (this._socket.Send(bytesSent, bytesSent.Length, 0) == bytesSent.Length)
			{
				int bytes = 0;
				string receivedData = string.Empty;

				// Got play response
				try
				{
					bytes = this._socket.Receive(_bytesReceived, _bytesReceived.Length, 0);
				}
				catch (Exception ex)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Got play response failure, Error Message: " + ex.ToString(), "");
					return string.Empty;
				}

				receivedData = receivedData + Encoding.ASCII.GetString(_bytesReceived, 0, bytes);

				this._session = FindSession(receivedData);
				return receivedData;
			}
			else
			{
				return "Send Play Request Fail";
			}
		}

		public string TearDown()
		{
			if (null != this._socket)
			{
				// Send tear down request
				this._nCseqj += 1;
				string request = string.Format(this._executeTearDownFormat, this._serverIP, this._nCseqj, this._session);
				Byte[] bytesSent = Encoding.ASCII.GetBytes(request);

				if (this._socket.Send(bytesSent, bytesSent.Length, 0) == bytesSent.Length)
				{
					int bytes = 0;
					string receivedData = string.Empty;

					// Got tear down response
					try
					{
						bytes = this._socket.Receive(_bytesReceived, _bytesReceived.Length, 0);
					}
					catch (Exception ex)
					{
						LogtoFile.WriteLog("", Configs.DebugLevelValue, "Got tear down response failure, Error Message: " + ex.ToString(), "");
						return string.Empty;
					}


					receivedData = receivedData + Encoding.ASCII.GetString(_bytesReceived, 0, bytes);

					this._session = FindSession(receivedData);
					return receivedData;
				}
				else
				{
					return "Send Tear Down Request Fail";
				}
			}

			return "No Socket is Existed";
		}

		public void CloseSocket()
		{
			if (null != this._socket)
			{
				try
				{
					this._socket.Close();
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "this._socket.Close()", "");
				}
				catch (Exception ex)
				{
					LogtoFile.WriteLog("", Configs.DebugLevelValue, "Socket close failure, Error Message: " + ex.ToString(), "");
				}

				this._socket = null;
			}
		}

		#endregion
	}
}
