﻿using AirFusion.Common;
using NativeWifi;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Net.NetworkInformation;
using System.Net.Sockets;
using System.Text;

namespace AirFusion.Common.Network
{
	/// <summary>
	/// Represents the network helper.
	/// </summary>
	public static partial class NetworkHelper
	{
		/// <summary>
		/// Gets the collection of network information of all available Ethernet / Wireless80211 network interfaces.
		/// </summary>
		/// <returns></returns>
		public static ICollection<NetworkInfo> GetAvailableSupportedNetworkList()
		{
			try
			{
				StaticLogger.TrackEnter();

				ICollection<NetworkInfo> networkList = new List<NetworkInfo>();

				foreach (var net in GetAvailableSupportedInterfaces().Where(item => item.Supports(NetworkInterfaceComponent.IPv4) && item.OperationalStatus == OperationalStatus.Up))
				{
					NetworkInfo networkInfo;

					// only get Ethernet and Wireless
					switch (net.NetworkInterfaceType)
					{
						case NetworkInterfaceType.Ethernet:
						case NetworkInterfaceType.Ethernet3Megabit:
						case NetworkInterfaceType.FastEthernetFx:
						case NetworkInterfaceType.FastEthernetT:
						case NetworkInterfaceType.GigabitEthernet:
							networkInfo = new NetworkInfo(net);
							networkList.Add(networkInfo);
							break;

						case NetworkInterfaceType.Wireless80211:
							networkInfo = new WirelessNetworkInfo(net);
							bool isWlanInfoGet = GetWlanNetInfo((WirelessNetworkInfo)networkInfo);
							//if (isWlanInfoGet) // workaround: maybe it's soft AP, TODO: get profile
							//{
							networkList.Add(networkInfo);
							//}
							break;
					}
				}

				return networkList;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#region Private/Protected/Internal Stuff

		private static IEnumerable<NetworkInterface> _availableSupportedInterfaces;
		private static object _locker;

		static NetworkHelper()
		{
			_availableSupportedInterfaces = null;
			_locker = new Object();
			NetworkChange.NetworkAvailabilityChanged += (s, e) =>
				{
					_availableSupportedInterfaces = GetAvailableSupportedInterfaces();
				};
		}

		/// <summary>
		/// Gets the available supported network interfaces.
		/// </summary>
		/// <returns></returns>
		private static IEnumerable<NetworkInterface> GetAvailableSupportedInterfaces()
		{
			lock (_locker)
			{
				return NetworkInterface.GetAllNetworkInterfaces().Where(
					ni => (ni.OperationalStatus == OperationalStatus.Up) &&
					(
						ni.NetworkInterfaceType == NetworkInterfaceType.Wireless80211 ||
						ni.NetworkInterfaceType == NetworkInterfaceType.Ethernet ||
						ni.NetworkInterfaceType == NetworkInterfaceType.Ethernet3Megabit ||
						ni.NetworkInterfaceType == NetworkInterfaceType.FastEthernetFx ||
						ni.NetworkInterfaceType == NetworkInterfaceType.FastEthernetT
					));
			}
		}

		/// <summary>
		/// Generates the encryption type of the connected wireless AP.
		/// </summary>
		/// <param name="auth">The wireless LAN authentication algorithm.</param>
		/// <param name="cipher">The wireless LAN cipher algorithm.</param>
		/// <returns>The encryption string, possible values: WEP, WPA, WPA2, (null).</returns>
		private static WirelessEncryptionProtocolType GetEncryption(Wlan.Dot11AuthAlgorithm auth, Wlan.Dot11CipherAlgorithm cipher)
		{
			switch (auth)
			{
				case Wlan.Dot11AuthAlgorithm.WPA:
				case Wlan.Dot11AuthAlgorithm.WPA_PSK:
				case Wlan.Dot11AuthAlgorithm.WPA_None:
					return WirelessEncryptionProtocolType.WPA;

				case Wlan.Dot11AuthAlgorithm.RSNA:
				case Wlan.Dot11AuthAlgorithm.RSNA_PSK:
					return WirelessEncryptionProtocolType.WPA2;

				default:
				case Wlan.Dot11AuthAlgorithm.IEEE80211_Open:
				case Wlan.Dot11AuthAlgorithm.IEEE80211_SharedKey:
					if (cipher != Wlan.Dot11CipherAlgorithm.None)
					{
						return WirelessEncryptionProtocolType.WEP;
					}
					break;
			}

			return WirelessEncryptionProtocolType.None;
		}

		/// <summary>
		/// Get the wireless AP information.
		/// </summary>
		/// <param name="networkInfo">The network info.</param>
		/// <returns>Indicates if find the correct network interface</returns>
		private static bool GetWlanNetInfo(WirelessNetworkInfo networkInfo)
		{
			bool findWlanInterfaceInfo = false;

			try
			{
				StaticLogger.TrackEnter();

				//WlanClient wlanClient = new WlanClient();
				WlanClient wlanClient = WlanClient.Client;
				foreach (WlanClient.WlanInterface wlanInterface in wlanClient.Interfaces)
				{
					string wlanInterfaceId = wlanInterface.InterfaceGuid.ToString();

					Debug.Assert(networkInfo.Interface != null);
					if (networkInfo.Interface.Id.IndexOf(wlanInterfaceId, 0, StringComparison.InvariantCultureIgnoreCase) >= 0)
					{
						Wlan.WlanConnectionAttributes currentConnection = wlanInterface.CurrentConnection;
						Wlan.WlanAssociationAttributes currentAssocaition = currentConnection.wlanAssociationAttributes;
						Wlan.WlanSecurityAttributes currentSecurity = currentConnection.wlanSecurityAttributes;

						// get SSID
						byte[] ssidArray = currentAssocaition.dot11Ssid.SSID;
						bool hasNulls = false;
						for (int i = 0; i < currentAssocaition.dot11Ssid.SSIDLength; i++)
						{
							if (ssidArray[i] == 0)
							{
								hasNulls = true;
								break;
							}
						}

						if (!hasNulls)
						{
							networkInfo.Ssid = Encoding.UTF8.GetString(ssidArray, 0, (int)currentAssocaition.dot11Ssid.SSIDLength);
						}

						// get BSSID
						networkInfo.Bssid = currentAssocaition.dot11Bssid;

						// get profile
						if (!string.IsNullOrEmpty(currentConnection.profileName))
						{
							// need to set Wlan_Profile_Get_Plaintext_key property on WlanGetProfile() function
							string profileXml = wlanInterface.GetProfileXml(currentConnection.profileName);
							int index = profileXml.IndexOf("<keyMaterial>", 0, StringComparison.InvariantCultureIgnoreCase);
							if (index != -1) // find password
							{
								networkInfo.Password = profileXml.Substring(index + 13, profileXml.IndexOf("</keyMaterial>") - index - 13);
							}

							index = profileXml.IndexOf("<nonBroadcast>", 0, StringComparison.InvariantCultureIgnoreCase);
							if (index != -1) // find Hidden property
							{
								string nonBroadcastValue = profileXml.Substring(index + 14, profileXml.IndexOf("</nonBroadcast>") - index - 14);
								networkInfo.IsHidden = (nonBroadcastValue.IndexOf("true", StringComparison.InvariantCultureIgnoreCase) >= 0);
							}
							else
							{
								networkInfo.IsHidden = false;
							}
						}

						// get Encryption
						networkInfo.Encryption = GetEncryption(currentSecurity.dot11AuthAlgorithm, currentSecurity.dot11CipherAlgorithm);

						// remove password if it's encrypted
						switch (networkInfo.Encryption)
						{
							case WirelessEncryptionProtocolType.WPA:
							case WirelessEncryptionProtocolType.WPA2:
								if (networkInfo.Password.Length < 8 || networkInfo.Password.Length > 63)
								{
									networkInfo.Password = null;
								}
								break;

							case WirelessEncryptionProtocolType.WEP:
								if (networkInfo.Password.Length != 5 && networkInfo.Password.Length != 13 && networkInfo.Password.Length != 29)
								{
									networkInfo.Password = null;
								}
								break;
						}

						findWlanInterfaceInfo = true;
						break;
					}
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("Failed getting Wlan network information, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}

			return findWlanInterfaceInfo;
		}

		#endregion Private/Protected/Internal Stuff
	}
}