﻿using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Reflection;

namespace AirFusion.Common.RegistryHolders
{
	internal class RegistryHoldersData : IDisposable
	{
		#region Definitions

		private const string LNRegistryKey = "RegistryKey";

		#endregion Definitions

		#region Private/Protected/Internal Variables

		private object _instance;
		private RegistryHive? _regHive;
		private string _regKeyPath;

		#endregion Private/Protected/Internal Variables

		#region Public Properties

		/// <summary>
		/// Gets the registry.
		/// </summary>
		public RegistryKey RegistryTree { get; private set; }

		#endregion Public Properties

		#region Public Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="RegistryHoldersData"/> class.
		/// </summary>
		/// <param name="instance">The instance.</param>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.ArgumentException" />
		public RegistryHoldersData(object instance)
		{
			// check parameters
			if (instance == null)
			{
				throw new ArgumentNullException("instance");
			}
			if (!instance.GetType().IsClass)
			{
				throw new ArgumentException("The specified instance is not an instance of a class");
			}

			// check if the instance has any properties with RegistryStoreTypeAttribute
			if (GetStoredProperties(instance).Count == 0)
			{
				throw new Exception("The specified does not contain any property with 'RegistryStoreTypeAttribute' attribute.");
			}

			this.RegistryTree = null;
			this._instance = instance;
			this._regHive = null;
			this._regKeyPath = null;
		}

		/// <summary>
		/// Opens or create the specified registry.
		/// </summary>
		/// <param name="regHive">The registry hive.</param>
		/// <param name="regKeyPath">The registry key path.</param>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.IO.IOException" />
		/// <exception cref="System.ObjectDisposedException" />
		/// <exception cref="System.Security.SecurityException" />
		/// <exception cref="System.UnauthorizedAccessException" />
		public void OpenOrCreateRegistry(RegistryHive regHive, string regKeyPath)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (string.IsNullOrWhiteSpace(regKeyPath))
				{
					throw new ArgumentException("Please specify a valid registry key path.", "regKeyPath");
				}

				lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNRegistryKey)))
				{
					if (this.RegistryTree != null)
					{
						this.RegistryTree.Close();
					}

					this.RegistryTree = CommonUtils.OpenOrCreateRegistry(regHive, Environment.Is64BitOperatingSystem ? RegistryView.Registry64 : RegistryView.Registry32, regKeyPath, true, true);
					this._regHive = regHive;
					this._regKeyPath = regKeyPath;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sets the registry key.
		/// </summary>
		/// <param name="regKey">The registry key.</param>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.ObjectDisposedException" />
		public void SetRegistry(RegistryKey regKey)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (regKey == null)
				{
					throw new ArgumentNullException("regKey");
				}

				lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNRegistryKey)))
				{
					if (this.RegistryTree != null)
					{
						this.RegistryTree.Close();
					}

					this.RegistryTree = regKey;

					// parse registry hive and registry key path
					var tokens = regKey.Name.Split('\\');
					if (tokens.Count() >= 1)
					{
						switch (tokens[0].ToUpper())
						{
							case "HKEY_LOCAL_MACHINE":
								this._regHive = RegistryHive.LocalMachine;
								break;

							case "HKEY_CURRENT_CONFIG":
								this._regHive = RegistryHive.CurrentConfig;
								break;

							case "HKEY_CLASSES_ROOT":
								this._regHive = RegistryHive.ClassesRoot;
								break;

							case "HKEY_CURRENT_USER":
								this._regHive = RegistryHive.CurrentUser;
								break;

							case "HKEY_USERS":
								this._regHive = RegistryHive.Users;
								break;

							case "HKEY_PERFORMANCE_DATA":
								this._regHive = RegistryHive.PerformanceData;
								break;

							case "HKEY_DYN_DATA":
								this._regHive = RegistryHive.DynData;
								break;

							default:
								throw new ArgumentException("The specified registry key is invalid with root key '" + tokens[0] + "'.");
						}
						this._regKeyPath = string.Join("", tokens.Skip(1));
					}
					else
					{
						throw new ArgumentException("The specified registry key is invalid.");
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Loads the specified property from registry.
		/// </summary>
		/// <param name="propertyName">The name of the property.</param>
		/// <returns></returns>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		public object LoadProperty(string propertyName)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (string.IsNullOrWhiteSpace(propertyName))
				{
					throw new ArgumentException("Please specify a valid property name.");
				}

				// check prerequisite
				if (this.RegistryTree == null)
				{
					throw new InvalidOperationException("No registry key is specified, please call OpenOrCreateRegistry() or SetRegistry() first.");
				}

				Debug.Assert(this._instance != null);
				var property = this._instance.GetType().GetProperty(propertyName);
				if (property == null)
				{
					throw new ArgumentException("The specified property name '" + propertyName + "' does not found in the type '" + this._instance.GetType().Name + "'.", "propertyName");
				}

				// get store type
				StoreType storeType;
				if (TryGetStoreType(property, out storeType) == false)
				{
					throw new Exception("The property '" + property.Name + "' doesn't have '" + typeof(RegistryStoreTypeAttribute).Name + "' attribute or marked as non-stored.");
				}

				try
				{
					var value = property.GetValue(this._instance);

					// re-create or open the registry key if such registry is invalid
					if (!IsRegistryAvailable(this.RegistryTree))
					{
						Debug.Assert(this._regHive.HasValue);
						Debug.Assert(!string.IsNullOrWhiteSpace(this._regKeyPath));

						try
						{
							this.OpenOrCreateRegistry(this._regHive.Value, this._regKeyPath);
						}
						catch (Exception ex)
						{
							throw new Exception("Failed opening the registry key '" + this._regKeyPath + "' from '" + this._regHive.Value + "'.", ex);
						}
					}

					bool instanceCreated;
					LoadProperty(this.RegistryTree, storeType, property, ref value, out instanceCreated);

					// set value if such value is value-type
					if (property.PropertyType.IsValueType || property.PropertyType.Equals(typeof(string)) || property.PropertyType.IsArray || instanceCreated)
					{
						property.SetValue(this._instance, value);
					}

					return value;
				}
				catch (Exception ex)
				{
					throw new Exception("Failed loading property '" + property.Name + "' under registry key '" + this.RegistryTree.ToString() + "'.", ex);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Loads properties from registry.
		/// </summary>
		/// <param name="throwOnLoadFail">A value indicating whether to throw exceptions when failed loading.</param>
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.Exception" />
		public void LoadProperties(bool throwOnLoadFail = false)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this.RegistryTree == null)
				{
					throw new InvalidOperationException("No registry key is specified, please call OpenOrCreateRegistry() or SetRegistry() first.");
				}

				// re-create or open the registry key if such registry is invalid
				if (!IsRegistryAvailable(this.RegistryTree))
				{
					Debug.Assert(this._regHive.HasValue);
					Debug.Assert(!string.IsNullOrWhiteSpace(this._regKeyPath));

					try
					{
						this.OpenOrCreateRegistry(this._regHive.Value, this._regKeyPath);
					}
					catch (Exception ex)
					{
						throw new Exception("Failed opening the registry key '" + this._regKeyPath + "' from '" + this._regHive.Value + "'.", ex);
					}
				}

				Debug.Assert(this._instance != null);
				foreach (var propertyEntry in GetStoredProperties(this._instance))
				{
					var property = propertyEntry.Key;
					var storeType = propertyEntry.Value;
					try
					{
						var value = propertyEntry.Key.GetValue(this._instance);
						bool instanceCreated;
						LoadProperty(this.RegistryTree, storeType, property, ref value, out instanceCreated);

						// set value if such value is value-type
						if (property.PropertyType.IsValueType || property.PropertyType.Equals(typeof(string)) || property.PropertyType.IsArray || instanceCreated)
						{
							property.SetValue(this._instance, value);
						}
					}
					catch (Exception ex)
					{
						if (throwOnLoadFail)
						{
							throw new Exception("Failed loading property '" + property.Name + "' under registry key '" + this.RegistryTree.ToString() + "'.", ex);
						}
						StaticLogger.LogWarn("Failed loading property '" + property.Name + "' under registry key '" + this.RegistryTree.ToString() + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Stores the property into registry.
		/// </summary>
		/// <param name="propertyName">The name of the property.</param>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		public void StoreProperty(string propertyName)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (string.IsNullOrWhiteSpace(propertyName))
				{
					throw new ArgumentException("Please specify a valid property name.");
				}

				// check prerequisite
				if (this.RegistryTree == null)
				{
					throw new InvalidOperationException("No registry key is specified, please call OpenOrCreateRegistry() or SetRegistry() first.");
				}

				Debug.Assert(this._instance != null);
				var property = this._instance.GetType().GetProperty(propertyName);
				if (property == null)
				{
					throw new ArgumentException("The specified property name '" + propertyName + "' does not found in the type '" + this._instance.GetType().Name + "'.", "propertyName");
				}

				// get store type
				StoreType storeType;
				if (TryGetStoreType(property, out storeType) == false)
				{
					throw new Exception("The property '" + property.Name + "' doesn't have '" + typeof(RegistryStoreTypeAttribute).Name + "' attribute or marked as non-stored.");
				}

				// re-create or open the registry key if such registry is invalid
				if (!IsRegistryAvailable(this.RegistryTree))
				{
					Debug.Assert(this._regHive.HasValue);
					Debug.Assert(!string.IsNullOrWhiteSpace(this._regKeyPath));

					try
					{
						this.OpenOrCreateRegistry(this._regHive.Value, this._regKeyPath);
					}
					catch (Exception ex)
					{
						throw new Exception("Failed opening the registry key '" + this._regKeyPath + "' from '" + this._regHive.Value + "'.", ex);
					}
				}

				StoreProperty(this.RegistryTree, storeType, property, property.GetValue(this._instance));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Stores all desired properties the into registry.
		/// </summary>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.NotSupportedException" />
		public void StoreProperties()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this.RegistryTree == null)
				{
					throw new InvalidOperationException("No registry key is specified, please call OpenOrCreateRegistry() or SetRegistry() first.");
				}

				// re-create or open the registry key if such registry is invalid
				if (!IsRegistryAvailable(this.RegistryTree))
				{
					Debug.Assert(this._regHive.HasValue);
					Debug.Assert(!string.IsNullOrWhiteSpace(this._regKeyPath));

					try
					{
						this.OpenOrCreateRegistry(this._regHive.Value, this._regKeyPath);
					}
					catch (Exception ex)
					{
						throw new Exception("Failed opening the registry key '" + this._regKeyPath + "' from '" + this._regHive.Value + "'.", ex);
					}
				}

				Debug.Assert(this._instance != null);
				foreach (var propertyEntry in GetStoredProperties(this._instance))
				{
					var property = propertyEntry.Key;
					var storeType = propertyEntry.Value;
					StoreProperty(this.RegistryTree, storeType, property, property.GetValue(this._instance));
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Closes this instance.
		/// </summary>
		public void Close()
		{
			lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNRegistryKey)))
			{
				if (this.RegistryTree != null)
				{
					this.RegistryTree.Close();
					this.RegistryTree = null;
				}
			}
		}

		#endregion Public Methods

		#region Static Functions

		/// <summary>
		/// Determines whether the specified object is nullable.
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="obj">The object.</param>
		/// <returns></returns>
		private static bool IsNullable<T>(T obj)
		{
			if (obj == null)
			{
				return true; // obvious
			}

			return IsNullable(typeof(T));
		}

		/// <summary>
		/// Determines whether the specified type is nullable.
		/// </summary>
		/// <param name="type">The type.</param>
		/// <returns></returns>
		private static bool IsNullable(Type type)
		{
			// check parameters
			if (type == null)
			{
				throw new ArgumentNullException("type");
			}

			return !type.IsValueType || Nullable.GetUnderlyingType(type) != null;
		}

		/// <summary>
		/// Determines if the specified registry key is still available.
		/// </summary>
		/// <param name="regKey">The registry key.</param>
		/// <returns></returns>
		private static bool IsRegistryAvailable(RegistryKey regKey)
		{
			if (regKey == null)
			{
				return false;
			}

			try
			{
				var count = regKey.SubKeyCount;
				return true;
			}
			catch
			{
				return false;
			}
		}

		/// <summary>
		/// Loads the specified property from the specified registry key.
		/// </summary>
		/// <param name="regKey">The registry key.</param>
		/// <param name="storeType">The store type.</param>
		/// <param name="property">The property.</param>
		/// <param name="propertyValue">The referenced value.</param>
		/// <param name="instanceCreated">True if an instance has been created.</param>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.Exception" />
		private static void LoadProperty(RegistryKey regKey, StoreType storeType, PropertyInfo property, ref object propertyValue, out bool instanceCreated)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (regKey == null)
				{
					throw new ArgumentNullException("key");
				}
				if (property == null)
				{
					throw new ArgumentNullException("property");
				}
				if (property.PropertyType.IsArray && (!property.PropertyType.GetElementType().Equals(typeof(byte)) && !property.PropertyType.GetElementType().Equals(typeof(string))))
				{
					throw new NotSupportedException("Array type is not supported.");
				}
				if (storeType == StoreType.NoStore)
				{
					throw new ArgumentException("The store type is non-stored.");
				}

				// assign default value
				instanceCreated = false;

				// get store name attribute
				var storeNameAttributes = property.GetCustomAttributes(typeof(RegistryStoreNameAttribute), false).OfType<RegistryStoreNameAttribute>();
				var regName = storeNameAttributes != null && storeNameAttributes.Count() > 0 ? storeNameAttributes.First().Name : property.Name;

				// get values by store type
				if (storeType == StoreType.Value) // read as value
				{
					try
					{
						// get store converter attribute
						var storeConverterAttributes = property.GetCustomAttributes(typeof(RegistryStoreConverterTypeAttribute), false).OfType<RegistryStoreConverterTypeAttribute>();
						var converterType = storeConverterAttributes != null && storeConverterAttributes.Count() > 0 ? storeConverterAttributes.First().ConverterType : null;
						var converter = converterType != null ? Activator.CreateInstance(converterType) as TypeConverter : null;

						propertyValue = LoadValue(regKey, regName, property.PropertyType, converter);
					}
					catch (Exception ex)
					{
						throw new Exception("Failed loading values of '" + regName + "' from registry key '" + regKey.ToString() + "'.", ex);
					}
				}
				else if (storeType == StoreType.SubTree) // read as subkey
				{
					try
					{
						var subKey = regKey.OpenSubKey(regName, true);
						if (subKey != null)
						{
							if (propertyValue == null)
							{
								propertyValue = Activator.CreateInstance(property.PropertyType);
								instanceCreated = true;
							}
							RegistryHolders.Create(propertyValue, subKey);
							RegistryHolders.Load(propertyValue, true);
						}
					}
					catch (Exception ex)
					{
						throw new Exception("Failed loading sub key '" + regName + "' under '" + regKey.ToString() + "'.", ex);
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Loads the specified value from the specified registry key.
		/// </summary>
		/// <param name="regKey">The registry key.</param>
		/// <param name="name">The registry name.</param>
		/// <param name="targetType">The target type.</param>
		/// <param name="converter">The converter.</param>
		/// <returns></returns>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.NotSupportedException" />
		private static object LoadValue(RegistryKey regKey, string name, Type targetType, TypeConverter converter = null)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (regKey == null)
				{
					throw new ArgumentNullException("key");
				}
				if (string.IsNullOrWhiteSpace(name))
				{
					throw new ArgumentException("Please specify a valid registry name.");
				}
				if (targetType == null)
				{
					throw new ArgumentNullException("targetType");
				}

				// get value
				object value;
				try
				{
					value = regKey.GetValue(name);
					if (value == null)
					{
						if (!IsNullable(targetType))
						{
							throw new NotSupportedException("The registry value is null or not existed but the target type is not nullable.");
						}
						else
						{
							return null;
						}
					}
				}
				catch (Exception ex)
				{
					throw new Exception("Failed getting value of '" + name + "' from registry '" + regKey.ToString() + "'.", ex);
				}

				if (converter != null)
				{
					if (converter.CanConvertFrom(value.GetType()) == false)
					{
						throw new InvalidCastException("The type of '" + name + "' ('" + value.GetType().Name + "') is not supported to be converted by the converter '" + converter.GetType().Name + "'.");
					}
					if (converter.CanConvertTo(targetType) == false)
					{
						throw new InvalidCastException("Unable to convert to '" + targetType.Name + "' by the converter '" + converter.GetType().Name + "'.");
					}

					return converter.ConvertTo(converter.ConvertFrom(value), targetType);
				}
				else
				{
					var valueType = value.GetType();

					if (targetType.IsGenericType)
					{
						var nullableType = Nullable.GetUnderlyingType(targetType);
						targetType = nullableType ?? targetType;
					}

					if (valueType.Equals(targetType) == false)
					{
						if (targetType.GetInterfaces().Contains(typeof(IConvertible)))
						{
							try
							{
								return Convert.ChangeType(value, targetType);
							}
							catch (Exception ex)
							{
								throw new InvalidCastException("The type of '" + valueType.Name + "' does not match to type '" + targetType.Name + "'.", ex);
							}
						}

						throw new InvalidCastException("The type of '" + valueType.Name + "' does not match to type '" + targetType.Name + "'.");
					}

					return value;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Stores the specified property into the specified registry.
		/// </summary>
		/// <param name="regKey">The registry key.</param>
		/// <param name="property">The property.</param>
		/// <param name="propertyValue">The property value.</param>
		/// <param name="storeType">The store type.</param>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.NotSupportedException" />
		private void StoreProperty(RegistryKey regKey, StoreType storeType, PropertyInfo property, object propertyValue)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (regKey == null)
				{
					throw new ArgumentNullException("key");
				}
				if (property == null)
				{
					throw new ArgumentNullException("property");
				}
				if (storeType == StoreType.NoStore)
				{
					throw new ArgumentException("The store type is non-stored.");
				}

				// get store name attribute
				var storeNameAttributes = property.GetCustomAttributes(typeof(RegistryStoreNameAttribute), false).OfType<RegistryStoreNameAttribute>();
				var regName = storeNameAttributes != null && storeNameAttributes.Count() > 0 ? storeNameAttributes.First().Name : property.Name;

				// store values by store type
				if (storeType == StoreType.Value) // read as value
				{
					try
					{
						// get store converter attribute
						var storeConverterAttributes = property.GetCustomAttributes(typeof(RegistryStoreConverterTypeAttribute), false).OfType<RegistryStoreConverterTypeAttribute>();
						var converterType = storeConverterAttributes != null && storeConverterAttributes.Count() > 0 ? storeConverterAttributes.First().ConverterType : null;
						var converter = converterType != null ? Activator.CreateInstance(converterType) as TypeConverter : null;

						StoreValue(regKey, regName, propertyValue, converter);
					}
					catch (Exception ex)
					{
						throw new Exception("Failed storing values of '" + regName + "' from registry key '" + regKey.ToString() + "'.", ex);
					}
				}
				else if (storeType == StoreType.SubTree) // save as subkey
				{
					if (propertyValue != null)
					{
						try
						{
							// create another registry holders for the class instance
							RegistryHolders.Create(propertyValue, regKey.OpenSubKey(regName, true));
							RegistryHolders.Store(propertyValue);
						}
						catch (Exception ex)
						{
							throw new Exception("Failed storing sub key '" + regName + "' under '" + regKey.ToString() + "'.", ex);
						}
					}
					else
					{
						try
						{
							// delete the whole sub key
							regKey.DeleteSubKeyTree(regName, false);
						}
						catch (Exception ex)
						{
							throw new Exception("Failed deleting sub key '" + regName + "' under '" + regKey.ToString() + "'.", ex);
						}
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Stores the specified value into the specified registry key.
		/// </summary>
		/// <param name="regKey">The registry key.</param>
		/// <param name="name">The registry name.</param>
		/// <param name="value">The value.</param>
		/// <param name="converter">The converter.</param>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.NotSupportedException" />
		private static void StoreValue(RegistryKey regKey, string name, object value, TypeConverter converter = null)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (regKey == null)
				{
					throw new ArgumentNullException("key");
				}
				if (string.IsNullOrWhiteSpace(name))
				{
					throw new ArgumentException("Please specify a valid registry name.");
				}

				if (value == null)
				{
					try
					{
						regKey.DeleteValue(name, false);
					}
					catch (Exception ex)
					{
						throw new Exception("Failed deleting registry name '" + name + "' under '" + regKey.ToString() + "'.", ex);
					}
				}
				else
				{
					if (converter != null)
					{
						if (converter.CanConvertFrom(value.GetType()) == false)
						{
							throw new InvalidCastException("The type of value ('" + value.GetType().Name + "') is not supported to be converted by the converter '" + converter.GetType().Name + "'.");
						}

						value = converter.ConvertFrom(value);
					}

					try
					{
						regKey.SetValue(name, value);
					}
					catch (Exception ex)
					{
						throw new Exception("Failed storing value of '" + name + "' into registry '" + regKey.ToString() + "'.", ex);
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Tries gets the store type of the specified instance.
		/// </summary>
		/// <param name="property">The property.</param>
		/// <param name="storeType">The store type.</param>
		/// <returns>A value indicating whether the property contains a store type.</returns>
		private static bool TryGetStoreType(PropertyInfo property, out StoreType storeType)
		{
			storeType = StoreType.NoStore;

			try
			{
				if (property == null)
				{
					return false;
				}

				storeType = property.GetCustomAttribute<RegistryStoreTypeAttribute>().StoreType;
				return true;
			}
			catch
			{
				return false;
			}
		}

		/// <summary>
		/// Gets stored properties.
		/// </summary>
		/// <param name="instance">The instance.</param>
		/// <returns></returns>
		private static IDictionary<PropertyInfo, StoreType> GetStoredProperties(object instance)
		{
			// check parameters
			if (instance == null)
			{
				throw new ArgumentNullException("instance");
			}
			if (instance.GetType().IsClass == false)
			{
				throw new ArgumentException("The specified instance is not an instance of a class");
			}

			var dictionary = new Dictionary<PropertyInfo, StoreType>();
			foreach (var property in instance.GetType().GetProperties())
			{
				StoreType storeType;
				if (TryGetStoreType(property, out storeType) && storeType != StoreType.NoStore)
				{
					dictionary.Add(property, storeType);
				}
			}

			return dictionary;
		}

		#endregion Static Functions

		#region IDisposable

		internal bool _disposed = false;

		internal virtual void Dispose(bool disposing)
		{
			if (!this._disposed)
			{
				// If disposing equals true, dispose all managed and unmanaged resources.
				if (disposing)
				{
					lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNRegistryKey)))
					{
						if (this.RegistryTree != null)
						{
							this.RegistryTree.Dispose();
							this.RegistryTree = null;
						}
					}
				}

				//
				// cleanup unmanaged resources
				//

				this._disposed = true;
			}
		}

		/// <summary>
		/// Disposes the object.
		/// </summary>
		public void Dispose()
		{
			this.Dispose(true);

			// This object will be cleaned up by the Dispose method.
			// Therefore, you should call GC.SupressFinalize to take this object off the finalization queue
			// and prevent finalization code for this object from executing a second time.
			GC.SuppressFinalize(this);
		}

		#endregion IDisposable
	}
}