﻿using AirFusion.Common;
using Bonjour;
using System;
using System.Diagnostics;
using System.Net;

namespace AirFusion.Manager.DeviceManager
{
	public partial class DnsSdHandler
	{
		private void OnServiceRegistered(DNSSDService service, DNSSDFlags flags, string instanceName, string serviceType, string domain)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("service: " + (service != null ? service.GetHashCode().ToString() : "N/A"));
				StaticLogger.LogDebug("flags: " + flags);
				StaticLogger.LogDebug("instanceName: " + instanceName);
				StaticLogger.LogDebug("serviceType: " + serviceType);
				StaticLogger.LogDebug("domain: " + domain);

				// check prerequisite
				{
					if (this.IsServiceTypeDesired(serviceType) == false)
					{
						return;
					}

					if (string.Compare(domain.TrimEnd('.'), this.Domain, true) != 0)
					{
						StaticLogger.LogWarn("Undesired domain: " + domain);
						return;
					}
				}

				// add into the pool of registered instance names
				this.RegisteredInstanceNames.Add(instanceName);

				// find the registrar
				Debug.Assert(service != null);
				uint processorID;
				ProcessorData data;
				if (!this.TryFindProcessor(service, out processorID, out data) || !(data is RegistrarData))
				{
					StaticLogger.LogWarn("No such registrar session exist or invalid data format.");
					return;
				}

				RegistrarData registrarData = data as RegistrarData;
				registrarData.InstanceName = instanceName;

				// invoke callback
				if (registrarData.Callback != null)
				{
					this.InvokeCallback<RegistrarCallbackArgs>(
						new Action<RegistrarCallbackArgs>(registrarData.Callback),
						new RegistrarCallbackArgs()
						{
							Handler = this,
							ProcessorID = processorID,
							InstanceName = instanceName,
							Hostname = registrarData.Hostname,
							Port = registrarData.Port,
							TxtRecord = registrarData.TxtRecord,
							Exception = null,
							IsCompleted = false
						}
					);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		private void OnBrowserEvent(DNSSDService browser, DNSSDFlags flags, uint ifIndex, string instanceName, string serviceType, string domain, BrowserCallbackArgs.BrowserEventType eventType)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("browser: " + (browser != null ? browser.GetHashCode().ToString() : "N/A"));
				StaticLogger.LogDebug("flags: " + flags);
				StaticLogger.LogDebug("ifIndex: " + ifIndex);
				StaticLogger.LogDebug("instanceName: " + instanceName);
				StaticLogger.LogDebug("serviceType: " + serviceType);
				StaticLogger.LogDebug("domain: " + domain);

				// check prerequisite
				{
					if (this.IsServiceTypeDesired(serviceType) == false)
					{
						return;
					}

					if (string.Compare(domain.TrimEnd('.'), this.Domain, true) != 0)
					{
						StaticLogger.LogWarn("Undesired domain: " + domain);
						return;
					}
				}

				// hide self ?
				if (this.HideSelf && this.RegisteredInstanceNames.Contains(instanceName))
				{
					StaticLogger.LogInfo("Skip self device.");
					return;
				}

				// find the browser
				Debug.Assert(browser != null);
				uint processorID;
				ProcessorData data;
				if (!this.TryFindProcessor(browser, out processorID, out data) || !(data is BrowserData))
				{
					StaticLogger.LogWarn("No such browser session exist or invalid data format.");
					return;
				}

				BrowserData browserData = data as BrowserData;

				// invoke callback
				if (browserData.Callback != null)
				{
					this.InvokeCallback<BrowserCallbackArgs>(
						new Action<BrowserCallbackArgs>(browserData.Callback),
						new BrowserCallbackArgs()
						{
							Handler = this,
							ProcessorID = processorID,
							InterfaceIndex = ifIndex,
							InstanceName = instanceName,
							EventType = eventType,
							Exception = null,
							IsCompleted = false
						}
					);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		private void OnServiceResolved(DNSSDService service, DNSSDFlags flags, uint ifIndex, string fullName, string hostname, ushort port, TXTRecord record)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("service: " + (service != null ? service.GetHashCode().ToString() : "N/A"));
				StaticLogger.LogDebug("flags: " + flags);
				StaticLogger.LogDebug("ifIndex: " + ifIndex);
				StaticLogger.LogDebug("fullName: " + fullName);
				StaticLogger.LogDebug("hostname: " + hostname);
				StaticLogger.LogDebug("port: " + port);
				if (record != null)
				{
					StaticLogger.LogDebug("record:\r\n" + record.GetRepresentation());
				}

				string instanceName, serviceName, domain;
				ConnectionProtocol connProtocol;

				// check prerequisite
				{
					if (this.IsFullNameDesired(fullName, out instanceName, out serviceName, out connProtocol, out domain) == false)
					{
						return;
					}

					if (string.IsNullOrWhiteSpace(hostname))
					{
						StaticLogger.LogWarn("'hostname' is unspecified.");
						return;
					}
				}

				// find the resolver
				Debug.Assert(service != null);
				uint processorID;
				ProcessorData data;
				if (!this.TryFindProcessor(service, out processorID, out data) || !(data is ResolverData))
				{
					StaticLogger.LogWarn("No such resolver session exist or invalid data format.");
					return;
				}

				ResolverData resolverData = data as ResolverData;

				// invoke callback
				if (resolverData.Callback != null)
				{
					this.InvokeCallback<ResolverCallbackArgs>(
						new Action<ResolverCallbackArgs>(resolverData.Callback),
						new ResolverCallbackArgs()
						{
							Handler = this,
							ProcessorID = processorID,
							InstanceName = instanceName,
							InterfaceIndex = ifIndex,
							Hostname = hostname,
							Port = port,
							TxtRecord = record.Transform(),
							MoreComing = (flags & DNSSDFlags.kDNSSDFlagsMoreComing) == DNSSDFlags.kDNSSDFlagsMoreComing,
							Exception = null,
							IsCompleted = false
						}
					);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		private void OnAddressFound(DNSSDService service, DNSSDFlags flags, uint ifIndex, string hostname, DNSSDAddressFamily addressFamily, string address, uint ttl)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("service: " + (service != null ? service.GetHashCode().ToString() : "N/A"));
				StaticLogger.LogDebug("flags: " + flags);
				StaticLogger.LogDebug("ifIndex: " + ifIndex);
				StaticLogger.LogDebug("hostname: " + hostname);
				StaticLogger.LogDebug("addressFamily: " + addressFamily);
				StaticLogger.LogDebug("address: " + address);
				StaticLogger.LogDebug("ttl: " + ttl);

				// parse flags
				bool moreComing = (flags & DNSSDFlags.kDNSSDFlagsMoreComing) != 0;
				bool isRemoved = ((int)flags & 2) == 0; // 2 == add

				// find the address querier
				Debug.Assert(service != null);
				uint processorID;
				ProcessorData data;
				if (!this.TryFindProcessor(service, out processorID, out data) || !(data is AddrQuerierData))
				{
					StaticLogger.LogWarn("No such address querier session exist ('" + hostname + "', '" + addressFamily + "') or invalid data format.");
					service.Stop();
					return;
				}

				AddrQuerierData addrQuerierData = data as AddrQuerierData;

				// parse address into IPAddress object
				IPAddress ipAddress;
				if (IPAddress.TryParse(address, out ipAddress) == false)
				{
					StaticLogger.LogWarn("The address '" + address + "' is invalid, ignore the result.");
					return;
				}

				// invoke callback
				if (addrQuerierData.Callback != null)
				{
					this.InvokeCallback<AddressQuerierCallbackArgs>(
						new Action<AddressQuerierCallbackArgs>(addrQuerierData.Callback),
						new AddressQuerierCallbackArgs()
						{
							Handler = this,
							ProcessorID = processorID,
							InterfaceIndex = ifIndex,
							Address = ipAddress,
							Hostname = hostname,
							QueryMethod = addrQuerierData.QueryMethod,
							MoreComing = moreComing,
							IsRemoved = isRemoved,
							Exception = null,
							IsCompleted = false
						}
					);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		private void OnQueryRecordAnswered(DNSSDService service, DNSSDFlags flags, uint ifIndex, string fullName, DNSSDRRType rrtype, DNSSDRRClass rrclass, object rdata, uint ttl)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("service: " + (service != null ? service.GetHashCode().ToString() : "N/A"));
				StaticLogger.LogDebug("flags: " + flags);
				StaticLogger.LogDebug("ifIndex: " + ifIndex);
				StaticLogger.LogDebug("fullName: " + fullName);
				StaticLogger.LogDebug("rrtype: " + rrtype);
				StaticLogger.LogDebug("rrclass: " + rrclass);
				StaticLogger.LogDebug("rdata: " + (rdata != null && rdata is byte[] ? "\r\n" + BitConverter.ToString((byte[])rdata) : "N/A"));
				StaticLogger.LogDebug("ttl: " + ttl);

				// parse flags
				bool moreComing = (flags & DNSSDFlags.kDNSSDFlagsMoreComing) != 0;
				bool isRemoved = ((int)flags & 2) == 0; // 2 == add

				// check prerequisite
				string instanceName, serviceName, domain;
				ConnectionProtocol connProtocol;
				if (this.IsFullNameDesired(fullName, out instanceName, out serviceName, out connProtocol, out domain) == false)
				{
					return;
				}

				// find the record querier
				Debug.Assert(service != null);
				uint processorID;
				ProcessorData data;
				if (!this.TryFindProcessor(service, out processorID, out data) || !(data is RecordQuerierData))
				{
					StaticLogger.LogWarn("No such record querier session exist ('" + instanceName + "', '" + rrtype + "') or invalid data format.");
					service.Stop();
					return;
				}

				RecordQuerierData recordQuerierData = data as RecordQuerierData;

				// TODO:
				// convert all the rdata to IResourceRecordData
				IResourceRecordData rrData = null;
				switch (rrtype)
				{
					case DNSSDRRType.kDNSSDType_A:
						rrData = new ResourceRecordDataA((byte[])rdata);
						break;
				}

				// invoke callback
				if (recordQuerierData.Callback != null)
				{
					this.InvokeCallback<RecordQuerierCallbackArgs>(
						new Action<RecordQuerierCallbackArgs>(recordQuerierData.Callback),
						new RecordQuerierCallbackArgs()
						{
							Handler = this,
							ProcessorID = processorID,
							InstanceName = instanceName,
							InterfaceIndex = ifIndex,
							RRType = rrtype.Transform(),
							RRData = rrData,
							QueryMethod = recordQuerierData.QueryMethod,
							MoreComing = moreComing,
							IsRemoved = isRemoved,
							Exception = null,
							IsCompleted = false
						}
					);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		private void OnOperationFailed(DNSSDService service, DNSSDError error)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("service: " + (service != null ? service.GetHashCode().ToString() : "N/A"));
				StaticLogger.LogDebug("error: " + error);

				// find the processor
				Debug.Assert(service != null);
				uint processorID;
				ProcessorData data;
				if (!this.TryFindProcessor(service, out processorID, out data))
				{
					StaticLogger.LogWarn("No such processor session exist.");
					service.Stop();
					return;
				}

				if (data is RegistrarData)
				{
					RegistrarData registrarData = data as RegistrarData;
					if (registrarData.Callback != null)
					{
						this.OnOperationFailed<RegistrarCallbackArgs>(
							new Action<RegistrarCallbackArgs>(registrarData.Callback),
							new RegistrarCallbackArgs()
							{
								Handler = this,
								ProcessorID = processorID,
								InstanceName = registrarData.InstanceName,
								Hostname = registrarData.Hostname,
								Port = registrarData.Port,
								TxtRecord = registrarData.TxtRecord,
								InterfaceIndex = registrarData.InterfaceIndex
							}, error);
					}
				}
				else if (data is BrowserData)
				{
					BrowserData browserData = data as BrowserData;
					if (browserData.Callback != null)
					{
						this.OnOperationFailed<BrowserCallbackArgs>(
							new Action<BrowserCallbackArgs>(browserData.Callback),
							new BrowserCallbackArgs()
							{
								Handler = this,
								ProcessorID = processorID,
								InterfaceIndex = browserData.InterfaceIndex
							}, error);
					}
				}
				else if (data is ResolverData)
				{
					ResolverData resolverData = data as ResolverData;
					if (resolverData.Callback != null)
					{
						this.OnOperationFailed<ResolverCallbackArgs>(
							new Action<ResolverCallbackArgs>(resolverData.Callback),
							new ResolverCallbackArgs()
							{
								Handler = this,
								ProcessorID = processorID,
								InstanceName = resolverData.InstanceName,
								InterfaceIndex = resolverData.InterfaceIndex,
								Hostname = null,
								Port = null,
								TxtRecord = null,
								MoreComing = null
							}, error);
					}
				}
				else if (data is AddrQuerierData)
				{
					AddrQuerierData addrQuerierData = data as AddrQuerierData;
					if (addrQuerierData.Callback != null)
					{
						this.OnOperationFailed<AddressQuerierCallbackArgs>(
							new Action<AddressQuerierCallbackArgs>(addrQuerierData.Callback),
							new AddressQuerierCallbackArgs()
							{
								Handler = this,
								ProcessorID = processorID,
								Hostname = null,
								InterfaceIndex = addrQuerierData.InterfaceIndex,
								Address = null,
								QueryMethod = addrQuerierData.QueryMethod,
								MoreComing = null,
								IsRemoved = null
							}, error);
					}
				}
				else if (data is RecordQuerierData)
				{
					RecordQuerierData recordQuerierData = data as RecordQuerierData;
					if (recordQuerierData.Callback != null)
					{
						this.OnOperationFailed<RecordQuerierCallbackArgs>(
							new Action<RecordQuerierCallbackArgs>(recordQuerierData.Callback),
							new RecordQuerierCallbackArgs()
							{
								Handler = this,
								ProcessorID = processorID,
								InstanceName = recordQuerierData.InstanceName,
								InterfaceIndex = recordQuerierData.InterfaceIndex,
								RRType = recordQuerierData.RRType,
								RRData = null,
								QueryMethod = recordQuerierData.QueryMethod,
								MoreComing = null,
								IsRemoved = null
							}, error);
					}
				}
				else
				{
					StaticLogger.LogWarn("Unrecognized DNSService.");
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		private void OnOperationFailed<T>(Action<T> callback, T args, DNSSDError error)
			where T : BaseCallbackArgs
		{
			StaticLogger.LogTrace("Invoke callback function with error...");

			if (args != null)
			{
				(args as BaseCallbackArgs).Exception = new Exception("Error Code: " + error);
				(args as BaseCallbackArgs).Exception.Data.Add("error", error);

				this.InvokeCallback<T>(callback, args);
			}
		}

		private void InvokeCallback<T>(Action<T> callback, T args)
			where T : BaseCallbackArgs
		{
			StaticLogger.LogTrace("Invoke callback function...");

			if (callback != null)
			{
				try
				{
					callback.Invoke(args);
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Unhandled exception of callback function, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				}
				/*
				callback.BeginInvoke(args, new AsyncCallback((result) =>
				{
					try
					{
						callback.EndInvoke(result);
					}
					catch (Exception ex)
					{
						StaticLogger.LogWarn("Unhandled exception of callback function, exception:\r\n" + AFCommonUtils.GetExceptionMessages(ex));
					}
				}), null);
				*/
			}
		}
	}
}
