﻿using AirFusion.Common;
//using InTheHand.Net; // 2015-05-20 Seoker: remove Bluetooth related functions since it's defeatured
using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;

namespace AirFusion.Manager.DeviceManager
{
	/// <summary>
	/// Represents the configuration of Device Management.
	/// </summary>
	public sealed partial class DeviceConfiguration
	{
		// TODO: replace this with RegistryHolders

		#region Definitions & Constants

		private const string KeyLocalInstanceID = "InstanceID";
		private const string KeyLocalDandelionServerHost = "DandelionServerHost";
		//private const string KeyLocalBluetoothAddress = "BluetoothAddress"; // 2015-05-20 Seoker: remove Bluetooth related functions since it's defeatured

		private const string KeyLocalDandelionServerPort = "DandelionServerPort";
		private const string KeyLocalDandelionServerSecure = "DandelionServerSecure";
		private const string KeyPairedDeviceDeviceName = "DeviceName";
		private const string KeyPairedDeviceModelName = "ModelName";
		private const string KeyPairedDeviceManufacturer = "Manufacturer";
		private const string KeyPairedDeviceOS = "OS";
		private const string KeyPairedDeviceOSVersion = "OSVersion";
		private const string KeyPairedDeviceUserName = "UserName";
		private const string KeyPairedDeviceAddress = "Address";
		private const string KeyPairedDeviceSerialNumber = "SerialNumber";
		private const string KeyPairedDeviceWebServicePort = "WebServicePort";
		private const string KeyPairedDeviceBluetoothAddress = "BluetoothAddress";
		private const string KeyPairedDeviceSessionID = "SessionID";

		#endregion Definitions & Constants

		#region Private/Protected/Internal Variables

		private RegistryKey _regKey;
		private IDictionary<Guid, AFUserDevice> _pairedDevice;

		#endregion Private/Protected/Internal Variables

		#region Properties

		/// <summary>
		/// Gets or sets the instance ID of local.
		/// </summary>
		public Guid? LocalInstanceID
		{
			get
			{
				try
				{
					Debug.Assert(this._regKey != null);
					object value = this._regKey.GetValue(KeyLocalInstanceID, null);
					if (value != null)
					{
						switch (this._regKey.GetValueKind(KeyLocalInstanceID))
						{
							case RegistryValueKind.String:
								return new Guid(value as string);

							case RegistryValueKind.Binary:
								return new Guid(value as byte[]);
						}
					}
				}
				catch (Exception ex)
				{
					StaticLogger.LogWarn("Failed getting or parsing '" + KeyLocalInstanceID + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				}

				return null;
			}
			set
			{
				try
				{
					Debug.Assert(this._regKey != null);
					if (value != null)
					{
						this._regKey.SetValue(KeyLocalInstanceID, value.Value.GetStringFormat());
					}
					else
					{
						this._regKey.DeleteValue(KeyLocalInstanceID);
					}
				}
				catch (Exception ex)
				{
					StaticLogger.LogWarn("Failed writing or deleting '" + KeyLocalInstanceID + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				}
			}
		}

		// 2015-05-20 Seoker: remove Bluetooth related functions since it's defeatured
		/*
		/// <summary>
		/// Gets or sets the Bluetooth address of local.
		/// </summary>
		public BluetoothAddress LocalBluetoothAddress
		{
			get
			{
				return this.GetBluetoothAddress(this._regKey, KeyLocalBluetoothAddress);
			}
			set
			{
				try
				{
					Debug.Assert(this._regKey != null);
					if (value != null)
					{
						this._regKey.SetValue(KeyLocalBluetoothAddress, value.ToByteArray(), RegistryValueKind.Binary);
					}
					else
					{
						this._regKey.DeleteValue(KeyLocalBluetoothAddress);
					}
				}
				catch (Exception ex)
				{
					StaticLogger.LogWarn("Failed writing or deleting '" + KeyLocalBluetoothAddress + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				}
			}
		}
		*/

		/// <summary>
		/// Gets or sets the address of dandelion server.
		/// </summary>
		public string DandelionServerHost
		{
			get
			{
				return this._regKey.GetValue(KeyLocalDandelionServerHost) as string;
			}
			set
			{
				try
				{
					Debug.Assert(this._regKey != null);
					if (value != null)
					{
						this._regKey.SetValue(KeyLocalDandelionServerHost, value);
					}
					else
					{
						this._regKey.DeleteValue(KeyLocalDandelionServerHost);
					}
				}
				catch (Exception ex)
				{
					StaticLogger.LogWarn("Failed writing or deleting '" + KeyLocalDandelionServerHost + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				}
			}
		}

		/// <summary>
		/// Gets or sets the port of dandelion server.
		/// </summary>
		public ushort? DandelionServerPort
		{
			get
			{
				return this.GetUnsignedShort(this._regKey, KeyLocalDandelionServerPort);
			}
			set
			{
				try
				{
					Debug.Assert(this._regKey != null);
					if (value != null)
					{
						this._regKey.SetValue(KeyLocalDandelionServerPort, value.Value, RegistryValueKind.DWord);
					}
					else
					{
						this._regKey.DeleteValue(KeyLocalDandelionServerPort);
					}
				}
				catch (Exception ex)
				{
					StaticLogger.LogWarn("Failed writing or deleting '" + KeyLocalDandelionServerPort + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				}
			}
		}

		/// <summary>
		/// Gets or sets the secure protocol of dandelion server.
		/// </summary>
		public bool? DandelionServerSecure
		{
			get
			{
				return this.GetBoolean(this._regKey, KeyLocalDandelionServerSecure);
			}
			set
			{
				try
				{
					Debug.Assert(this._regKey != null);
					if (value != null)
					{
						this._regKey.SetValue(KeyLocalDandelionServerSecure, value.Value, RegistryValueKind.DWord);
					}
					else
					{
						this._regKey.DeleteValue(KeyLocalDandelionServerSecure);
					}
				}
				catch (Exception ex)
				{
					StaticLogger.LogWarn("Failed writing or deleting '" + KeyLocalDandelionServerSecure + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				}
			}
		}

		/// <summary>
		/// Gets the list of paired devices.
		/// </summary>
		public IList<AFUserDevice> PairedDeviceList
		{
			get
			{
				if (this._pairedDevice == null)
				{
					this.LoadPairedDeviceList();
				}

				return this._pairedDevice.Select(item => item.Value).ToList();
			}
		}

		#endregion Properties

		#region Public Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="DeviceConfiguration"/> class.
		/// </summary>
		/// <param name="hive">The registry hive.</param>
		/// <param name="keyPath">The key path.</param>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.ObjectDisposedException" />
		/// <exception cref="System.Security.SecurityException" />
		/// <exception cref="System.UnauthorizedAccessException" />
		internal DeviceConfiguration(RegistryHive hive, string keyPath)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("hive: " + hive.ToString());
				StaticLogger.LogDebug("keyPath: " + keyPath);

				// check parameters
				if (string.IsNullOrEmpty(keyPath))
				{
					throw new ArgumentException("The 'keyPath' is unspecified.");
				}

				// open or create the registry key
				this._regKey = CommonUtils.OpenOrCreateRegistry(hive, Environment.Is64BitOperatingSystem ? RegistryView.Registry64 : RegistryView.Registry32, keyPath, true, true);

				this._pairedDevice = null;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Stores the device into the configuration.
		/// </summary>
		/// <param name="device">The device.</param>
		public void StoreDevice(AFUserDevice device)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("device" + (device != null ? ".InstanceID: " + device.InstanceID.GetStringFormat() : ": N/A"));

				// check parameters
				if (device == null)
				{
					throw new ArgumentNullException("device");
				}

				if (this._pairedDevice == null)
				{
					this.LoadPairedDeviceList();
				}

				// update the list of paired device
				if (this._pairedDevice.ContainsKey(device.InstanceID) == false)
				{
					this._pairedDevice.Add(device.InstanceID, device);
				}
				else
				{
					this._pairedDevice[device.InstanceID] = device;
				}

				// write into registry key
				Debug.Assert(this._regKey != null);
				RegistryKey key = CommonUtils.OpenOrCreateRegistry(this._regKey, device.InstanceID.GetStringFormat());

				// TODO:
				// save everything using RegistryHolders

				// SessionID
				try
				{
					if (device.SessionID != null)
					{
						key.SetValue(KeyPairedDeviceSessionID, device.SessionID.GetByteArray());
					}
					else
					{
						key.DeleteValue(KeyPairedDeviceSessionID);
					}
				}
				catch { }

				// DeviceName
				try
				{
					if (device.DeviceName != null)
					{
						key.SetValue(KeyPairedDeviceDeviceName, device.DeviceName);
					}
					else
					{
						key.DeleteValue(KeyPairedDeviceDeviceName);
					}
				}
				catch { }

				// ModelName
				try
				{
					if (device.ModelName != null)
					{
						key.SetValue(KeyPairedDeviceModelName, device.ModelName);
					}
					else
					{
						key.DeleteValue(KeyPairedDeviceModelName);
					}
				}
				catch { }

				// Manufacturer
				try
				{
					if (device.Manufacturer != null)
					{
						key.SetValue(KeyPairedDeviceManufacturer, device.Manufacturer);
					}
					else
					{
						key.DeleteValue(KeyPairedDeviceManufacturer);
					}
				}
				catch { }

				// OS
				try
				{
					if (device.OS != null)
					{
						key.SetValue(KeyPairedDeviceOS, device.OS);
					}
					else
					{
						key.DeleteValue(KeyPairedDeviceOS);
					}
				}
				catch { }

				// OSVersion
				try
				{
					if (device.OSVersion != null)
					{
						key.SetValue(KeyPairedDeviceOSVersion, device.OSVersion);
					}
					else
					{
						key.DeleteValue(KeyPairedDeviceOSVersion);
					}
				}
				catch { }

				// UserName
				try
				{
					if (device.UserName != null)
					{
						key.SetValue(KeyPairedDeviceUserName, device.UserName);
					}
					else
					{
						key.DeleteValue(KeyPairedDeviceUserName);
					}
				}
				catch { }

				// SerialNumber
				try
				{
					if (device.SerialNumber != null)
					{
						key.SetValue(KeyPairedDeviceSerialNumber, device.SerialNumber);
					}
					else
					{
						key.DeleteValue(KeyPairedDeviceSerialNumber);
					}
				}
				catch { }

				// Address
				try
				{
					if (device.Address != null)
					{
						key.SetValue(KeyPairedDeviceAddress, device.Address.GetAddressBytes());
					}
					else
					{
						key.DeleteValue(KeyPairedDeviceAddress);
					}
				}
				catch { }

				// WebServicePort
				try
				{
					key.SetValue(KeyPairedDeviceWebServicePort, device.WebServicePort, RegistryValueKind.DWord);
				}
				catch { }

				// 2015-05-20 Seoker: remove Bluetooth related functions since it's defeatured
				/*
				// Bluetooth
				try
				{
					if (device.Bluetooth != null && device.Bluetooth.Address != null)
					{
						key.SetValue(KeyPairedDeviceBluetoothAddress, device.Bluetooth.Address.ToByteArray(), RegistryValueKind.Binary);
					}
					else
					{
						key.DeleteValue(KeyPairedDeviceBluetoothAddress);
					}
				}
				catch { }
				*/

				key.Close();
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Removes the device from the configuration.
		/// </summary>
		/// <param name="device">The device.</param>
		public void RemoveDevice(AFUserDevice device)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("device" + (device != null ? ".InstanceID: " + device.InstanceID.GetStringFormat() : ": N/A"));

				// check parameters
				if (device == null)
				{
					throw new ArgumentNullException("device");
				}

				if (this._pairedDevice == null)
				{
					this.LoadPairedDeviceList();
				}

				// remove from the list of paired device
				this._pairedDevice.Remove(device.InstanceID);

				// delete the entire tree of the registry key
				var regKeyName = device.InstanceID.GetStringFormat();
				Debug.Assert(this._regKey != null);
				if (this._regKey.GetSubKeyNames().Contains(regKeyName, StringComparer.InvariantCultureIgnoreCase))
				{
					this._regKey.DeleteSubKeyTree(regKeyName);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Gets the device with the specified instance ID from configuration.
		/// </summary>
		/// <param name="instanceId">The instance ID.</param>
		/// <returns>The device of the specified instance Id; null if no such device found.</returns>
		public AFUserDevice GetDevice(Guid instanceId)
		{
			return this._pairedDevice.ContainsKey(instanceId) ? this._pairedDevice[instanceId] : null;
		}

		#endregion Public Methods

		#region Private/Protected/Internal Methods

		/// <summary>
		/// Loads the list of paired device.
		/// </summary>
		internal void LoadPairedDeviceList()
		{
			try
			{
				StaticLogger.TrackEnter();

				this._pairedDevice = new Dictionary<Guid, AFUserDevice>();

				Debug.Assert(this._regKey != null);
				foreach (var subKeyName in this._regKey.GetSubKeyNames())
				{
					try
					{
						Guid instanceId;
						if (Guid.TryParse(subKeyName, out instanceId))
						{
							this._pairedDevice.Add(instanceId, this.LoadDevice(subKeyName, instanceId));
						}
						else
						{
							StaticLogger.LogWarn("Invalid subkey '" + subKeyName + "', skipped.");
						}
					}
					catch (Exception ex)
					{
						StaticLogger.LogWarn("Failed loading '" + subKeyName + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
						continue;
					}
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogWarn("Failed loading paired device list, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				this._pairedDevice = null;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Loads the specified device.
		/// </summary>
		/// <param name="subKeyName">The sub-key.</param>
		/// <param name="instanceId">The instance ID.</param>
		/// <exception cref="System.Exception" />
		private AFUserDevice LoadDevice(string subKeyName, Guid instanceId)
		{
			// log parameters
			StaticLogger.LogDebug("subKeyName: " + (subKeyName ?? "N/A"));
			StaticLogger.LogDebug("instanceId: " + instanceId.GetStringFormat());

			Debug.Assert(!string.IsNullOrEmpty(subKeyName));
			using (RegistryKey key = this._regKey.OpenSubKey(subKeyName))
			{
				if (key == null)
				{
					throw new Exception("Failed openinig subkey '" + subKeyName + "'.");
				}

				AFUserDevice device = new AFUserDevice();

				// TODO:
				// load everything using RegistryHolders

				device.InstanceID = instanceId;
				device.DeviceName = key.GetValue(KeyPairedDeviceDeviceName, null) as string;
				device.ModelName = key.GetValue(KeyPairedDeviceModelName, null) as string;
				device.Manufacturer = key.GetValue(KeyPairedDeviceManufacturer, null) as string;
				device.OS = key.GetValue(KeyPairedDeviceOS, null) as string;
				device.OSVersion = key.GetValue(KeyPairedDeviceOSVersion, null) as string;
				device.UserName = key.GetValue(KeyPairedDeviceUserName, null) as string;
				device.SerialNumber = key.GetValue(KeyPairedDeviceSerialNumber, null) as string;
				device.Address = this.GetIPAddress(key, KeyPairedDeviceAddress);

				try
				{
					device.SessionID = PairingSessionId.CreateSessionId(key.GetValue(KeyPairedDeviceSessionID, null) as byte[]);
				}
				catch { }

				var port = this.GetUnsignedShort(key, KeyPairedDeviceWebServicePort);
				if (port.HasValue)
				{
					device.WebServicePort = port.Value;
				}

				// 2015-05-20 Seoker: remove Bluetooth related functions since it's defeatured
				/*
				device.Bluetooth = new AFUserDevice.BluetoothInfo()
				{
					Address = this.GetBluetoothAddress(key, KeyPairedDeviceBluetoothAddress)
				};
				*/

				StaticLogger.LogDebug("device.InstanceID: " + device.InstanceID.GetStringFormat());
				StaticLogger.LogDebug("device.SessionID: " + (device.SessionID != null ? device.SessionID.GetByteArray().ToHexString() : "N/A"));
				StaticLogger.LogDebug("device.DeviceName: " + device.DeviceName);
				StaticLogger.LogDebug("device.ModelName: " + device.ModelName);
				StaticLogger.LogDebug("device.Manufacturer: " + device.Manufacturer);
				StaticLogger.LogDebug("device.OS: " + device.OS);
				StaticLogger.LogDebug("device.OSVersion: " + device.OSVersion);
				StaticLogger.LogDebug("device.UserName: " + device.UserName);
				StaticLogger.LogDebug("device.SerialNumber: " + device.SerialNumber);
				StaticLogger.LogDebug("device.Address: " + (device.Address != null ? device.Address.ToString() : "N/A"));
				StaticLogger.LogDebug("device.WebServicePort: " + device.WebServicePort);
				//StaticLogger.LogDebug("device.Bluetooth:\r\n" + device.Bluetooth.ToString()); // 2015-05-20 Seoker: remove Bluetooth related functions since it's defeatured

				return device;
			}
		}

		/// <summary>
		/// Gets the IP address from the specified key/name.
		/// </summary>
		/// <param name="key">The key.</param>
		/// <param name="name">The name.</param>
		/// <returns></returns>
		private IPAddress GetIPAddress(RegistryKey key, string name)
		{
			try
			{
				Debug.Assert(key != null);
				Debug.Assert(!string.IsNullOrEmpty(name));
				object value = key.GetValue(name, null);
				if (value != null)
				{
					switch (key.GetValueKind(name))
					{
						case RegistryValueKind.String:
							return IPAddress.Parse(key.GetValue(name, null) as string);

						case RegistryValueKind.Binary:
							return new IPAddress(key.GetValue(name, null) as byte[]);

						case RegistryValueKind.QWord:
							return new IPAddress(Convert.ToInt64(key.GetValue(name, null)));
					}
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogWarn("Failed getting or parsing '" + name + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}

			return null;
		}

		// 2015-05-20 Seoker: remove Bluetooth related functions since it's defeatured
		/*
		/// <summary>
		/// Gets the Bluetooth address from the specified key/name.
		/// </summary>
		/// <param name="key">The key.</param>
		/// <param name="name">The name.</param>
		/// <returns></returns>
		private BluetoothAddress GetBluetoothAddress(RegistryKey key, string name)
		{
			try
			{
				Debug.Assert(key != null);
				Debug.Assert(!string.IsNullOrEmpty(name));
				object value = key.GetValue(name, null);
				if (value != null)
				{
					switch (key.GetValueKind(name))
					{
						case RegistryValueKind.String:
							return BluetoothAddress.Parse(key.GetValue(name, null) as string);

						case RegistryValueKind.Binary:
							return new BluetoothAddress((key.GetValue(name, null) as byte[]).Take(6).ToArray());
					}
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogWarn("Failed getting or parsing '" + name + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}

			return null;
		}
		*/

		/// <summary>
		/// Gets the unsigned short value from the specified key/name.
		/// </summary>
		/// <param name="key">The key.</param>
		/// <param name="name">The name.</param>
		/// <returns></returns>
		private ushort? GetUnsignedShort(RegistryKey key, string name)
		{
			try
			{
				Debug.Assert(key != null);
				Debug.Assert(!string.IsNullOrEmpty(name));
				object value = key.GetValue(name, null);
				if (value != null)
				{
					switch (key.GetValueKind(name))
					{
						case RegistryValueKind.String:
							return ushort.Parse(key.GetValue(name, null) as string);

						case RegistryValueKind.DWord:
							Int32 intValue = (Int32)key.GetValue(name, null);
							if (intValue <= ushort.MaxValue)
							{
								return (ushort)intValue;
							}
							break;
					}
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogWarn("Failed getting or parsing '" + name + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}

			return null;
		}

		/// <summary>
		/// Gets the unsigned short value from the specified key/name.
		/// </summary>
		/// <param name="key">The key.</param>
		/// <param name="name">The name.</param>
		/// <returns></returns>
		private bool? GetBoolean(RegistryKey key, string name)
		{
			try
			{
				Debug.Assert(key != null);
				Debug.Assert(!string.IsNullOrEmpty(name));
				object value = key.GetValue(name, null);
				if (value != null)
				{
					switch (key.GetValueKind(name))
					{
						case RegistryValueKind.Binary:
							return (key.GetValue(name, null) as byte[])[0] != 0;

						case RegistryValueKind.DWord:
							return (Int32)key.GetValue(name, null) != 0;

						case RegistryValueKind.QWord:
							return (Int64)key.GetValue(name, null) != 0;
					}
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogWarn("Failed getting or parsing '" + name + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}

			return null;
		}

		#endregion Private/Protected/Internal Methods
	}

	#region Disposable

	public sealed partial class DeviceConfiguration : IDisposable
	{
		internal bool _disposed = false;

		internal void Dispose(bool disposing)
		{
			if (!this._disposed)
			{
				// If disposing equals true, dispose all managed and unmanaged resources.
				if (disposing)
				{
					if (this._regKey != null)
					{
						this._regKey.Dispose();
						this._regKey = null;
					}
				}

				//
				// cleanup unmanaged resources
				//

				this._disposed = true;
			}
		}

		/// <summary>
		/// Disposes the object.
		/// </summary>
		public void Dispose()
		{
			this.Dispose(true);

			// This object will be cleaned up by the Dispose method.
			// Therefore, you should call GC.SupressFinalize to take this object off the finalization queue
			// and prevent finalization code for this object from executing a second time.
			GC.SuppressFinalize(this);
		}
	}

	#endregion Disposable
}
