﻿using AirFusion.Common;
using AirFusion.Common.JsonConverters;
using AirFusion.Common.Network;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Text;

namespace AirFusion.Manager.DeviceManager
{
	/// <summary>
	/// Represents the instance profile on Dandelion/QRCode.
	/// </summary>
	public sealed partial class InstanceProfile
	{
		/// <summary>
		/// Gets or sets the instance ID.
		/// </summary>
		[JsonProperty("ID"), JsonConverter(typeof(GuidConverter))]
		public Guid? InstanceID { get; internal set; }

		/// <summary>
		/// Gets or sets the device name.
		/// </summary>
		[JsonProperty("DN")]
		public string DeviceName { get; internal set; }

		/// <summary>
		/// Gets or sets the user name.
		/// </summary>
		[JsonProperty("UN")]
		public string UserName { get; internal set; }

		/// <summary>
		/// Gets or sets the type of operating system of the device.
		/// </summary>
		[JsonProperty("OS"), JsonConverter(typeof(EnumConverter))]
		public AFUserDevice.OSTypeValue? OS { get; internal set; }

		/// <summary>
		/// Gets or sets the port number of web service.
		/// </summary>
		[JsonProperty("P")]
		public ushort? WebServicePort { get; internal set; }

		/// <summary>
		/// Gets or sets the network information.
		/// </summary>
		[JsonProperty("N")]
		public NetworkInfo[] Network
		{
			get
			{
				return this._network;
			}
			internal set
			{
				this._network = value;
			}
		}

		[JsonIgnore]
		private NetworkInfo[] _network;

		/// <summary>
		/// Initializes a new instance of the <see cref="InstanceProfile"/> class.
		/// </summary>
		internal InstanceProfile()
		{
			this.InstanceID = null;
			this.DeviceName = null;
			this.UserName = null;
			this.OS = null;
			this.WebServicePort = null;
			this.Network = null;
		}

		/// <summary>
		/// Returns a <see cref="System.String" /> that represents this instance.
		/// </summary>
		/// <returns>
		/// A <see cref="System.String" /> that represents this instance.
		/// </returns>
		public override string ToString()
		{
			StringBuilder builder = new StringBuilder();
			builder.AppendLine(this.GetType().GetFullName() + ".InstanceID: " + (this.InstanceID.HasValue ? this.InstanceID.Value.GetStringFormat() : "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".DeviceName: " + (this.DeviceName ?? "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".UserName: " + (this.UserName ?? "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".OS: " + (this.OS.HasValue ? this.OS.Value.ToString() : "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".WebServicePort: " + (this.WebServicePort.HasValue ? this.WebServicePort.Value.ToString() : "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".Network: " + (this.Network != null ? "\r\n" + string.Join("\r\n", this.Network.Select(item => item.ToString())) : "N/A"));

			return builder.ToString().TrimEnd();
		}

		/// <summary>
		/// Returns a <see cref="System.String" /> that represents this instance with JSON format.
		/// </summary>
		/// <returns></returns>
		public string ToJsonString()
		{
			return CommonUtils.SerializeAsJson(this, false);
		}

		/// <summary>
		/// Returns a <see cref="System.String" /> that represents this instance with base64 encodings.
		/// </summary>
		/// <returns></returns>
		public string ToBase64String()
		{
			return Convert.ToBase64String(Encoding.UTF8.GetBytes(this.ToJsonString()));
		}

		/// <summary>
		/// Returns an AFProfile instance from the specified JSON format string.
		/// </summary>
		/// <param name="jsonString">The JSON-format string.</param>
		/// <returns></returns>
		public static InstanceProfile FromJsonString(string jsonString)
		{
			return CommonUtils.DeserializeFromJson<InstanceProfile>(jsonString);
		}

		/// <summary>
		/// Returns an AFProfile instance from the specified JSON base64-encoded string.
		/// </summary>
		/// <param name="base64String">The base64-encoded string.</param>
		/// <returns></returns>
		public static InstanceProfile FromBase64String(string base64String)
		{
			return FromJsonString(Encoding.UTF8.GetString(Convert.FromBase64String(base64String)));
		}
	}

	public partial class InstanceProfile
	{
		/// <summary>
		/// Represents the network info.
		/// </summary>
		public class NetworkInfo
		{
			#region Enumerations

			/// <summary>
			/// Specifies the encryption type.
			/// </summary>
			public enum EncryptionType
			{
				/// <summary>
				/// Signifies WEP.
				/// </summary>
				[ParameterValue("WEP")]
				WEP,

				/// <summary>
				/// Signifies WPA.
				/// </summary>
				[ParameterValue("WPA")]
				WPA,

				/// <summary>
				/// Signifies WPA2.
				/// </summary>
				[ParameterValue("WPA2")]
				WPA2,

				/// <summary>
				/// Signifies no encryption.
				/// </summary>
				[ParameterValue(null)]
				None
			}

			#endregion Enumerations

			/// <summary>
			/// Gets or sets the address.
			/// </summary>
			[JsonProperty("IP"), JsonConverter(typeof(IPAddressConverter))]
			public IPAddress Address { get; internal set; }

			/// <summary>
			/// Gets or sets the SSID of the access point;
			/// </summary>
			///// <remarks>This value should be ignored if the <seealso cref="IsWireless"/> is false.</remarks>
			[JsonProperty("S")]
			public string Ssid { get; internal set; }

			/// <summary>
			/// Gets or sets the encryption of the access point;
			/// </summary>
			///// <remarks>This value should be ignored if the <seealso cref="IsWireless"/> is false.</remarks>
			[JsonProperty("E"), JsonConverter(typeof(EnumConverter))]
			public EncryptionType? Encryption { get; internal set; }

			/// <summary>
			/// Gets or sets the password of the access point;
			/// </summary>
			///// <remarks>This value should be ignored if the <seealso cref="IsWireless"/> is false.</remarks>
			[JsonProperty("P")]
			public string Password { get; internal set; }

			/// <summary>
			/// Gets a value that indicates whether the access point is hidden.
			/// </summary>
			///// <remarks>This value should be ignored if the <seealso cref="IsWireless"/> is false.</remarks>
			[JsonProperty("H")]
			public bool? IsHidden { get; internal set; }

			/// <summary>
			/// Initializes a new instance of the <see cref="NetworkInfo"/> class.
			/// </summary>
			internal NetworkInfo()
			{
				this.Address = null;
				this.Ssid = null;
				this.Encryption = null;
				this.Password = null;
				this.IsHidden = null;
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".Address: " + (this.Address != null ? this.Address.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Ssid: " + (this.Ssid ?? "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Encryption: " + (this.Encryption.HasValue ? this.Encryption.GetStringValue() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Password: " + (this.Password ?? "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".IsHidden: " + (this.IsHidden.HasValue ? this.IsHidden.Value.ToString() : "N/A"));

				return builder.ToString().TrimEnd();
			}
		}
	}

	#region Extensions

	internal static partial class AFExtensions
	{
		/// <summary>
		/// Gets the profile representation for the specified device.
		/// </summary>
		/// <param name="device">The device.</param>
		/// <returns></returns>
		internal static InstanceProfile ToProfile(this AFUserDevice device)
		{
			return new InstanceProfile()
			{
				InstanceID = device.InstanceID,
				DeviceName = device.DeviceName,
				UserName = device.UserName,
				OS = device.OSType,
				WebServicePort = device.WebServicePort
			};
		}

		/// <summary>
		/// Gets the <see cref="AFUserDevice"/> form for the specified profile representation.
		/// </summary>
		/// <param name="profile">The profile.</param>
		/// <returns></returns>
		internal static AFUserDevice ToDevice(this InstanceProfile profile)
		{
			AFUserDevice device = new AFUserDevice();

			device.InstanceID = profile.InstanceID.HasValue ? profile.InstanceID.Value : Guid.Empty;
			device.DeviceName = profile.DeviceName;
			device.UserName = profile.UserName;
			device.OSType = profile.OS.HasValue ? profile.OS.Value : AFUserDevice.OSTypeValue.Unknown;
			if (profile.WebServicePort.HasValue)
			{
				device.WebServicePort = profile.WebServicePort.Value;
			}

			return device;
		}

		/// <summary>
		/// Fills from the specified <see cref="InstanceProfile"/> object if such values are specified.
		/// </summary>
		/// <param name="device">The device.</param>
		/// <param name="profile">The profile.</param>
		internal static void FillFromIfSpecified(this AFUserDevice device, InstanceProfile profile)
		{
			// debug assertion
			Debug.Assert(profile != null);

			if (profile.InstanceID.HasValue && profile.InstanceID.Value.Equals(Guid.Empty) == false)
			{
				device.InstanceID = profile.InstanceID.Value;
			}

			if (profile.DeviceName != null)
			{
				device.DeviceName = profile.DeviceName;
			}

			if (profile.UserName != null)
			{
				device.UserName = profile.UserName;
			}

			if (profile.OS.HasValue)
			{
				device.OSType = profile.OS.Value;
			}

			if (profile.WebServicePort.HasValue)
			{
				device.WebServicePort = profile.WebServicePort.Value;
			}
		}

		/// <summary>
		/// Converts the format of network information.
		/// </summary>
		internal static InstanceProfile.NetworkInfo Transform(this NetworkInfo networkInfo)
		{
			try
			{
				StaticLogger.TrackEnter();

				InstanceProfile.NetworkInfo info = new InstanceProfile.NetworkInfo()
				{
					Address = networkInfo.Address,
				};

				if (networkInfo.IsWireless)
				{
					WirelessNetworkInfo wirelessNetworkInfo = (WirelessNetworkInfo)networkInfo;
					if (wirelessNetworkInfo != null)
					{
						info.Ssid = wirelessNetworkInfo.Ssid;
						info.IsHidden = wirelessNetworkInfo.IsHidden;
						info.Password = wirelessNetworkInfo.Password;

						switch (wirelessNetworkInfo.Encryption)
						{
							case WirelessEncryptionProtocolType.WEP:
								info.Encryption = InstanceProfile.NetworkInfo.EncryptionType.WEP;
								break;

							case WirelessEncryptionProtocolType.WPA:
								info.Encryption = InstanceProfile.NetworkInfo.EncryptionType.WPA;
								break;

							case WirelessEncryptionProtocolType.WPA2:
								info.Encryption = InstanceProfile.NetworkInfo.EncryptionType.WPA2;
								break;

							default:
								info.Encryption = InstanceProfile.NetworkInfo.EncryptionType.None;
								break;
						}
					}
				}
				else
				{
					info.Ssid = null;
					info.IsHidden = null;
					info.Password = null;
					info.Encryption = null;
				}

				return info;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}
	}

	#endregion Extensions
}