﻿using AirFusion.Common;
using InTheHand.Net;
using InTheHand.Net.Sockets;
using System;
using System.Collections.Generic;
using System.Threading.Tasks;

namespace AirFusion.Manager.DeviceManager
{
	/// <summary>
	/// Represents the Bluetooth wake up approach.
	/// </summary>
	internal sealed class BluetoothWakeUp : IWakeUp
	{
		#region Private/Protected/Internal

		private Dictionary<BluetoothAddress, Task> _wakeUpTask;

		#endregion Private/Protected/Internal

		#region Public

		public BluetoothWakeUp()
		{
			this._wakeUpTask = new Dictionary<BluetoothAddress, Task>();
		}

		#endregion Public

		#region Override IWakeUp

		/// <summary>
		/// Wakes up the specified device asynchronously.
		/// </summary>
		/// <param name="device">The device.</param>
		/// <returns></returns>
		/// <exception cref="System.NotSupportedException" />
		public async Task WakeUpAsync(AFUserDevice device)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequsite
				if (!CommonUtils.IsBluetoothSupported())
				{
					throw new NotSupportedException("There's no Bluetooth installed, unable to wake up by Bluetooth.");
				}
				if (!CommonUtils.IsBluetoothEnabled())
				{
					throw new NotSupportedException("There's no Bluetooth radio enabled, unable to wake up by Bluetooth.");
				}
				if (device.Bluetooth == null || device.Bluetooth.Address == null || device.Bluetooth.Address.Equals(BluetoothAddress.None))
				{
					throw new NotSupportedException("There's no Bluetooth address specified, unable to wake up by Bluetooth.");
				}

				// check if there's any previous connecting task, if yes, wait for its result
				if (this._wakeUpTask.ContainsKey(device.Bluetooth.Address))
				{
					var task = this._wakeUpTask[device.Bluetooth.Address];
					if (task != null && !task.IsCompleted)
					{
						try
						{
							StaticLogger.LogTrace("There's a previous connecting task running, wait for the result...");
							await task;
							StaticLogger.LogInfo("The device has been waken up successfully by previous task.");
							return;
						}
						catch (Exception ex)
						{
							throw new Exception("Failed waking up the '" + device + "' by the previous task.", ex);
						}
					}
				}

				// TBD:
				// 1. design the wakeup string or packet
				// 2. use existing socket if there's any
				// 3. if there's multiple listener (multiple user) in a single device, the connection might unable to be created since the previous one might stay open

				StaticLogger.LogTrace("Try to wake up the device by creating the Bluetooth connection...");
				StaticLogger.LogInfo("Bluetooth address: " + device.Bluetooth.Address.GetStringFormat());

				try
				{
					using (var client = new BluetoothClient())
					{
						var task = Task.Factory.FromAsync(client.BeginConnect, client.EndConnect, device.Bluetooth.Address, UserDeviceManager.BluetoothServiceUuid, null);
						this._wakeUpTask[device.Bluetooth.Address] = task;
						await task;
						StaticLogger.LogInfo("The Bluetooth connection has been connected successfully.");
						client.Close();
					}
				}
				catch (Exception ex)
				{
					throw new Exception("Failed connecting to the '" + device + "' by its Bluetooth address '" + device.Bluetooth.Address.GetStringFormat() + "'.", ex);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Override IWakeUp
	}
}