﻿using AirFusion.Common;
using AirFusion.WebService.Adapters;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.Manager.DeviceManager
{
	public sealed partial class AFUserDevice
	{
		/// <summary>
		/// Initializes a new instance of the <see cref="AFUserDevice"/> class.
		/// </summary>
		internal AFUserDevice()
		{
			// generic
			this.InstanceID = Guid.Empty;
			this.SessionID = null;

			// device-specific
			this.DeviceName = null;
			this.ModelName = null;
			this.Manufacturer = null;
			this.OS = null;
			this.OSVersion = null;
			this.SerialNumber = null;
			this.Display = new DisplayInfo();
			//this.Bluetooth = new BluetoothInfo(); // 2015-05-20 Seoker: remove Bluetooth related functions since it's been defeatured

			// user-specific
			this.UserName = null;

			// application-specific
			this.ApplicationVersion = null;
			this.HasSystemPermission = null;
			this.Services = new ServicesInfo();

			// others
			this._pairStatus = PairingStatus.UnPaired; // this way won't invoke the event unintentionally
			this.Address = null;
			this.WebServicePort = 0;
			this.Owner = null;
		}

		/// <summary>
		/// Fills from the specified device if such values are specified.
		/// </summary>
		/// <param name="source">The source object.</param>
		internal void FillFromIfSpecified(AFUserDevice source)
		{
			Trace.Assert(source != null, "The 'source' is null.");

			if (source.InstanceID.Equals(Guid.Empty) == false)
			{
				this.InstanceID = source.InstanceID;
			}

			if (source.SessionID != null)
			{
				this.SessionID = source.SessionID;
			}

			if (source.DeviceName != null)
			{
				this.DeviceName = source.DeviceName;
			}

			if (source.ModelName != null)
			{
				this.ModelName = source.ModelName;
			}

			if (source.Manufacturer != null)
			{
				this.Manufacturer = source.Manufacturer;
			}

			if (source.OS != null)
			{
				this.OS = source.OS;
			}

			if (source.OSVersion != null)
			{
				this.OSVersion = source.OSVersion;
			}

			if (source.SerialNumber != null)
			{
				this.SerialNumber = source.SerialNumber;
			}

			if (source.UserName != null)
			{
				this.UserName = source.UserName;
			}

			if (source.ApplicationVersion != null)
			{
				this.ApplicationVersion = source.ApplicationVersion;
			}

			if (source.HasSystemPermission != null)
			{
				this.HasSystemPermission = source.HasSystemPermission;
			}

			if (source.Address != null)
			{
				this.Address = source.Address;
			}

			this.WebServicePort = source.WebServicePort;

			if (source.Display != null)
			{
				if (this.Display == null)
				{
					this.Display = new DisplayInfo();
				}
				this.Display.FillFromIfSpecified(source.Display);
			}

			// 2015-05-20 Seoker: remove Bluetooth related functions since it's been defeatured
			/*
			if (source.Bluetooth != null)
			{
				if (this.Bluetooth == null)
				{
					this.Bluetooth = new BluetoothInfo();
				}
				this.Bluetooth.FillFromIfSpecified(source.Bluetooth);
			}
			*/

			if (source.Services != null)
			{
				if (this.Services == null)
				{
					this.Services = new ServicesInfo();
				}
				this.Services.FillFromIfSpecified(source.Services);
			}
		}

		/// <summary>
		/// Fills from the specified <see cref="DeviceInformation"/> object if such values are specified.
		/// </summary>
		/// <param name="source">The source object.</param>
		internal void FillFromIfSpecified(DeviceInformation source)
		{
			Trace.Assert(source != null, "The 'source' is null.");

			if (source.InstanceID.HasValue && source.InstanceID.Value.Equals(Guid.Empty) == false)
			{
				this.InstanceID = source.InstanceID.Value;
			}

			if (source.DeviceName != null)
			{
				this.DeviceName = source.DeviceName;
			}

			if (source.ModelName != null)
			{
				this.ModelName = source.ModelName;
			}

			if (source.Manufacturer != null)
			{
				this.Manufacturer = source.Manufacturer;
			}

			if (source.OS != null)
			{
				this.OS = source.OS;
			}

			if (source.OSVersion != null)
			{
				this.OSVersion = source.OSVersion;
			}

			if (source.SerialNumber != null)
			{
				this.SerialNumber = source.SerialNumber;
			}

			if (source.UserName != null)
			{
				this.UserName = source.UserName;
			}

			if (source.ApplicationVersion != null)
			{
				this.ApplicationVersion = source.ApplicationVersion;
			}

			if (source.HasSystemPermission != null)
			{
				this.HasSystemPermission = source.HasSystemPermission;
			}

			if (source.Display != null)
			{
				if (this.Display == null)
				{
					this.Display = new DisplayInfo();
				}
				this.Display.FillFromIfSpecified(source.Display);
			}

			// 2015-05-20 Seoker: remove Bluetooth related functions since it's been defeatured
			/*
			if (source.Bluetooth != null)
			{
				if (this.Bluetooth == null)
				{
					this.Bluetooth = new BluetoothInfo();
				}
				this.Bluetooth.FillFromIfSpecified(source.Bluetooth);
			}
			*/

			if (source.Services != null)
			{
				if (this.Services == null)
				{
					this.Services = new ServicesInfo();
				}
				this.Services.FillFromIfSpecified(source.Services);
			}
		}

		/// <summary>
		/// Fills the values into the specified <see cref="DeviceInformation"/> object.
		/// </summary>
		/// <param name="target">The target object.</param>
		internal void FillTo(DeviceInformation target)
		{
			Trace.Assert(target != null, "The 'target' is null.");

			target.InstanceID = this.InstanceID;
			target.DeviceName = this.DeviceName;
			target.ModelName = this.ModelName;
			target.UserName = this.UserName;
			target.Manufacturer = this.Manufacturer;
			target.OS = this.OS;
			target.OSVersion = this.OSVersion;
			target.SerialNumber = this.SerialNumber;
			target.ApplicationVersion = this.ApplicationVersion;
			target.HasSystemPermission = this.HasSystemPermission;

			if (this.Display != null)
			{
				if (target.Display == null)
				{
					target.Display = new DeviceInformation.DisplayInfo();
				}

				this.Display.FillTo(target.Display);
			}
			else
			{
				target.Display = null;
			}

			// 2015-05-20 Seoker: remove Bluetooth related functions since it's been defeatured
			/*
			if (this.Bluetooth != null)
			{
				if (target.Bluetooth == null)
				{
					target.Bluetooth = new DeviceInformation.BluetoothInfo();
				}
				this.Bluetooth.FillTo(target.Bluetooth);
			}
			else
			{
				target.Bluetooth = null;
			}
			*/

			if (this.Services != null)
			{
				if (target.Services == null)
				{
					target.Services = new DeviceInformation.ServicesInfo();
				}
				this.Services.FillTo(target.Services);
			}
			else
			{
				target.Services = null;
			}
		}

		/// <summary>
		/// Updates the pairing status asynchronously.
		/// </summary>
		/// <returns></returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		internal async Task UpdatePairingStatusAsync()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this.Address == null || this.Address.Equals(IPAddress.None))
				{
					throw new InvalidOperationException("The address is not specified, might due to the device is not found yet.");
				}

				// debug assertion
				Trace.Assert(this.Owner != null, "The 'Owner' is not specified.");
				Trace.Assert(this.Owner.PairingService != null, "The 'Owner.PairingService' is not specified.");
				Trace.Assert(this.Owner.LocalDevice != null, "The 'Owner.LocalDevice' is not specified.");
				Trace.Assert(this.Owner.LocalDevice.InstanceID.Equals(Guid.Empty) == false, "The 'Owner.LocalDevice.InstanceID' is Guid.Empty.");

				try
				{
					var pairingStatusData = new PairingStatusInfo()
					{
						PairingStatus = this.PairStatus == AFUserDevice.PairingStatus.Paired
					};

					var address = this.Address;
					var port = this.WebServicePort;
					StaticLogger.LogTrace("Update pairing status to '" + address + ":" + port + "'...");

					var response = await this.Owner.PairingService.UpdateStatusAsync(address, port, this.Owner.LocalDevice.InstanceID, pairingStatusData).ConfigureAwait(continueOnCapturedContext: false);
					if (response.IsMessageResponse)
					{
						StaticLogger.LogInfo("PairingService.UpdateStatusAsync() is done successfully: " + response.Message);
					}
					else
					{
						StaticLogger.LogInfo("PairingService.UpdateStatusAsync() is done successfully.");
					}
				}
				catch (ArgumentNullException ex)
				{
					Trace.Fail(ex.Message);
					throw;
				}
				catch (WebServiceException ex)
				{
					StaticLogger.LogWarn("PairingService.UpdateStatusAsync() returns: " + ex.Message);
					throw;
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed updating pairing status to device '" + this + "': " + ex.Message);
					throw;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Gets or creates the wake up tasks.
		/// </summary>
		/// <returns>The dictionary of the created tasks.</returns>
		private Dictionary<string, Task<bool>> GetOrCreateWakeUpTasks()
		{
			var wakeUpTasks = new Dictionary<string, Task<bool>>();

			Trace.Assert(_wakeUpObjects != null, "The '_wakeUpObjects' is null.");

			// add tasks that doesn't exist in the list
			foreach (var entry in _wakeUpObjects)
			{
				var taskName = entry.Key;
				var wakeUp = entry.Value;

				try
				{
					var task = wakeUp.WakeUpAsync(this).ContinueWith(delegate(Task finishedTask)
					{
						if (finishedTask.IsFaulted)
						{
							Exception ex = finishedTask.Exception;
							while (ex is AggregateException && ex.InnerException != null)
							{
								AggregateException ae = ex as AggregateException;

								foreach (var e in ae.Flatten().InnerExceptions)
								{
									if (e is WebServiceVersionNotMatchedException)
									{
										StaticLogger.LogError("WakeUpAsynce Version'" + this + "': " + e.Message);

										throw e;
									}									
								}

								break;
							}

							StaticLogger.LogWarn("Failed waking up the device (" + this.ToString() + ") by method '" + taskName + "', exception:\r\n" + CommonUtils.GetExceptionMessages(finishedTask.Exception));

							return false;
						}
						else
						{
							StaticLogger.LogInfo("The device '" + this + "' should be waken up by method '" + taskName + "'.");
							return true;
						}
					}, TaskContinuationOptions.ExecuteSynchronously);
					task.ConfigureAwait(continueOnCapturedContext: false);
					wakeUpTasks[entry.Key] = task;
				}
				catch (AggregateException ae)
				{
					foreach (var e in ae.Flatten().InnerExceptions)
					{
						if (e is WebServiceVersionNotMatchedException)
						{
							// Recover from the exception. Here we just
							// print the message for demonstration purposes.
							StaticLogger.LogError("WakeUpAsynce Version'" + this + "': " + e.Message);
							Console.WriteLine(e.Message);
						}
						else
						{
							throw;
						}
					}
				}
				catch (WebServiceVersionNotMatchedException wex)
				{
					StaticLogger.LogWarn("Failed waking up the device (" + this.ToString() + ") by method '" + taskName + "', exception:\r\n" + CommonUtils.GetExceptionMessages(wex));
					throw;
				}
				catch (Exception ex)
				{
					StaticLogger.LogWarn("Failed waking up the device (" + this.ToString() + ") by method '" + taskName + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
				}
			}

			return wakeUpTasks;
		}
	}
}