﻿using AirFusion.Common;
using AirFusion.Common.RegistryHolders;
using AirFusion.WebService.Adapters;
using System;
using System.Diagnostics;
using System.Text;

namespace AirFusion.Manager.DeviceManager
{
	public sealed partial class AFUserDevice
	{
		/// <summary>
		/// Represents the AirFusion services information of the device.
		/// </summary>
		public sealed partial class ServicesInfo
		{
			/// <summary>
			/// Gets or sets the properties of input control.
			/// </summary>
			[RegistryStoreTypeAttribute(StoreType.SubTree)]
			public InputControlInfo InputControl { get; set; }

			/// <summary>
			/// Gets or sets the properties of screencast.
			/// </summary>
			[RegistryStoreTypeAttribute(StoreType.SubTree)]
			public ScreencastInfo Screencast { get; set; }

			#region Private/Protected/Internal

			/// <summary>
			/// Initializes a new instance of the <see cref="ServicesInfo"/> class.
			/// </summary>
			internal ServicesInfo()
			{
				this.InputControl = new InputControlInfo();
				this.Screencast = new ScreencastInfo();
			}

			/// <summary>
			/// Fills from the specified <see cref="ServicesInfo"/> object if such values are specified.
			/// </summary>
			/// <param name="source">The source object.</param>
			internal void FillFromIfSpecified(ServicesInfo source)
			{
				Debug.Assert(source != null);

				// copy InputControl
				if (source.InputControl != null)
				{
					if (this.InputControl == null)
					{
						this.InputControl = new InputControlInfo();
					}
					this.InputControl.FillFromIfSpecified(source.InputControl);
				}

				// copy Screencast
				if (source.Screencast != null)
				{
					if (this.Screencast == null)
					{
						this.Screencast = new ScreencastInfo();
					}
					this.Screencast.FillFromIfSpecified(source.Screencast);
				}
			}

			/// <summary>
			/// Fills from the specified <see cref="DeviceInformation.ServicesInfo"/> object if such values are specified.
			/// </summary>
			/// <param name="source">The source object.</param>
			internal void FillFromIfSpecified(DeviceInformation.ServicesInfo source)
			{
				Debug.Assert(source != null);

				// copy InputControl
				if (source.InputControl != null)
				{
					if (this.InputControl == null)
					{
						this.InputControl = new InputControlInfo();
					}
					this.InputControl.FillFromIfSpecified(source.InputControl);
				}

				// copy Screencast
				if (source.Screencast != null)
				{
					if (this.Screencast == null)
					{
						this.Screencast = new ScreencastInfo();
					}
					this.Screencast.FillFromIfSpecified(source.Screencast);
				}
			}

			/// <summary>
			/// Fills the values into the specified <see cref="DeviceInformation.ServicesInfo"/> object.
			/// </summary>
			/// <param name="target">The target object.</param>
			internal void FillTo(DeviceInformation.ServicesInfo target)
			{
				Debug.Assert(target != null);

				// copy InputControl
				if (this.InputControl != null)
				{
					if (target.InputControl == null)
					{
						target.InputControl = new DeviceInformation.ServicesInfo.InputControlInfo();
					}
					this.InputControl.FillTo(target.InputControl);
				}
				else
				{
					target.InputControl = null;
				}

				// copy Screencast
				if (this.Screencast != null)
				{
					if (target.Screencast == null)
					{
						target.Screencast = new DeviceInformation.ServicesInfo.ScreencastInfo();
					}
					this.Screencast.FillTo(target.Screencast);
				}
				else
				{
					target.Screencast = null;
				}
			}

			#endregion Private/Protected/Internal

			#region Override

			/// <summary>
			/// Determines whether the specified object is equal to the current object.
			/// </summary>
			/// <param name="obj">The object to compare with the current object.</param>
			/// <returns></returns>
			public override bool Equals(object obj)
			{
				return (obj is ServicesInfo && this == (ServicesInfo)obj);
			}

			/// <summary>
			/// Returns the hash code of the instance.
			/// </summary>
			/// <returns></returns>
			public override int GetHashCode()
			{
				return base.GetHashCode() ^ this.InputControl.GetHashCode() ^ this.Screencast.GetHashCode();
			}

			/// <summary>
			/// Implements the operator ==.
			/// </summary>
			/// <param name="x">The first object.</param>
			/// <param name="y">The second object.</param>
			public static bool operator ==(ServicesInfo x, ServicesInfo y)
			{
				return (((object)x == null && (object)y == null) || Object.ReferenceEquals(x, y) ||
					((object)x != null && (object)y != null) &&
					x.InputControl == y.InputControl &&
					x.Screencast == y.Screencast);
			}

			/// <summary>
			/// Implements the operator !=.
			/// </summary>
			/// <param name="x">The first object.</param>
			/// <param name="y">The second object.</param>
			public static bool operator !=(ServicesInfo x, ServicesInfo y)
			{
				return !(x == y);
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".InputControl: " + (this.InputControl != null ? "\r\n" + this.InputControl.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Screencast: " + (this.Screencast != null ? "\r\n" + this.Screencast.ToString() : "N/A"));

				return builder.ToString().TrimEnd();
			}

			#endregion Override
		}
	}
}
