﻿using System;
using System.Diagnostics;
using System.Net.Http;

namespace AirFusion.WebService
{
	internal sealed class VersionLocalResponseFilter : WebServiceLocalResponseFilter
	{
		#region Definitions

		private const string ApplicationVersionKey = "Application-Version";
		private const string MinimumTargetVersionKey = "Min-Target-Version";

		#endregion Definitions

		/// <summary>
		/// Gets or sets the version of application.
		/// </summary>
		public string ApplicationVersion { get; internal set; }

		/// <summary>
		/// Gets or sets the minimum target version.
		/// </summary>
		public string MinimumTargetVersion { get; internal set; }

		/// <summary>
		/// Initializes a new instance of the <see cref="VersionLocalResponseFilter"/> class.
		/// </summary>
		/// <param name="applicationVersion">The application version.</param>
		/// <param name="minimumTargetVersion">The minimum target version.</param>
		public VersionLocalResponseFilter(string applicationVersion, string minimumTargetVersion)
		{
			this.ApplicationVersion = applicationVersion;
			this.MinimumTargetVersion = minimumTargetVersion;
		}

		/// <summary>
		/// Filters a local response.
		/// </summary>
		/// <param name="responseMessage">The request message, any modification could be made to this instance.</param>
		/// <returns>True if to continue next filters.</returns>
		/// <remarks>The exception thrown by this method should be considered as an error response, and stop executing next filters.</remarks>
		public override bool Filter(HttpResponseMessage responseMessage)
		{
			// check parameters
			if (responseMessage == null)
			{
				throw new ArgumentNullException("responseMessage");
			}

			Trace.Assert(responseMessage.Headers != null, "responseMessage.Headers = null");

			// fill Application-Version
			if (this.ApplicationVersion != null)
			{
				responseMessage.Headers.Add(ApplicationVersionKey, this.ApplicationVersion.ToString());
			}

			// fill Min-Target-Version
			if (this.MinimumTargetVersion != null)
			{
				responseMessage.Headers.Add(MinimumTargetVersionKey, this.MinimumTargetVersion.ToString());
			}

			return true;
		}
	}
}
