﻿using AirFusion.WebService.Adapters;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Net;
using System.Net.Http;

namespace AirFusion.WebService
{
	internal sealed class VersionRemoteResponseFilter : WebServiceRemoteResponseFilter
	{
		#region Definitions

		private const string ApplicationVersionKey = "Application-Version";
		private const string MinimumTargetVersionKey = "Min-Target-Version";
		private const HttpStatusCode UnMatchedVersionStatusCode = (HttpStatusCode)499;

		#endregion Definitions

		/// <summary>
		/// Gets or sets the version of application.
		/// </summary>
		public string ApplicationVersion { get; internal set; }

		/// <summary>
		/// Gets or sets the minimum target version.
		/// </summary>
		public string MinimumTargetVersion { get; internal set; }

		/// <summary>
		/// Initializes a new instance of the <see cref="VersionRemoteResponseFilter"/> class.
		/// </summary>
		/// <param name="applicationVersion">The application version.</param>
		/// <param name="minimumTargetVersion">The minimum target version.</param>
		public VersionRemoteResponseFilter(string applicationVersion, string minimumTargetVersion)
		{
			this.ApplicationVersion = applicationVersion;
			this.MinimumTargetVersion = minimumTargetVersion;
		}

		/// <summary>
		/// Filters a local request.
		/// </summary>
		/// <param name="requestMessage">The request message, any modification could be made to this instance.</param>
		/// <returns>True if to continue next filters.</returns>
		/// <remarks>The exception thrown by this method should be considered as an error response, and stop executing next filters.</remarks>
		public override bool Filter(HttpResponseMessage responseMessage)
		{
			// check parameters
			if (responseMessage == null)
			{
				throw new ArgumentNullException("responseMessage");
			}

			if (responseMessage.Headers == null)
			{
				return true;
			}

			if (responseMessage.StatusCode.Equals(UnMatchedVersionStatusCode))
			{
				throw new WebServiceVersionNotMatchedException(responseMessage);
			}

			return true;
		}
	}
}
