﻿using AirFusion.Common;
using Newtonsoft.Json;
using System;
using System.Net;
using System.Text;

namespace AirFusion.Manager.FeatureHelper
{
	/// <summary>
	/// Represents the data of screencast control.
	/// </summary>
	public class ScreencastControlInfo
	{
		/// <summary>
		/// Specifies the screencast operation.
		/// </summary>
		public enum ScreencastOperation
		{
			/// <summary>
			/// Signifies to play.
			/// </summary>
			[ParameterValue("play")]
			Play,

			/// <summary>
			/// Signifies to stop.
			/// </summary>
			[ParameterValue("stop")]
			Stop,

			/// <summary>
			/// Signifies to pause.
			/// </summary>
			[ParameterValue("pause")]
			Pause,

			/// <summary>
			/// Signifies to resume.
			/// </summary>
			[ParameterValue("resume")]
			Resume
		}

		/// <summary>
		/// Specifies the connection protocol.
		/// </summary>
		public enum ConnectionProtocol
		{
			/// <summary>
			/// Signifies UDP connection.
			/// </summary>
			[ParameterValue("udp")]
			Udp,

			/// <summary>
			/// Signifies TCP connection.
			/// </summary>
			[ParameterValue("tcp")]
			Tcp
		}

		/// <summary>
		/// Represents the resolution of screencast.
		/// </summary>
		public class ResolutionInfo
		{
			/// <summary>
			/// Gets or sets the width.
			/// </summary>
			[JsonProperty("Width")]
			public int? Width { get; set; }

			/// <summary>
			/// Gets or sets the height.
			/// </summary>
			[JsonProperty("Height")]
			public int? Height { get; set; }

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				return (this.Width.HasValue ? this.Width.Value.ToString() : "?") + " * " + (this.Height.HasValue ? this.Height.Value.ToString() : "?");
			}
		}

		/// <summary>
		/// Gets or sets the FPS, in frame/second.
		/// </summary>
		[JsonProperty("FPS")]
		public uint? Fps { get; set; }

		/// <summary>
		/// Gets or sets the interval of key frames, in seconds.
		/// </summary>
		[JsonProperty("IFrameInterval")]
		public uint? KeyFrameInterval { get; set; }

		/// <summary>
		/// Gets or sets the bitrate, in bps.
		/// </summary>
		[JsonProperty("Bitrate")]
		public uint? Bitrate { get; set; }

		/// <summary>
		/// Gets or sets the port number of receiver.
		/// </summary>
		[JsonProperty("DestinationPort")]
		public ushort? DestinationPort { get; set; }

		/// <summary>
		/// Gets or sets the connection protocol.
		/// </summary>
		[JsonProperty("Protocol"), JsonConverter(typeof(AirFusion.Common.JsonConverters.EnumConverter))]
		public ConnectionProtocol? Protocol { get; set; }

		/// <summary>
		/// Gets or sets the resolution.
		/// </summary>
		[JsonProperty("Resolution")]
		public ResolutionInfo Resolution { get; set; }

		/// <summary>
		/// Gets or sets the instance ID of requester.
		/// </summary>
		[JsonProperty("InstanceID"), JsonConverter(typeof(AirFusion.Common.JsonConverters.GuidConverter))]
		internal Guid? InstanceID { get; set; }

		/*
		/// <summary>
		/// Gets or sets the IP address of receiver.
		/// </summary>
		[JsonProperty("DestinationIP"), JsonConverter(typeof(AirFusion.Common.JsonConverters.IPAddressConverter))]
		[Obsolete("The IP address should be determined by Android end, this property is ignored if 'InstanceID' is set since Android version v1.00.1005.")]
		internal IPAddress DestinationIP { get; set; }
		*/

		/// <summary>
		/// Initializes a new instance of the <see cref="ScreencastControlInfo"/> class.
		/// </summary>
		public ScreencastControlInfo()
		{
			this.Fps = null;
			this.KeyFrameInterval = null;
			this.Bitrate = null;
			this.DestinationPort = null;
			this.Protocol = null;
			this.Resolution = null;
			this.InstanceID = null;
			//this.DestinationIP = null;
		}

		/// <summary>
		/// Returns a <see cref="System.String" /> that represents this instance.
		/// </summary>
		/// <returns>
		/// A <see cref="System.String" /> that represents this instance.
		/// </returns>
		public override string ToString()
		{
			StringBuilder builder = new StringBuilder();
			builder.AppendLine(this.GetType().GetFullName() + ".Fps: " + (this.Fps.HasValue ? this.Fps.Value.ToString() : "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".KeyFrameInterval: " + (this.KeyFrameInterval.HasValue ? this.KeyFrameInterval.Value.ToString() : "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".Bitrate: " + (this.Bitrate.HasValue ? this.Bitrate.Value.ToString() : "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".DestinationPort: " + (this.DestinationPort.HasValue ? this.DestinationPort.Value.ToString() : "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".Protocol: " + (this.Protocol.HasValue ? this.Protocol.Value.GetStringValue() : "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".Resolution: " + (this.Resolution != null ? this.Resolution.ToString() : "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".InstanceID: " + (this.InstanceID.HasValue ? this.InstanceID.Value.ToString() : "N/A"));
			//builder.AppendLine(this.GetType().GetFullName() + ".DestinationIP: " + (this.DestinationIP != null ? this.DestinationIP.ToString() : "N/A"));

			return builder.ToString().TrimEnd();
		}
	}
}