﻿using AirFusion.Common;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.Manager.IncomingCall
{
	/// <summary>
	/// Represents the handler of using native bluetooth APIs
	/// </summary>
	public sealed class PCBluetoothHandler
	{
		private static PCBluetoothHandler _bt_handler = null;

		/// <summary>
		/// Gets instance of Bluetooth Handler
		/// </summary>
		public static PCBluetoothHandler Instance
		{
			get
			{
				if (_bt_handler == null)
				{
					_bt_handler = new PCBluetoothHandler();
				}

				return _bt_handler;
			}
		}

		#region Native Bluetooth APIs & Structure

		private const int BLUETOOTH_MAX_NAME_SIZE = 248;

		#region Structures

		/// <summary>
		/// Provides the address of a Bluetooth device.
		/// </summary>
		[StructLayout(LayoutKind.Sequential)]
		private struct BLUETOOTH_ADDRESS
		{
			[MarshalAs(UnmanagedType.ByValArray, SizeConst = 6)]
			public byte[] address;

			public override string ToString()
			{
				return String.Format("{0}:{1}:{2}:{3}:{4}:{5}",
					address[5].ToString("x2"),
					address[4].ToString("x2"),
					address[3].ToString("x2"),
					address[2].ToString("x2"),
					address[1].ToString("x2"),
					address[0].ToString("x2")
					);
			}
		}

		/// <summary>
		/// Facilitates enumerating installed Bluetooth radios.
		/// </summary>
		private struct BLUETOOTH_FIND_RADIO_PARAMS
		{
			public int dwSize;
		}

		/// <summary>
		/// Specifies search criteria for Bluetooth device
		/// </summary>
		[StructLayout(LayoutKind.Sequential, Pack = 4)]
		private struct BLUETOOTH_DEVICE_SEARCH_PARAMS
		{
			public UInt32 size;         // 32 (pack = 4!)
			public bool returnAuthenticated;
			public bool returnRemembered;
			public bool returnUnknown;
			public bool returnConnected;
			public bool issueInquiry;
			public Byte timeoutMultiplier;
			public IntPtr hRadio;
		}

		/// <summary>
		/// Specifies a date and time, using individual members for the month, day, year, weekday, hour, minute, second, and millisecond.
		/// </summary>
		[StructLayout(LayoutKind.Sequential, Pack = 1)]
		private struct SYSTEMTIME
		{
			public UInt16 year;
			public UInt16 month;
			public UInt16 dayOfWeek;
			public UInt16 day;
			public UInt16 hour;
			public UInt16 minute;
			public UInt16 second;
			public UInt16 milliseconds;
		}

		/// <summary>
		/// Provides information about a Bluetooth device.
		/// </summary>
		[StructLayout(LayoutKind.Sequential, CharSet = CharSet.Auto, Pack = 1)]
		private struct BLUETOOTH_DEVICE_INFO
		{
			public UInt32 size;        // 560
			public UInt32 dummya;
			public BLUETOOTH_ADDRESS address;
			public UInt16 dummyb;
			public UInt32 classofDevice;
			public bool connected;
			public bool remembered;
			public bool authenticated;
			public SYSTEMTIME lastSeen;
			public SYSTEMTIME lastUsed;
			[MarshalAs(UnmanagedType.ByValTStr, SizeConst = BLUETOOTH_MAX_NAME_SIZE)]
			public String name;
		}

		#endregion Structures

		#region Functions

		/// <Summary>
		/// Begins the enumeration of local Bluetooth radios.
		/// </Summary>
		/// <param name="pbtfrp">Pointer to a BLUETOOTH_FIND_RADIO_PARAMS structure.</param>
		/// <param name="phRadio">Pointer to where the first enumerated radio handle will be returned.</param>
		/// <returns>Returns NULL upon failure.</returns>
		[DllImport("Bthprops.cpl", SetLastError = true)]
		private static extern IntPtr BluetoothFindFirstRadio(ref BLUETOOTH_FIND_RADIO_PARAMS pbtfrp, ref IntPtr phRadio);

		/// <summary>
		/// Closes the enumeration handle associated with finding Bluetooth radios
		/// </summary>
		/// <param name="hFind">Enumeration handle to close, obtained with a previous call to the BluetoothFindFirstRadio function.</param>
		/// <returns></returns>
		[DllImport("Bthprops.cpl", SetLastError = true)]
		private static extern bool BluetoothFindRadioClose(IntPtr hFind);

		/// <summary>
		///  Begins the enumeration Bluetooth devices.
		/// </summary>
		/// <param name="pbtsp">Pointer to a BLUETOOTH_DEVICE_SEARCH_PARAMS structure.</param>
		/// <param name="pbtdi">Pointer to a BLUETOOTH_DEVICE_INFO structure into which information about the first Bluetooth device found is placed. </param>
		/// <returns>Returns NULL upon failure.</returns>
		[DllImport("Bthprops.cpl", SetLastError = true)]
		private static extern IntPtr BluetoothFindFirstDevice(ref BLUETOOTH_DEVICE_SEARCH_PARAMS pbtsp, ref BLUETOOTH_DEVICE_INFO pbtdi);

		/// <summary>
		/// Finds the next local Bluetooth device.
		/// </summary>
		/// <param name="hFind">Handle for the query obtained in a previous call to the BluetoothFindFirstDevice function.</param>
		/// <param name="pbtdi">Pointer to a BLUETOOTH_DEVICE_INFO structure into which information about the next Bluetooth device found is placed.</param>
		/// <returns>
		/// Returns TRUE when the next device is successfully found.
		/// Returns FALSE upon error.
		/// </returns>
		[DllImport("Bthprops.cpl", SetLastError = true)]
		[return: MarshalAs(UnmanagedType.Bool)]
		private static extern bool BluetoothFindNextDevice(IntPtr hFind, ref BLUETOOTH_DEVICE_INFO pbtdi);

		/// <summary>
		/// Closes an enumeration handle that is associated with a device query.
		/// </summary>
		/// <param name="hFind">Handle for the query to be closed.</param>
		/// <returns>Returns TRUE when the handle is successfully closed. Returns FALSE upon error.</returns>
		[DllImport("Bthprops.cpl", SetLastError = true)]
		[return: MarshalAs(UnmanagedType.Bool)]
		private static extern bool BluetoothFindDeviceClose(IntPtr hFind);

		#endregion Functions

		#endregion Native Bluetooth APIs & Structure

		#region Methods

		/// <summary>
		/// Checks if the Bluetooth radio of local device is enabled or not.
		/// </summary>
		/// <returns>True: Enabled; False: Disabled</returns>
		public bool IsBluetoothEnabled()
		{
			IntPtr radioHandle = IntPtr.Zero;

			BLUETOOTH_FIND_RADIO_PARAMS radioParams = default(BLUETOOTH_FIND_RADIO_PARAMS);
			radioParams.dwSize = Marshal.SizeOf(radioParams);
			IntPtr m_bt = BluetoothFindFirstRadio(ref radioParams, ref radioHandle);

			if (m_bt == IntPtr.Zero)
				return false;

			BluetoothFindRadioClose(m_bt);
			return true;
		}

		/// <summary>
		/// Checks if the specified Bluetooth device has been paired with local device.
		/// </summary>
		/// <param name="bluetoothAddress">The target bluetooth device</param>
		/// <returns>True: Paired device; False: No pair</returns>
		public bool IsPairedBluetoothDevice(string bluetoothAddress)
		{
			if ((bluetoothAddress == null) || (bluetoothAddress == string.Empty))
				return false;

			BLUETOOTH_DEVICE_SEARCH_PARAMS searchParams = new BLUETOOTH_DEVICE_SEARCH_PARAMS();
			searchParams.returnAuthenticated = true;
			searchParams.returnConnected = false;
			searchParams.returnRemembered = false;
			searchParams.returnUnknown = false;
			searchParams.issueInquiry = false;
			searchParams.timeoutMultiplier = 5;
			searchParams.size = (uint)Marshal.SizeOf(searchParams);

			BLUETOOTH_DEVICE_INFO deviceInfo = default(BLUETOOTH_DEVICE_INFO);
			deviceInfo.size = (uint)Marshal.SizeOf(deviceInfo);

			// Get the first paired bluetooth device
			IntPtr findHandle = IntPtr.Zero;
			findHandle = BluetoothFindFirstDevice(ref searchParams, ref deviceInfo);
			if (findHandle == IntPtr.Zero)
			{
				return false;
			}

			// Check the first paired bluetooth device is the target bluetooth device
			if (0 == string.Compare(bluetoothAddress, deviceInfo.address.ToString(), true))
			{
				return true;
			}

			// Check the other paired bluetooth devices is the target
			do
			{
				if (!BluetoothFindNextDevice(findHandle, ref deviceInfo))
				{
					if (0 == string.Compare(bluetoothAddress, deviceInfo.address.ToString(), true))
					{
						return true;
					}

					break;
				}
			} while (true);

			return false;
		}

		#endregion Methods

		#region Workaround: Force to close QCA's windows

		/// <summary>
		/// Sent as a signal that a window or an application should terminate.
		/// </summary>
		private const int WM_CLOSE = 0x0010;

		/// <summary>
		/// Hides the window and activates another window.
		/// </summary>
		private const int SW_HIDE = 0;

		/// <summary>
		/// Finds window by Caption only. Note you must pass IntPtr.Zero as the first parameter.
		/// </summary>
		[DllImport("user32.dll", EntryPoint = "FindWindow", SetLastError = true)]
		private static extern IntPtr FindWindowByCaption(IntPtr ZeroOnly, string lpWindowName);

		/// <summary>
		/// Sends the specified message to a window or windows.
		/// </summary>
		/// <param name="hWnd">A handle to the window whose window procedure will receive the message.</param>
		/// <param name="Msg">The message to be sent.</param>
		/// <param name="wParam">Additional message-specific information.</param>
		/// <param name="lParam">Additional message-specific information.</param>
		/// <returns>The return value specifies the result of the message processing; it depends on the message sent.</returns>
		[DllImport("user32.dll", CharSet = CharSet.Auto)]
		private static extern IntPtr SendMessage(IntPtr hWnd, UInt32 Msg, IntPtr wParam, IntPtr lParam);

		/// <summary>
		/// Sets the specified window's show state.
		/// </summary>
		/// <param name="hWnd">A handle to the window.</param>
		/// <param name="nCmdShow">Controls how the window is to be shown.</param>
		/// <returns>
		///  If the window was previously visible, the return value is nonzero.
		///  If the window was previously hidden, the return value is zero.
		/// </returns>
		[DllImport("user32.dll")]
		private static extern bool ShowWindow(IntPtr hWnd, int nCmdShow);

		/// <summary>
		/// Closes the windows of QCA phone call if the windows are opened
		/// - EX: Window title = Acer S510 (00:18:60:d3:ae:0b)
		/// </summary>
		/// <param name="phone_name">Phone's name</param>
		/// <param name="phone_bt_mac">Phone's BT mac address</param>
		/// <param name="max_time_ms">The max time (milliseconds) that to find the windows of QCA</param>
		/// <param name="hide_only">
		///  - True: Only to hide windows, not to close
		///  - False: To close the windows
		/// </param>
		public void CloseQcaBluetoothWindows(string phone_name, string phone_bt_mac, int max_time_ms, bool hide_only)
		{
			StaticLogger.LogInfo(string.Format("Start to close ... (phone_name = {0}, phone_bt_mac = {1})", phone_name, phone_bt_mac));
			if ((phone_name == string.Empty) || (phone_bt_mac == string.Empty))
			{
				return;
			}

			bool flag_close_notifier = false;
			bool flag_close_phone_ui = false;
			int max_check_count = max_time_ms / 200;
			int check_count = 0;

			string notifier_title = "PhoneAnswerNotifier";
			string phone_ui_title = string.Format("{0} ({1})", phone_name, phone_bt_mac);

			IntPtr notifier_handle = IntPtr.Zero;
			IntPtr window_handle = IntPtr.Zero;

			try
			{
				do
				{
					// Wait for 0.2 seconds
					Thread.Sleep(200);

					// Close the notifier of QCA phone call
					if (false == flag_close_notifier)
					{
						notifier_handle = FindWindowByCaption(IntPtr.Zero, notifier_title);
						if (notifier_handle != IntPtr.Zero)
						{
							StaticLogger.LogInfo("Find the phone notifier (QCA) => To close the notifier !!");
							SendMessage(notifier_handle, WM_CLOSE, IntPtr.Zero, IntPtr.Zero);
							flag_close_notifier = true;
						}
					}

					// Close the window UI of QCA phone call
					if (false == flag_close_phone_ui)
					{
						window_handle = FindWindowByCaption(IntPtr.Zero, phone_ui_title);
						if (window_handle != IntPtr.Zero)
						{
							StaticLogger.LogInfo("Find the window UI (QCA) => To hide the window !!");
							ShowWindow(window_handle, SW_HIDE);
							flag_close_phone_ui = true;
						}
					}

					// If the notifier and window UI have been closed => exit this loop
					if (flag_close_notifier && flag_close_phone_ui)
					{
						break;
					}

					check_count++;
				} while (check_count < max_check_count);
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("Exception => " + ex.ToString());
			}

			if (!flag_close_notifier) { StaticLogger.LogDebug("Failed to close notifier (QCA) !!"); }
			if (!flag_close_phone_ui)
			{
				StaticLogger.LogDebug("Failed to close window UI (QCA) !!");
			}
			else
			{
				if (!hide_only)
				{
					Thread.Sleep(5000);
					StaticLogger.LogInfo("Waiting for 5 seconds => To close the QCA window !!");
					// Close the window UI of QCA
					SendMessage(window_handle, WM_CLOSE, IntPtr.Zero, IntPtr.Zero);
				}
			}
		}

		#endregion Workaround: Force to close QCA's windows
	}
}
