﻿using AirFusion.Common;
using AirFusion.Manager.DeviceManager;
using AirFusion.Manager.FeatureHelper;
//using AirFusion.Manager.IncomingCall; // 2015-05-20 Seoker: remove incoming call manager since it's defeatured
using AirFusion.Manager.Notification;
using AirFusion.WebService;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.Manager
{
	#region Public Methods

	/// <summary>
	/// Represents the manager of Air Fusion.
	/// </summary>
	public sealed partial class AirFusionManager
	{
		#region Properties

		/// <summary>
		/// Gets the web service manager of Air Fusion.
		/// </summary>
		public WebServicesManager WebServiceManager { get; private set; }

		/// <summary>
		/// Gets the user device manager of Air Fusion.
		/// </summary>
		public UserDeviceManager UserDeviceManager { get; private set; }

		/// <summary>
		/// Gets the notification manager of Air Fusion.
		/// </summary>
		public NotificationManager NotificationManager { get; private set; }

		// 2015-05-20 Seoker: remove incoming call manager since it's defeatured
		/*
		/// <summary>
		/// Gets the incoming call manager
		/// </summary>
		public IncomingCallManager IncomingCallManager { get; private set; }
		*/

		/// <summary>
		/// Gets the configuration helper.
		/// </summary>
		public ConfigurationHelper ConfigurationHelper { get; private set; }

		/// <summary>
		/// Gets the screencast helper.
		/// </summary>
		public ScreencastHelper ScreencastHelper { get; private set; }

		/// <summary>
		/// Gets the input control helper.
		/// </summary>
		public InputControlHelper InputControlHelper { get; private set; }

		/// <summary>
		/// Gets the mix helper.
		/// </summary>
		public MixHelper MixHelper { get; private set; }

		/// <summary>
		/// Gets or sets the application version.
		/// </summary>
		public Version ApplicationVersion { get; set; }

		/// <summary>
		/// Gets or sets the minimum target version.
		/// </summary>
		public Version MinimumTargetVersion { get; set; }

		#endregion Properties

		#region Public Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="AirFusionManager"/> class.
		/// </summary>
		/// <remarks>This instance must be initialized by the main thread.</remarks>
		public AirFusionManager()
		{
			// initialize web service
			this.WebServiceManager = new WebServicesManager();

			// initialize the device manager
			this.UserDeviceManager = new UserDeviceManager(); // this instance must be initialized by the main thread.

			// initialize the notification manager
			this.NotificationManager = new NotificationManager();

			// 2015-05-20 Seoker: remove incoming call manager since it's defeatured
			/*
			// initialize the incoming call manager
			this.IncomingCallManager = new AFIncomingCallManager();
			*/

			// initialize the configuration helper
			this.ConfigurationHelper = new ConfigurationHelper(this.UserDeviceManager, this.WebServiceManager);

			// initialize the screencast helper
			this.ScreencastHelper = new ScreencastHelper(this.UserDeviceManager, this.WebServiceManager);

			// initialize the input control helper
			this.InputControlHelper = new InputControlHelper(this.UserDeviceManager, this.WebServiceManager);

			// initialize the mix helper
			this.MixHelper = new MixHelper(this.UserDeviceManager, this.WebServiceManager);
		}

		/// <summary>
		/// Initializes the instance.
		/// </summary>
		/// <remarks>
		/// Event handlers must be set before initializing this object, otherwise some of the event might not be triggered before the event handlers being set.
		/// </remarks>
		public void Initialize()
		{
			this.NotificationManager.SetWebService(this.WebServiceManager);
			this.NotificationManager.SetUserDeviceManager(this.UserDeviceManager);

			// 2015-05-20 Seoker: remove incoming call manager since it's defeatured
			/*
			this.IncomingCallManager.SetWebService(this.WebServiceManager);
			this.IncomingCallManager.SetUserDeviceManager(this.UserDeviceManager);
			*/

			// start the web service
			this.StartWebService();

			// initialize AFUserDeviceManager
			this.UserDeviceManager.Initialize(new InitialParameters()
			{
				WebServiceManager = this.WebServiceManager
			});
		}

		#endregion Public Methods

		#region Private/Protected/Internal Methods

		/// <summary>
		/// Starts the web service.
		/// </summary>
		/// <exception cref="System.Exception" />
		private void StartWebService()
		{
			bool opened = false;
			foreach (var port in WebServicePort)
			{
				try
				{
					this.WebServiceManager.StartListen(port);
					opened = true;
					return;
				}
				catch (System.Net.Sockets.SocketException ex)
				{
					StaticLogger.LogWarn("Failed starting web service listening with port=" + port + ", try another port. Exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
					continue;
				}
				catch (Exception ex)
				{
					throw new Exception("Failed starting web service listening.", ex);
				}
			}

			if (!opened)
			{
				throw new Exception("Failed starting web service listening since all the desired ports are unavailable.");
			}
		}

		#endregion Private/Protected/Internal Methods

		#region Disposable

		internal bool _disposed = false;

		internal void Dispose(bool disposing)
		{
			if (!this._disposed)
			{
				// If disposing equals true, dispose all managed and unmanaged resources.
				if (disposing)
				{
					if (this.UserDeviceManager != null)
					{
						this.UserDeviceManager.Dispose();
						this.UserDeviceManager = null;
					}
				}

				//
				// cleanup unmanaged resources
				//

				this._disposed = true;
			}
		}

		/// <summary>
		/// Disposes the object.
		/// </summary>
		public void Dispose()
		{
			this.Dispose(true);

			// This object will be cleaned up by the Dispose method.
			// Therefore, you should call GC.SupressFinalize to take this object off the finalization queue
			// and prevent finalization code for this object from executing a second time.
			GC.SuppressFinalize(this);
		}

		#endregion Disposable
	}

	#endregion Public Methods
}
