﻿using AirFusion.Common;
using AirFusion.Common.JsonConverters;
using Newtonsoft.Json;
using System;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of notification launch (PC => Phone)
	/// (POST /notification/launch).
	/// </summary>
	internal sealed class NotificationLaunchAdapter : WebServiceAdapter<NotificationLaunchAdapter.NotificationLaunchInfo, DBNull>
	{
		#region Definitions

		internal const string WebServiceUri = "/notification/launch";
		internal static HttpMethod WebServiceMethod = HttpMethod.Post;

		#endregion Definitions

		#region Classes

		/// <summary>
		/// Represents the request data of notification launch.
		/// </summary>
		public class NotificationLaunchInfo
		{
			/// <summary>
			/// Gets the notification id
			/// </summary>
			[JsonProperty("id")]
			public int? Id { get; set; }

			/// <summary>
			/// Gets the notification user id
			/// </summary>
			[JsonProperty("user_id")]
			public int? UserId { get; set; }

			/// <summary>
			/// Gets the notification tag
			/// </summary>
			[JsonProperty("tag")]
			public string Tag { get; set; }

			/// <summary>
			/// Gets the package name of apk which generates the notification
			/// </summary>
			[JsonProperty("package_name")]
			public string PackageName { get; set; }

			/// <summary>
			/// Gets the device id
			/// </summary>
			[JsonProperty("InstanceID"), JsonConverter(typeof(GuidConverter))]
			public Guid? DeviceInstanceID { get; set; }

			/// <summary>
			/// Initializes a new instance of the <see cref="NotificationLaunchInfo"/> class.
			/// </summary>
			public NotificationLaunchInfo()
			{
				this.Id = null;
				this.UserId = null;
				this.Tag = null;
				this.PackageName = null;
				this.DeviceInstanceID = null;
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".Id: " + (this.Id.HasValue ? this.Id.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".UserId: " + (this.UserId.HasValue ? this.UserId.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Tag: " + this.Tag);
				builder.AppendLine(this.GetType().GetFullName() + ".PackageName: " + this.PackageName);
				builder.AppendLine(this.GetType().GetFullName() + ".DeviceInstanceID: " + (this.DeviceInstanceID.HasValue ? this.DeviceInstanceID.Value.GetStringFormat() : "N/A"));

				return builder.ToString().TrimEnd();
			}
		}

		#endregion Classes

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="NotificationLaunchAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		internal NotificationLaunchAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager)
		{
		}

		/// <summary>
		/// Send a notification launch request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="launchData">The required launch data to specify which notification to launch.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> RequestLaunchAsync(IPAddress address, ushort port, NotificationLaunchInfo launchData)
		{
			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}
			if (launchData == null)
			{
				throw new ArgumentNullException("launchData");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri).Uri,
				launchData);

			return await this.SendRequestAsync(request, this.Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
