﻿using AirFusion.Common;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.WebService
{
	/// <summary>
	/// Represents the HTTP remote request handler.
	/// </summary>
	public class HttpRemoteRequestHandler
	{
		#region Properties

		/// <summary>
		/// Gets the client.
		/// </summary>
		public TcpClient Client { get; internal set; }

		/// <summary>
		/// Gets the request message.
		/// </summary>
		public HttpRemoteRequestMessage RequestMessage { get; internal set; }

		#endregion Properties

		#region Private/Protected/Internal Variables

		internal Stream _inputStream;
		internal Stream _outputStream;

		#endregion Private/Protected/Internal Variables

		#region Methods

		/// <summary>
		/// Creates the <seealso cref="HttpRemoteRequestHandler"/> instance.
		/// </summary>
		/// <param name="client">The client.</param>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.ArgumentNullException" />
		internal HttpRemoteRequestHandler(TcpClient client)
		{
			if (client == null)
			{
				throw new ArgumentNullException("client");
			}

			this.Client = client;
			this._inputStream = new BufferedStream(this.Client.GetStream());
			this._outputStream = new BufferedStream(this.Client.GetStream());
			this.RequestMessage = HttpRemoteRequestMessage.Create(this._inputStream);
		}

		/// <summary>
		/// Reads the HTTP request asynchronously.
		/// </summary>
		/// <exception cref="System.Exception" />
		/// <exception cref="System.FormatException" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.NotSupportedException" />
		/// <exception cref="System.ObjectDisposedException" />
		/// <exception cref="System.UriFormatException" />
		public async Task ReadRequestAsync()
		{
			try
			{
				StaticLogger.TrackEnter();

				Trace.Assert(this.RequestMessage != null, "this.RequestMessage = null");

				try
				{
					await this.RequestMessage.ReadRequestAsync();
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed reading request, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
					throw;
				}

				try
				{
					await this.RequestMessage.ReadHeaderAsync();
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed reading header, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
					throw;
				}
			}
			catch (ArgumentException ex)
			{
				Trace.Fail("The exception is an ArgumentException: " + ex.Message);
				throw;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Writes the response to the client asynchronously.
		/// </summary>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.IO.IOException" />
		/// <exception cref="System.NotSupportedException" />
		/// <exception cref="System.ObjectDisposedException" />
		public async Task WriteResponseAsync(HttpResponseMessage responseMessage)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (responseMessage == null)
				{
					throw new ArgumentNullException("responseMessage");
				}

				// check prerequisite
				if (!this._outputStream.CanWrite)
				{
					throw new NotSupportedException("The output stream is not able to write.");
				}

				if (responseMessage.RequestMessage == null)
				{
					responseMessage.RequestMessage = this.RequestMessage;
				}

				// debug assertion
				Trace.Assert(this._outputStream != null, "this._outputStream = null");

				// message type
				Debug.Assert(responseMessage.Version != null);
				var messageType = "HTTP/" + responseMessage.Version.ToString() + " " + (int)responseMessage.StatusCode + " " + responseMessage.ReasonPhrase;
				StaticLogger.LogDebug("Message type to response:\r\n" + messageType);
				await WriteAsync(messageType, this._outputStream);

				// response headers
				if (responseMessage.Headers != null)
				{
					var headers = responseMessage.Headers.ToString().TrimEnd('\r', '\n');
					StaticLogger.LogDebug("Headers to response:\r\n" + headers);
					if (!string.IsNullOrWhiteSpace(headers))
					{
						await WriteAsync(headers, this._outputStream);
					}
				}

				// content
				if (responseMessage.Content != null)
				{
					await WriteContentAsync(responseMessage.Content, this._outputStream);

					// experiment: failed
					// await responseMessage.Content.CopyToAsync(this._outputStream.BaseStream);
				}
				else
				{
					await WriteContentAsync(new StringContent(string.Empty, Encoding.UTF8), this._outputStream);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Writes the specified text asynchronously.
		/// </summary>
		/// <param name="text">The text.</param>
		/// <param name="outputStream">The output stream.</param>
		/// <param name="newLine">True to append a newline.</param>
		/// <exception cref="System.NotSupportedException" />
		/// <exception cref="System.ObjectDisposedException" />
		/// <exception cref="System.InvalidOperationException" />
		private static async Task WriteAsync(string text, Stream outputStream, bool newLine = true)
		{
			// debug assertion
			Trace.Assert(text != null, "text = null");
			Trace.Assert(outputStream != null, "outputStream = null");

			byte[] output = Encoding.UTF8.GetBytes(text + "\r\n");
			try
			{
				await outputStream.WriteAsync(output, 0, output.Length);
				await outputStream.FlushAsync();
			}
			catch (ArgumentException ex)
			{
				Trace.Fail("The exception is an ArgumentException: " + ex.Message);
				throw;
			}
		}

		/// <summary>
		/// Writes the specified content to the specified output stream asynchronously.
		/// </summary>
		/// <param name="content">The content.</param>
		/// <param name="outputStream">The output stream.</param>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.IO.IOException" />
		/// <exception cref="System.NotSupportedException" />
		/// <exception cref="System.ObjectDisposedException" />
		private async static Task WriteContentAsync(HttpContent content, Stream outputStream)
		{
			// debug assertion
			Trace.Assert(content != null, "content = null");
			Trace.Assert(outputStream != null, "outputStream = null");

			// content header
			if (content.Headers != null)
			{
				string contentHeaders = content.Headers.ToString().TrimEnd('\r', '\n');
				StaticLogger.LogDebug("Content Headers:\r\n" + contentHeaders);
				if (!string.IsNullOrWhiteSpace(contentHeaders))
				{
					await WriteAsync(contentHeaders, outputStream);
				}
			}

			// empty line
			await WriteAsync("", outputStream);

			if (content is StringContent)
			{
				var text = await (content as StringContent).ReadAsStringAsync();
				await WriteAsync(text, outputStream, false);
			}
			else if (content is ByteArrayContent)
			{
				var byteArrayContent = await (content as ByteArrayContent).ReadAsByteArrayAsync();
				await outputStream.WriteAsync(byteArrayContent, 0, byteArrayContent.Length);
			}
			else if (content is StreamContent)
			{
				var streamContent = await (content as StreamContent).ReadAsStreamAsync();
				await streamContent.CopyToAsync(outputStream); // TODO: not verified
			}
			else if (content is MultipartContent)
			{
				foreach (var c in (content as MultipartContent))
				{
					await WriteContentAsync(c, outputStream);
				}
			}

			await outputStream.FlushAsync();
		}

		#endregion Methods
	}
}
