﻿using AirFusion.Common;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.WebService
{
	/// <summary>
	/// Represnets the HTTP request message of remote request.
	/// </summary>
	public class HttpRemoteRequestMessage : HttpRequestMessage
	{
		#region Properties

		/// <summary>
		/// Gets a value indicating whether the header has been read.
		/// </summary>
		public bool HasRequestRead { get; private set; }

		/// <summary>
		/// Gets a value indicating whether the header has been read.
		/// </summary>
		public bool HasHeaderRead { get; private set; }

		/// <summary>
		/// Gets a value indicating whether the content has been read.
		/// </summary>
		public bool HasContentRead
		{
			get
			{
				return this.Content != null;
			}
		}

		/// <summary>
		/// Gets a value that indicates whether this request contains a message body.
		/// </summary>
		public bool HasContent
		{
			get
			{
				return this._contentLength > 0;
			}
		}

		/// <summary>
		/// Gets the payload content.
		/// </summary>
		/// <remarks>
		/// This property will be in string type, byte[] type, or Stream type.
		/// </remarks>
		public object PayloadContent { get; internal set; }

		#endregion Properties

		#region Private/Protected/Internal Variables

		private Stream _stream;
		private long _contentLength;
		private string _contentType;
		private string _charset;

		#endregion Private/Protected/Internal Variables

		#region Definitions

		private const int _maxContentLength = 10485760; // 10MB
		private const int _bufSize = 4096;

		#endregion Definitions

		#region Static Methods

		/// <summary>
		/// Creates an <seealso cref="HttpRemoteRequestMessage" /> instance with the specified input stream.
		/// </summary>
		/// <param name="inputStream">The input stream.</param>
		/// <returns>The created <seealso cref="HttpRemoteRequestMessage" /> instance.</returns>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.ArgumentNullException" />
		public static HttpRemoteRequestMessage Create(Stream inputStream)
		{
			return new HttpRemoteRequestMessage(inputStream);
		}

		/// <summary>
		/// Reads a line from the input stream asynchronously.
		/// </summary>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.NotSupportedException" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.ObjectDisposedException" />
		/// <returns>The next line.</returns>
		private static async Task<string> ReadLineAsync(Stream stream)
		{
			// check parameters
			if (stream == null)
			{
				throw new ArgumentNullException("stream");
			}
			if (!stream.CanRead)
			{
				throw new NotSupportedException("The specified stream is unable to read.");
			}

			int tryTimes = 0;
			byte[] buffer = new byte[_bufSize];
			int offset = 0;
			while (offset < _bufSize)
			{
				int read = await stream.ReadAsync(buffer, offset, 1);
				if (read == 0)
				{
					tryTimes++;
					if (tryTimes > 100)
					{
						throw new Exception("Failed reading any data in 100ms.");
					}
					Thread.Sleep(1);
					continue;
				}
				else
				{
					tryTimes = 0;
					var readChar = Convert.ToChar(buffer[offset]);
					if (readChar.Equals('\n'))
					{
						break;
					}
					offset++;
				}
			}

			return Encoding.UTF8.GetString(buffer, 0, offset).TrimEnd('\r', '\n');
		}

		/// <summary>
		/// Reads stream into a byte array with the specified length.
		/// </summary>
		/// <param name="stream">The input stream.</param>
		/// <param name="length">The length to read.</param>
		/// <returns>The bytes array read.</returns>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.ArgumentOutOfRangeException" />
		/// <exception cref="System.NotSupportedException" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.ObjectDisposedException" />
		private static async Task<byte[]> ReadFromStreamAsync(Stream stream, int length)
		{
			// check parameters
			if (stream == null)
			{
				throw new ArgumentNullException("stream");
			}
			if (!stream.CanRead)
			{
				throw new NotSupportedException("The specified stream is unable to read.");
			}
			if (length <= 0)
			{
				throw new ArgumentOutOfRangeException("length", "The length should be larger than zero.");
			}

			int tryTimes = 0;
			int toRead = length;
			int offset = 0;
			byte[] byteArray = new byte[length];

			while (toRead > 0)
			{
				try
				{
					int read = await stream.ReadAsync(byteArray, offset, toRead);
					if (read == 0)
					{
						tryTimes++;
						if (tryTimes > 100)
						{
							throw new Exception("Failed reading any data in 100ms.");
						}
						Thread.Sleep(1);
						continue;
					}
					else
					{
						tryTimes = 0;
						toRead -= read;
						offset += read;
					}
				}
				catch (ArgumentException ex)
				{
					Trace.Fail("The exception is an ArgumentException: " + ex.Message);
					return null;
				}
			}

			return byteArray;
		}

		#endregion Static Methods

		#region Public Methods

		/// <summary>
		/// Prevents a default instance of the <see cref="HttpRemoteRequestMessage"/> class from being created.
		/// </summary>
		private HttpRemoteRequestMessage()
		{
			this.HasRequestRead = false;
			this.HasHeaderRead = false;
			this.PayloadContent = null;
			this._stream = null;
			this._contentLength = 0;
			this._contentType = null;
			this._charset = null;
		}

		/// <summary>
		/// Creates a <seealso cref="HttpRemoteRequestMessage" /> instance.
		/// </summary>
		/// <param name="stream">The input stream.</param>
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.NotSupportedException" />
		private HttpRemoteRequestMessage(Stream stream)
			: this()
		{
			// check prerequisite
			if (stream == null)
			{
				throw new ArgumentNullException("stream");
			}
			if (!stream.CanRead)
			{
				throw new NotSupportedException("The specified stream is unable to read.");
			}

			this._stream = stream;
		}

		/// <summary>
		/// Reads the HTTP message type asynchronously.
		/// </summary>
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.FormatException" />
		/// <exception cref="System.NotSupportedException" />
		/// <exception cref="System.ObjectDisposedException" />
		/// <exception cref="System.UriFormatException" />
		public async Task<Tuple<HttpMethod, Uri, Version>> ReadRequestAsync()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this.HasRequestRead)
				{
					throw new InvalidOperationException("The request line has been read.");
				}

				// read the message type
				string messageType = await ReadLineAsync(this._stream);
				if (messageType == null)
				{
					throw new Exception("Failed reading line.");
				}

				StaticLogger.LogInfo(messageType);
				string[] tokens = messageType.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries).Select(token => token.Trim()).ToArray();
				if (tokens == null || tokens.Length != 3)
				{
					throw new Exception("Invalid HTTP message type: " + messageType);
				}

				// Method
				string httpMethod = tokens[0].ToUpper();
				switch (httpMethod)
				{
					case "DELETE":
						this.Method = HttpMethod.Delete;
						break;

					case "GET":
						this.Method = HttpMethod.Get;
						break;

					case "HEAD":
						this.Method = HttpMethod.Head;
						break;

					case "OPTIONS":
						this.Method = HttpMethod.Options;
						break;

					case "POST":
						this.Method = HttpMethod.Post;
						break;

					case "PUT":
						this.Method = HttpMethod.Put;
						break;

					case "TRACE":
						this.Method = HttpMethod.Trace;
						break;

					default:
						this.Method = new HttpMethod(httpMethod);
						break;
				}

				// Request URI & QueryString
				this.RequestUri = new Uri(tokens[1], UriKind.Relative);

				// Version
				if (tokens[2].StartsWith("HTTP/"))
				{
					var version = tokens[2].Substring(5).Trim();
					try
					{
						this.Version = new Version(version);
					}
					catch
					{
						StaticLogger.LogWarn("Bad version: " + version);
						this.Version = new Version("1.1");
					}
				}

				this.HasRequestRead = true;

				return new Tuple<HttpMethod, Uri, Version>(this.Method, this.RequestUri, this.Version);
			}
			catch (ArgumentException ex)
			{
				Trace.Fail("The exception is an ArgumentException: " + ex.Message);
				throw;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Reads the HTTP request headers asynchronously.
		/// </summary>
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.NotSupportedException" />
		/// <exception cref="System.ObjectDisposedException" />
		public async Task<HttpRequestHeaders> ReadHeaderAsync()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (this.HasHeaderRead)
				{
					throw new InvalidOperationException("The header has been read.");
				}

				// debug assertion
				Trace.Assert(this.Headers != null, "this.Headers = null");

				try
				{
					string line;
					while (!string.IsNullOrEmpty(line = await ReadLineAsync(this._stream)))
					{
						int separator = line.IndexOf(':');
						if (separator == -1)
						{
							throw new Exception("Invalid HTTP header line: " + line);
						}

						string name = line.Substring(0, separator);
						string value = (separator < line.Length - 1 ? line.Substring(separator + 1) : string.Empty).Trim();

						// Content-Type
						if (string.Compare(name, "Content-Type", true) == 0)
						{
							int indexOfSemicolon = value.IndexOf(';');
							if (indexOfSemicolon >= 0)
							{
								this._contentType = value.Substring(0, indexOfSemicolon).Trim();
								string remains = value.Substring(indexOfSemicolon).Trim();
								int indexOfCharset = remains.IndexOf("charset=");
								if (indexOfCharset >= 0 && indexOfCharset + 8 < remains.Length)
								{
									this._charset = remains.Substring(indexOfCharset + 8);
								}
							}
							else
							{
								this._contentType = value;
							}
						}
						// Content-Length
						else if (string.Compare(name, "Content-Length", true) == 0)
						{
							try
							{
								this._contentLength = Convert.ToInt64(value);
								if (this._contentLength > _maxContentLength)
								{
									throw new Exception("The Content-Length '" + value + "' is too large.");
								}
							}
							catch (Exception ex)
							{
								throw new Exception("Invalid Content-Length '" + value + "'.", ex);
							}
						}
						else
						{
							try
							{
								this.Headers.Add(name, value);
							}
							catch (Exception ex)
							{
								StaticLogger.LogWarn("Failed reading the header '" + name + ": " + value + "' as request header, skipped it, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
							}
						}
					}

					StaticLogger.LogInfo("HTTP Headers:");
					foreach (var entry in this.Headers)
					{
						StaticLogger.LogInfo("\t" + entry.Key + ": " + string.Join(", ", entry.Value));
					}

					this.HasHeaderRead = true;

					return this.Headers;
				}
				catch (ArgumentException ex)
				{
					Trace.Fail("The exception is an ArgumentException: " + ex.Message);
					throw;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Reads the content as a byte array asynchronously.
		/// </summary>
		/// <exception cref="System.NotSupportedException" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.ObjectDisposedException" />
		/// <exception cref="System.OverflowException" />
		public async Task<ByteArrayContent> ReadContentAsByteArrayAsync()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (!this.HasHeaderRead)
				{
					throw new InvalidOperationException("Please read the header first.");
				}
				if (this._contentLength <= 0)
				{
					throw new InvalidOperationException("There's no Content-Length or the Content-Length is invalid.");
				}
				if (this._contentLength > int.MaxValue)
				{
					throw new OverflowException("The Content-Type is too large, cannot read into byte array.");
				}

				try
				{
					var byteArrayContent = await ReadFromStreamAsync(this._stream, (int)this._contentLength);
					this.PayloadContent = byteArrayContent;
					var content = new ByteArrayContent(byteArrayContent);
					this.Content = content;

					this.SetContentHeaders();
					StaticLogger.LogInfo("Request payload (byte array):\r\n" + (byteArrayContent.Length > 32 ? byteArrayContent.ToHexString(0, 32) + "..." : byteArrayContent.ToHexString()));

					return content;
				}
				catch (ArgumentException ex)
				{
					Trace.Fail("The exception is an ArgumentException: " + ex.Message);
					throw;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Reads the content as a string with the specified encodings asynchronously.
		/// </summary>
		/// <param name="encoding">The encoding.</param>
		/// <exception cref="System.NotSupportedException" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.ObjectDisposedException" />
		/// <exception cref="System.OverflowException" />
		/// <exception cref="System.Text.DecoderFallbackException" />
		public async Task<StringContent> ReadContentAsStringAsync(Encoding encoding)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (encoding == null)
				{
					throw new ArgumentNullException("encoding");
				}

				// check prerequisite
				if (!this.HasHeaderRead)
				{
					throw new InvalidOperationException("Please read the header first.");
				}
				if (this._contentLength <= 0)
				{
					throw new InvalidOperationException("There's no Content-Length or the Content-Length is invalid.");
				}
				if (this._contentLength > int.MaxValue)
				{
					throw new OverflowException("The Content-Type is too large, cannot read into string value.");
				}

				try
				{
					var byteArrayContent = await ReadFromStreamAsync(this._stream, (int)this._contentLength);

					var stringContent = encoding.GetString(byteArrayContent);
					this.PayloadContent = stringContent;
					var content = string.IsNullOrEmpty(this._contentType) ? new StringContent(stringContent, encoding) : new StringContent(stringContent, encoding, this._contentType);
					this.Content = content;

					this.SetContentHeaders();
					StaticLogger.LogInfo("Request payload (string):\r\n" + stringContent.GetSubString());

					return content;
				}
				catch (ArgumentException ex)
				{
					Trace.Fail("The exception is an ArgumentException: " + ex.Message);
					throw;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Reads the content as a string asynchronously.
		/// </summary>
		/// <exception cref="System.NotSupportedException" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.ObjectDisposedException" />
		/// <exception cref="System.OverflowException" />
		/// <exception cref="System.Text.DecoderFallbackException" />
		public async Task<StringContent> ReadContentAsStringAsync()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (!this.HasHeaderRead)
				{
					throw new InvalidOperationException("Please read the header first.");
				}

				// get encodings
				var encoding = Encoding.UTF8;
				if (this._charset != null)
				{
					try
					{
						encoding = Encoding.GetEncoding(this._charset);
					}
					catch (ArgumentException)
					{
						StaticLogger.LogWarn("The specified charset '" + this._charset + "' is not a valid code page, use UTF8 instead.");
					}
				}

				return await this.ReadContentAsStringAsync(encoding);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Reads the content as a MemoryStream with size as described in Content-Length from the stream asynchronously.
		/// </summary>
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.NotSupportedException" />
		/// <exception cref="System.ObjectDisposedException" />
		[Obsolete("This function is not verified.")]
		public async Task<StreamContent> ReadContentAsStreamAsync()
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (!this.HasHeaderRead)
				{
					throw new InvalidOperationException("Please read the header first.");
				}
				if (this._contentLength <= 0)
				{
					throw new InvalidOperationException("There's no Content-Length or the Content-Length is invalid.");
				}

				var ms = new MemoryStream();

				byte[] buf = new byte[_bufSize];
				long toRead = this._contentLength;

				while (toRead > 0)
				{
					try
					{
						int read = await this._stream.ReadAsync(buf, 0, (int)Math.Min(toRead, _bufSize));
						if (read == 0)
						{
							throw new Exception("The client is disconnected during reading data.");
						}
						toRead -= read;
						await ms.WriteAsync(buf, 0, read);
					}
					catch (ArgumentException ex)
					{
						Trace.Fail("The exception is an ArgumentException: " + ex.Message);
						throw;
					}
				}

				this.PayloadContent = ms;
				var streamContent = new StreamContent(ms);
				this.Content = streamContent;

				this.SetContentHeaders();
				StaticLogger.LogInfo("Request payload is read (stream).");

				return streamContent;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Returns a <see cref="System.String" /> that represents this instance.
		/// </summary>
		/// <returns>
		/// A <see cref="System.String" /> that represents this instance.
		/// </returns>
		public override string ToString()
		{
			var builder = new StringBuilder();
			builder.AppendLine(this.GetType().GetFullName() + ".Method: " + (this.Method != null ? this.Method.ToString() : "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".RequestUri: " + (this.RequestUri != null ? this.RequestUri.ToString() : "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".Version: " + (this.Version != null ? this.Version.ToString() : "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".Headers: " + (this.Headers != null ? "\r\n" + this.Headers.ToString() : "N/A"));

			return builder.ToString();
		}

		#endregion Public Methods

		#region Private/Protected/Internal Methods

		/// <summary>
		/// Sets the content headers.
		/// </summary>
		private void SetContentHeaders()
		{
			Debug.Assert(this.Content != null);

			if (!this.Content.Headers.ContentLength.HasValue && this._contentLength > 0)
			{
				this.Content.Headers.ContentLength = this._contentLength;
			}

			// Content-Type
			if (!string.IsNullOrEmpty(this._contentType))
			{
				if (this.Content.Headers.ContentType == null)
				{
					this.Content.Headers.ContentType = new MediaTypeHeaderValue(this._contentType);
				}
				else
				{
					this.Content.Headers.ContentType.MediaType = this._contentType;
				}
			}

			// Content-Length
			if (!this.Content.Headers.ContentLength.HasValue && this._contentLength > 0)
			{
				this.Content.Headers.ContentLength = this._contentLength;
			}
		}

		#endregion Private/Protected/Internal Methods
	}
}
