﻿using AirFusion.Common;
using System;
using System.Diagnostics;
using System.Net;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Text;

namespace AirFusion.WebService.Adapters
{
	public partial class WebServiceAdapter<TRequest, TResponse>
	{
		/// <summary>
		/// Represents the response of a web service.
		/// </summary>
		public sealed class WebServiceResponse
		{
			#region Public Properties

			/// <summary>
			/// Gets a value indicating whether this instance is a message response.
			/// </summary>
			public bool IsMessageResponse
			{
				get
				{
					return this._message != null;
				}
			}

			/// <summary>
			/// Gets or sets the status code.
			/// </summary>
			public HttpStatusCode StatusCode
			{
				get
				{
					Trace.Assert(this.ResponseMessage != null, "this.ResponseMessage = null");
					return this.ResponseMessage.StatusCode;
				}
				set
				{
					Trace.Assert(this.ResponseMessage != null, "this.ResponseMessage = null");
					this.ResponseMessage.StatusCode = value;
				}
			}

			/// <summary>
			/// Gets the response data.
			/// </summary>
			public TResponse ResponseData
			{
				get
				{
					return this._response;
				}
				private set
				{
					this._response = value;
					Trace.Assert(this.ResponseMessage != null, "this.ResponseMessage = null");
					this.ResponseMessage.Content = new StringContent(CommonUtils.SerializeAsJson(value), Encoding.UTF8, "application/json");
				}
			}

			/// <summary>
			/// Gets the response message.
			/// </summary>
			public string Message
			{
				get
				{
					return this._message;
				}
				private set
				{
					this._message = value;
					Trace.Assert(this.ResponseMessage != null, "this.ResponseMessage = null");
					this.ResponseMessage.Content = new StringContent(value);
				}
			}

			/// <summary>
			/// Gets the response headers.
			/// </summary>
			public HttpResponseHeaders Headers
			{
				get
				{
					Trace.Assert(this.ResponseMessage != null, "this.ResponseMessage = null");
					return this.ResponseMessage.Headers;
				}
			}

			#endregion Public Properties

			#region Public Methods

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();

				builder.AppendLine(this.GetType().GetFullName() + ".StatusCode: " + (int)this.StatusCode + " " + this.StatusCode.ToString());
				builder.AppendLine(this.GetType().GetFullName() + ".ResponseData: " + (this.ResponseData != null ? this.ResponseData.ToString().GetSubString() : "N/A"));

				return builder.ToString().TrimEnd();
			}

			#endregion Public Methods

			#region Private/Protected/Internal

			/// <summary>
			/// Gets the response message.
			/// </summary>
			internal HttpResponseMessage ResponseMessage { get; private set; }

			private TResponse _response;
			private string _message;

			/// <summary>
			/// Prevents a default instance of the <see cref="WebServiceResponse"/> class from being created.
			/// </summary>
			private WebServiceResponse()
			{
				this.ResponseMessage = new HttpResponseMessage(HttpStatusCode.OK);
				this._response = default(TResponse);
				this._message = null;
			}

			#endregion Private/Protected/Internal

			#region Public Static

			/// <summary>
			/// Creates a <see cref="WebServiceResponse"/> object with the specified parameters.
			/// </summary>
			/// <param name="statusCode">The status code.</param>
			/// <returns>The created <see cref="WebServiceResponse"/> instance.</returns>
			public static WebServiceResponse Create(HttpStatusCode statusCode)
			{
				return new WebServiceResponse()
				{
					Message = string.Empty,
					StatusCode = statusCode
				};
			}

			/// <summary>
			/// Creates a <see cref="WebServiceResponse"/> instance with the specified parameters.
			/// </summary>
			/// <param name="statusCode">The status code.</param>
			/// <param name="responseData">The response data.</param>
			/// <returns>The created <see cref="WebServiceResponse"/> instance.</returns>
			public static WebServiceResponse Create(HttpStatusCode statusCode, TResponse responseData)
			{
				return new WebServiceResponse()
				{
					ResponseData = responseData,
					StatusCode = statusCode
				};
			}

			/// <summary>
			/// Creates a <see cref="WebServiceResponse"/> instance with the specified parameters.
			/// </summary>
			/// <param name="statusCode">The status code.</param>
			/// <param name="message">The message.</param>
			/// <returns>The created <see cref="WebServiceResponse"/> instance.</returns>
			public static WebServiceResponse Create(HttpStatusCode statusCode, string message)
			{
				message = message ?? string.Empty; // the null string will be considered as IsMessageResponse = false

				return new WebServiceResponse()
				{
					Message = message,
					StatusCode = statusCode
				};
			}

			#endregion Public Static
		}
	}
}
