﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;

namespace VNC
{
	public enum FileProgressState
	{
		FPS_Transfering,
		FPS_Cancel,
		FPS_Complete,
		FPS_Error
	}

	public enum FileTransferType
	{
		UploadToServer,
		UploadFilesFinish,
		DownloadFromServer,
		DownloadFileFinish
	}

	// Windows Virtual-Key Codes
	// reference: http://msdn.microsoft.com/en-us/library/windows/desktop/dd375731%28v=vs.85%29.aspx
	public enum VIRTUAL_KEY_TYPE
	{
		BACKSPACE = 0x08,	// = VK_BACK = 0x08
		ENTER = 0x0D,		// = VK_RETURN = 0x0D
		ESC = 0x1B,			// = VK_ESCAPE = 0x1B
		PAGE_UP = 0x21,		// = VK_PRIOR = 0x21
		PAGE_DOWN = 0x22,	// = VK_NEXT = 0x22
		END = 0x23,			// = VK_END = 0x23
		HOME = 0x24,		// = VK_HOME = 0x24
		DELETE = 0x2E,		// = VK_DELETE = 0x2E
		History = 0xE9
		// ...
	}

	[StructLayout(LayoutKind.Sequential, Pack = 1)]
	public class SCREEN_SIZE_INFO
	{
		public uint uiWidth;	// The current width of Miracast screen on Air Fusion UI
		public uint uiHeight;	// The current height of Miracast screen on Air Fusion UI
		public int iDeltaX;	// 在X軸上 Miracast Screen 距離Air Fusion Windows 原點座標的偏移量
		public int iDeltaY;	// 在Y軸上 Miracast Screen 距離Air Fusion Windows 原點座標的偏移量
		public float fDpiRate;	// DPI rate (Ex: If DPI = 125%, fDpiRate = 100/125 = 0.8
	}

	/// <summary>
	/// 為了用來計算現在的 File Transfer Total Progress
	/// </summary>
	public class FTCalcProgressInfo
	{
		private int _iTotalFilesCount;
		private int _iSendCompleteFilesCount;
		private float _fOneFilePercentUnit;
		private int _iCurrentTotalProgress;		// 暫存現在 Uploading Files 的 Total Progress

		public FTCalcProgressInfo(int TotalFilesCount)
		{
			_iTotalFilesCount = TotalFilesCount;

			_fOneFilePercentUnit = 100.0f / (float)TotalFilesCount;
			_iSendCompleteFilesCount = 0;
			_iCurrentTotalProgress = 0;
		}

		public int TotalFilesCount { get { return _iTotalFilesCount; } }
		public float OneFilePercentUnit { get { return _fOneFilePercentUnit; } }

		public int SendCompleteFilesCount
		{
			set { _iSendCompleteFilesCount = value; }
			get { return _iSendCompleteFilesCount; }
		}

		public int CurrentTotalProgress
		{
			set { _iCurrentTotalProgress = value; }
			get { return _iCurrentTotalProgress; }
		}
	}

	/// <summary>
	/// File Transfer Update Request Item class
	/// - It will be added into one Queue (i.e. VNCClient::FTProgUpdateEventQueue)
	/// </summary>
	public class FTUpdateItem
	{
		public FTUpdateItem(int vncId, FileTransferType type)
		{
			_iVncID = vncId;
			_TransferType = type;
		}

		public FTUpdateItem(int vncId, FileTransferType type, string fileName, int progress, FileProgressState state)
		{
			_iVncID = vncId;
			_TransferType = type;
			_sDownloadFileName = fileName;
			_iProgressValue = progress;
			_state = state;
		}

		public FTUpdateItem(int vncId, FileTransferType type, int fileId, int progress, FileProgressState state)
		{
			_iVncID = vncId;
			_TransferType = type;
			_iFileID = fileId;
			_iProgressValue = progress;
			_state = state;
		}

		private int _iVncID;
		private FileTransferType _TransferType;
		private int _iProgressValue;

		private int _iFileID;								// For Upload File Case
		private string _sDownloadFileName = string.Empty;	// For Download File Case
		private FileProgressState _state;

		public int VncID { get { return _iVncID; } }
		public FileTransferType Type { get { return _TransferType; } }
		public int FileID { get { return _iFileID; } }
		public int ProgressValue { get { return _iProgressValue; } }
		public string DownloadFileName { get { return _sDownloadFileName; } }

		public int CurrentCompleteFileCount = -1;
		public FileProgressState State { get { return _state; } }
	}

	/// <summary>
	/// The request item that sending VNC Handler Message
	/// - It will be added into the 'VNCClient::VNCMsgEventQueue' Queue
	/// </summary>
	public class VNCMsgItem
	{
		private int _iVncID;

		public VNCMsgItem(int vncId, VNC_MSG_TYPE msg_type, VNC_FUN_CODE fun_code, VNC_ERR_CODE err_code, string msg)
		{
			_iVncID = vncId;
			MsgType = msg_type;
			FunctionCode = fun_code;
			ErrorCode = err_code;
			MessageText = msg;
		}

		public int VncID { get { return _iVncID; } }
		public VNC_MSG_TYPE MsgType;
		public VNC_FUN_CODE FunctionCode;
		public VNC_ERR_CODE ErrorCode;
		public string MessageText;
	}
}
