#include "RfbKeySym.h"
#include "keysymdef.h"

#define XK_MISCELLANY

RfbKeySym::RfbKeySym(ClientConnection *mclient, Log *mlog)
:	m_allowProcessCharEvent(false),
	m_leftMetaIsPressed(false),
	m_rightMetaIsPressed(false),
	m_client(mclient),
	m_log(mlog)
{
	clearKeyState();
}

RfbKeySym::~RfbKeySym()
{
}

void RfbKeySym::processKeyEvent(unsigned short virtKey, unsigned int addKeyData)
{
	m_log->Print(8, _T("processKeyEvent() function called: virtKey = %#4.4x, addKeyData") _T(" = %#x\n"), (unsigned int)virtKey, addKeyData);

	// Ignoring win key (when fullscreen mode is off).
	if (virtKey == VK_LWIN || virtKey == VK_RWIN) { // Ignoring the Win key
		m_log->Print(8, _T("Ignoring the Win key event\n"));
		return;
	}

	bool down = (addKeyData & 0x80000000) == 0;
	m_log->Print(8, _T("down = %u\n"), (unsigned int)down);

	bool ctrlPressed = isPressed(VK_LCONTROL) || isPressed(VK_RCONTROL);
	bool altPressed = isPressed(VK_LMENU) || isPressed(VK_RMENU);
	bool shiftPressed = isPressed(VK_LSHIFT) || isPressed(VK_RSHIFT);
	bool capsToggled = (GetKeyState(VK_CAPITAL) & 1) != 0;
	m_log->Print(8, _T("ctrl = %u, alt = %u, shift = %u, caps toggled = %u\n"),
					(unsigned int)ctrlPressed,
					(unsigned int)altPressed,
					(unsigned int)shiftPressed,
					(unsigned int)capsToggled);

	// Without distinguishing between left and right modifiers.
	m_viewerKeyState[virtKey & 255] = down ? 128 : 0;
	m_viewerKeyState[VK_CAPITAL & 255] = capsToggled ? 1 : 0;

	bool extended = (addKeyData & 0x1000000) != 0; // 24 bit
	virtKey = distinguishLeftRightModifier((unsigned char)virtKey, extended);

	// With distinguishing between left and right modifiers.
	m_serverKeyState[virtKey & 255] = down ? 128 : 0;

	UINT32 rfbSym;
	if (m_keyMap.virtualCodeToKeySym(&rfbSym, virtKey & 255)) {
		// Special case for VK_RETURN that have no self numpad code.
		// FIXME: May be replace this code to the virtualCodeToKeySym() function?
		if (rfbSym == XK_Return && extended) {
			rfbSym = XK_KP_Enter;
		}

		m_log->Print(8, _T("The key has been mapped to the %#4.4x rfb symbol\n"), rfbSym);
		sendKeySymEvent(rfbSym, down);
	}
	else
	{
		WCHAR outBuff[20];
		HKL currentLayout = GetKeyboardLayout(0);

		int count = ToUnicodeEx(virtKey, 0, m_viewerKeyState, outBuff,
								sizeof(outBuff) / sizeof(WCHAR), 0, currentLayout);

		if (count > 0) {
			count = ToUnicodeEx(virtKey, 0, m_viewerKeyState, outBuff,
								sizeof(outBuff) / sizeof(WCHAR), 0, currentLayout);
		}

		m_log->Print(8, _T("ToUnicodeEx() return %d\n"), count);

		if (count == 1 && !m_allowProcessCharEvent || count > 1) {
			bool ctrlAltPressed = ctrlPressed && altPressed;
			bool onlyCtrlPressed = ctrlPressed && !altPressed;

			if (ctrlAltPressed) {
				m_log->Print(8, _T("Release the ctrl and alt") _T(" modifiers before send the key event(s)\n"));
				releaseModifiers();
			}

			for (int i = 0; i < count; i++) {
				if (onlyCtrlPressed && outBuff[i] < 32) {
					if (onlyCtrlPressed && outBuff[i] >= 1 && outBuff[i] <= 26 && !shiftPressed) {
						m_log->Print(8, _T("The %u char is a control symbol then")
										_T(" it will be increased by 96 to %u\n"), (unsigned int)outBuff[i], (unsigned int)outBuff[i] + 96);
						outBuff[i] += 96;
					} else {
						m_log->Print(8, _T("The %u char is a control symbol then")
										_T(" it will be increased by 64 to %u\n"), (unsigned int)outBuff[i], (unsigned int)outBuff[i] + 64);
						outBuff[i] += 64;
					}
				}

				if (m_keyMap.unicodeCharToKeySym(outBuff[i], &rfbSym)) {
					m_log->Print(8, _T("Sending the %#4.4x rfb symbol\n"), rfbSym);
					sendKeySymEvent(rfbSym, down);
				} else {
					m_log->Print(8, _T("[Error] Can't translate the %#4.4x unicode character to an")
									_T(" rfb symbol to send it\n"), (unsigned int)outBuff[i]);
				}
			}

			if (ctrlAltPressed) {
				m_log->Print(8, _T("Restore the ctrl and alt")
								_T(" modifiers after send the key event(s)\n"));
				restoreModifiers();
			}
		}
		else if (count == 0)
		{
			m_log->Print(8, _T("Was get a not printable symbol then try get a printable")
							_T(" with turned off the ctrl and alt modifiers\n"));

			// E.g if pressed Ctrl + Alt + A
			// Try found char without modificators
			unsigned char withoutCtrlAltKbdState[256];
			memcpy(withoutCtrlAltKbdState, m_viewerKeyState, sizeof(withoutCtrlAltKbdState));
			withoutCtrlAltKbdState[VK_LCONTROL] = 0;
			withoutCtrlAltKbdState[VK_RCONTROL] = 0;
			withoutCtrlAltKbdState[VK_CONTROL] = 0;
			withoutCtrlAltKbdState[VK_LMENU] = 0;
			withoutCtrlAltKbdState[VK_RMENU] = 0;
			withoutCtrlAltKbdState[VK_MENU] = 0;

			count = ToUnicodeEx(virtKey, 0, withoutCtrlAltKbdState, outBuff,
								sizeof(outBuff) / sizeof(WCHAR), 0, currentLayout);
			m_log->Print(8, _T("ToUnicodeEx() without ctrl and alt return %d\n"), count);

			if (count == 1) { // other case will be ignored
				if (m_keyMap.unicodeCharToKeySym(outBuff[0], &rfbSym)) {
					sendKeySymEvent(rfbSym, down);
				} else {
					m_log->Print(8, _T("[Error] Can't translate the %#4.4x unicode character to an")
									_T(" rfb symbol to send it\n"), (unsigned int)outBuff[0]);
				}
			}
		}
		else if (count == -1 && down)
		{
			m_log->Print(8, _T("Dead key pressed, wait for a char event\n"));
			m_allowProcessCharEvent = true;
		}
	}
}

void RfbKeySym::processCharEvent(WCHAR charCode, unsigned int addKeyData)
{
	if (m_allowProcessCharEvent) {
		m_log->Print(8, _T("processCharEvent() function called with alowed processing:")
						_T(" charCode = %#4.4x, addKeyData = %#x\n"), (unsigned int)charCode, addKeyData);

		m_allowProcessCharEvent = false;

		UINT32 rfbSym;
		if (m_keyMap.unicodeCharToKeySym(charCode, &rfbSym))
		{
			sendKeySymEvent(rfbSym, true);
			sendKeySymEvent(rfbSym, false);
		}
		else
		{
			m_log->Print(8, _T("[Error] Can't translate the %#4.4x unicode character to an")
							_T(" rfb symbol to send it\n"), (unsigned int)charCode);
		}
	}
}

void RfbKeySym::processVirtualKey(unsigned short virtKey)
{
	UINT32 rfbSym;
	if (m_keyMap.virtualCodeToKeySym(&rfbSym, virtKey & 255))
		sendKeySymEvent(rfbSym, true);
	else
		m_log->Print(8, _T("[Error][RfbKeySym::processVirtualKey] Not mapped ! virtKey = %d\n"), virtKey);
}

void RfbKeySym::releaseModifiers()
{
	releaseModifier(VK_LCONTROL);
	releaseModifier(VK_RCONTROL);
	releaseModifier(VK_LMENU);
	releaseModifier(VK_RMENU);
}

void RfbKeySym::restoreModifiers()
{
	restoreModifier(VK_LCONTROL);
	restoreModifier(VK_RCONTROL);
	restoreModifier(VK_LMENU);
	restoreModifier(VK_RMENU);
}

void RfbKeySym::releaseModifier(unsigned char modifier)
{
	UINT32 rfbSym;
	if (isPressed(modifier)) {
		bool success = m_keyMap.virtualCodeToKeySym(&rfbSym, modifier);
		_ASSERT(success);
		sendKeySymEvent(rfbSym, false);
	}
}

void RfbKeySym::releaseMeta()
{
	if (m_leftMetaIsPressed) {
		sendKeySymEvent(XK_Meta_L, false);
	}
	if (m_rightMetaIsPressed) {
		sendKeySymEvent(XK_Meta_R, false);
	}
}

void RfbKeySym::restoreMeta()
{
	if (m_leftMetaIsPressed) {
		sendKeySymEvent(XK_Meta_L, true);
	}
	if (m_rightMetaIsPressed) {
		sendKeySymEvent(XK_Meta_R, true);
	}
}

void RfbKeySym::restoreModifier(unsigned char modifier)
{
	UINT32 rfbSym;
	if (isPressed(modifier)) {
		bool success = m_keyMap.virtualCodeToKeySym(&rfbSym, modifier);
		_ASSERT(success);
		sendKeySymEvent(rfbSym, true);
	}
}

void RfbKeySym::clearKeyState()
{
		memset(m_viewerKeyState, 0, sizeof(m_viewerKeyState));
	  memset(m_serverKeyState, 0, sizeof(m_serverKeyState));
}

bool RfbKeySym::isPressed(unsigned char virtKey)
{
	return (m_serverKeyState[virtKey] & 128) != 0;
}

void RfbKeySym::sendKeySymEvent(unsigned int rfbKeySym, bool down)
{
  // Translate Alt to Meta if Scroll Lock is on.
  if (rfbKeySym == XK_Alt_L || rfbKeySym == XK_Alt_R) {
    bool scrollLockOn = (GetKeyState(VK_SCROLL) & 1) != 0;
    if (scrollLockOn) {
      if (rfbKeySym == XK_Alt_L) {
        rfbKeySym = XK_Meta_L;
        m_leftMetaIsPressed = down;
      } else { // assuming (rfbKeySym == XK_Alt_R)
        rfbKeySym = XK_Meta_R;
        m_rightMetaIsPressed = down;
      }
    }
  }

  // Send the keyboard event.
  sendVerbatimKeySymEvent(rfbKeySym, down);
}

void RfbKeySym::sendVerbatimKeySymEvent(unsigned int rfbKeySym, bool down)
{
	////m_extKeySymListener->onRfbKeySymEvent(rfbKeySym, down);
	sendKeyboardEvent(down, rfbKeySym);
}

unsigned char RfbKeySym::distinguishLeftRightModifier(unsigned char virtKey, bool isRightHint)
{
  if (virtKey == VK_CONTROL) {
    virtKey = isRightHint ? VK_RCONTROL : VK_LCONTROL;
  }
  if (virtKey == VK_MENU) {
    virtKey = isRightHint ? VK_RMENU : VK_LMENU;
  }
  if (virtKey == VK_SHIFT) {
    virtKey = isRightHint ? VK_RSHIFT : VK_LSHIFT;
  }
  return virtKey;
}

// [***][TightVNC v2.7.10] RemoteViewerCore.cpp: void RemoteViewerCore::sendKeyboardEvent(...)
// Solution: Directly call ClientConnection::sendKeyboardEvent(...)
void RfbKeySym::sendKeyboardEvent(bool downFlag, UINT32 key)
{
	m_client->sendKeyboardEvent(downFlag, key);
}